@@ -7,8 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#ifndef INCLUDED_UTIL_HPP
-#define INCLUDED_UTIL_HPP
+#pragma once
 
 #include <cassert>
 #include <cerrno>
@@ -22,6 +21,7 @@
 #include <set>
 #include <sstream>
 #include <string>
+#include <map>
 #include <inttypes.h>
 
 #include <memory.h>
@@ -32,12 +32,13 @@
 
 #include <Poco/File.h>
 #include <Poco/Path.h>
-#include <Poco/Process.h>
 #include <Poco/RegularExpression.h>
 
 #define LOK_USE_UNSTABLE_API
 #include <LibreOfficeKit/LibreOfficeKitEnums.h>
 
+#include <StringVector.hpp>
+
 namespace Util
 {
     namespace rng
@@ -68,8 +69,9 @@ namespace Util
 
     /// Spawn a process if stdInput is non-NULL it contains a writable descriptor
     /// to send data to the child.
-    int spawnProcess(const std::string &cmd, const std::vector<std::string> &args,
-                     int *stdInput = nullptr);
+    int spawnProcess(const std::string &cmd, const StringVector &args,
+                     const std::vector<int>* fdsToKeep = nullptr, int *stdInput = nullptr);
+
 #endif
 
     /// Hex to unsigned char
@@ -133,18 +135,25 @@ namespace Util
     size_t getTotalSystemMemoryKb();
 
     /// Returns the process PSS in KB (works only when we have perms for /proc/pid/smaps).
-    size_t getMemoryUsagePSS(const Poco::Process::PID pid);
+    size_t getMemoryUsagePSS(const pid_t pid);
 
     /// Returns the process RSS in KB.
-    size_t getMemoryUsageRSS(const Poco::Process::PID pid);
+    size_t getMemoryUsageRSS(const pid_t pid);
 
     /// Returns the RSS and PSS of the current process in KB.
     /// Example: "procmemstats: pid=123 rss=12400 pss=566"
     std::string getMemoryStats(FILE* file);
 
-    size_t getCpuUsage(const Poco::Process::PID pid);
+    /// Reads from SMaps file Pss and Private_Dirty values and
+    /// returns them as a pair in the same order
+    std::pair<size_t, size_t> getPssAndDirtyFromSMaps(FILE* file);
 
-    size_t getStatFromPid(const Poco::Process::PID pid, int ind);
+    size_t getCpuUsage(const pid_t pid);
+
+    size_t getStatFromPid(const pid_t pid, int ind);
+
+    /// Sets priorities for a given pid & the current thread
+    void setProcessAndThreadPriorities(const pid_t pid, int prio);
 #endif
 
     std::string replace(std::string s, const std::string& a, const std::string& b);
@@ -938,9 +947,12 @@ int main(int argc, char**argv)
     /// Convert a time_point to iso8601 formatted string.
     std::string time_point_to_iso8601(std::chrono::system_clock::time_point tp);
 
-    //// Convert time from ISO8061 fraction format
+    /// Convert time from ISO8061 fraction format
     std::chrono::system_clock::time_point iso8601ToTimestamp(const std::string& iso8601Time, const std::string& logName);
 
+    /// conversion from steady_clock for debugging / tracing
+    std::string getSteadyClockAsString(const std::chrono::steady_clock::time_point &time);
+
     /// Automatically execute code at end of current scope.
     /// Used for exception-safe code.
     class ScopeGuard
@@ -960,8 +972,43 @@ int main(int argc, char**argv)
 
         std::function<void()> m_func;
     };
+
+    /**
+     * Avoid using the configuration layer and rely on defaults which is only useful for special
+     * test tool targets (typically fuzzing) where start-up speed is critical.
+     */
+    bool isFuzzing();
+
+    /**
+     * Splits string into vector<string>. Does not accept referenced variables for easy
+     * usage like (splitString("test", ..)) or (splitString(getStringOnTheFly(), ..))
+     */
+    inline std::vector<std::string> splitStringToVector(std::string const str, const char delim)
+    {
+        size_t start;
+        size_t end = 0;
+
+        std::vector<std::string> result;
+
+        while ((start = str.find_first_not_of(delim, end)) != std::string::npos)
+        {
+            end = str.find(delim, start);
+            result.push_back(str.substr(start, end - start));
+        }
+        return result;
+    }
+
+    /**
+     * Converts vector of strings to map. Strings should have formed like this: key + delimiter + value.
+     * In case of a misformed string or zero length vector, passes that item and warns the developer.
+     */
+    std::map<std::string, std::string> stringVectorToMap(std::vector<std::string> sVector, const char delimiter);
+
+    #if !MOBILEAPP
+        // If OS is not mobile, it must be Linux.
+        std::string getLinuxVersion();
+    #endif
+
 } // end namespace Util
 
-#endif
-
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
