@@ -3,7 +3,7 @@
 * Control.ColumnHeader
 */
 
-/* global $ _UNO */
+/* global $ _UNO Hammer */
 L.Control.ColumnHeader = L.Control.Header.extend({
 	options: {
 		cursor: 'col-resize'
@@ -36,6 +36,7 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		this._canvasContext = this._canvas.getContext('2d');
 		this._setCanvasWidth();
 		this._setCanvasHeight();
+		this._canvasBaseHeight = this._canvasHeight;
 
 		var scale = L.getDpiScaleFactor();
 		this._canvasContext.scale(scale, scale);
@@ -60,84 +61,82 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		this._startOffset = 0;
 		this._position = 0;
 
-		L.DomEvent.on(this._cornerCanvas, 'contextmenu', L.DomEvent.preventDefault);
 		L.DomEvent.addListener(this._cornerCanvas, 'click', this._onCornerHeaderClick, this);
 
-		var colHeaderObj = this;
-		L.installContextMenu({
-			selector: '.spreadsheet-header-columns',
-			className: 'loleaflet-font',
-			items: {
-				'insertcolbefore': {
-					name: _UNO('.uno:InsertColumnsBefore', 'spreadsheet', true),
-					callback: function() {
-						var index = colHeaderObj._lastMouseOverIndex;
-						if (index) {
-							colHeaderObj.insertColumn.call(colHeaderObj, index);
-						}
-					}
-				},
-				'deleteselectedcol': {
-					name: _UNO('.uno:DeleteColumns', 'spreadsheet', true),
-					callback: function() {
-						var index = colHeaderObj._lastMouseOverIndex;
-						if (index) {
-							colHeaderObj.deleteColumn.call(colHeaderObj, index);
-						}
-					}
-				},
-				'optimalwidth': {
-					name: _UNO('.uno:SetOptimalColumnWidth', 'spreadsheet', true),
-					callback: function() {
-						var index = colHeaderObj._lastMouseOverIndex;
-						if (index) {
-							colHeaderObj.optimalWidth.call(colHeaderObj, index);
-						}
-					}
-				},
-				'hideColumn': {
-					name: _UNO('.uno:HideColumn', 'spreadsheet', true),
-					callback: function() {
-						var index = colHeaderObj._lastMouseOverIndex;
-						if (index) {
-							colHeaderObj.hideColumn.call(colHeaderObj, index);
-						}
-					}
-				},
-				'showColumn': {
-					name: _UNO('.uno:ShowColumn', 'spreadsheet', true),
-					callback: function() {
-						var index = colHeaderObj._lastMouseOverIndex;
-						if (index) {
-							colHeaderObj.showColumn.call(colHeaderObj, index);
-						}
-					}
-				}
+		this._menuItem = {
+			'.uno:InsertColumnsBefore': {
+				name: _UNO('.uno:InsertColumnsBefore', 'spreadsheet', true),
+				callback: (this._insertColBefore).bind(this)
 			},
-			zIndex: 10
-		});
+			'.uno:InsertColumnsAfter': {
+				name: _UNO('.uno:InsertColumnsAfter', 'spreadsheet', true),
+				callback: (this._insertColAfter).bind(this)
+			},
+			'.uno:DeleteColumns': {
+				name: _UNO('.uno:DeleteColumns', 'spreadsheet', true),
+				callback: (this._deleteSelectedCol).bind(this)
+			},
+			'.uno:SetOptimalColumnWidth': {
+				name: _UNO('.uno:SetOptimalColumnWidth', 'spreadsheet', true),
+				callback: (this._optimalWidth).bind(this)
+			},
+			'.uno:HideColumn': {
+				name: _UNO('.uno:HideColumn', 'spreadsheet', true),
+				callback: (this._hideColumn).bind(this)
+			},
+			'.uno:ShowColumn': {
+				name: _UNO('.uno:ShowColumn', 'spreadsheet', true),
+				callback: (this._showColumn).bind(this)
+			}
+		};
+
+		if (!window.mode.isMobile()) {
+			L.installContextMenu({
+				selector: '.spreadsheet-header-columns',
+				className: 'loleaflet-font',
+				items: this._menuItem,
+				zIndex: 10
+			});
+		} else {
+			var menuData = L.Control.JSDialogBuilder.getMenuStructureForMobileWizard(this._menuItem, true, '');
+			(new Hammer(this._canvas, {recognizers: [[Hammer.Press]]}))
+			.on('press', L.bind(function () {
+				if (this._map._permission === 'edit') {
+					window.contextMenuWizard = true;
+					this._map.fire('mobilewizard', menuData);
+				}
+			}, this));
+		}
 	},
 
 	optimalWidth: function(index) {
-		if (this._map._docLayer._selections.getLayers().length === 0) {
+		if (!this.isHighlighted(index)) {
 			this._selectColumn(index, 0);
 		}
 		this._map.sendUnoCommand('.uno:SetOptimalColumnWidth');
 	},
 
-	insertColumn: function(index) {
+	insertColumnBefore: function(index) {
 		// First select the corresponding column because
 		// .uno:InsertColumn doesn't accept any column number
 		// as argument and just inserts before the selected column
-		if (this._map._docLayer._selections.getLayers().length === 0) {
+		if (!this.isHighlighted(index)) {
 			this._selectColumn(index, 0);
 		}
 		this._map.sendUnoCommand('.uno:InsertColumns');
 		this._updateColumnHeader();
 	},
 
+	insertColumnAfter: function(index) {
+		if (!this.isHighlighted(index)) {
+			this._selectColumn(index, 0);
+		}
+		this._map.sendUnoCommand('.uno:InsertColumnsAfter');
+		this._updateColumnHeader();
+	},
+
 	deleteColumn: function(index) {
-		if (this._map._docLayer._selections.getLayers().length === 0) {
+		if (!this.isHighlighted(index)) {
 			this._selectColumn(index, 0);
 		}
 		this._map.sendUnoCommand('.uno:DeleteColumns');
@@ -145,7 +144,7 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 	},
 
 	hideColumn: function(index) {
-		if (this._map._docLayer._selections.getLayers().length === 0) {
+		if (!this.isHighlighted(index)) {
 			this._selectColumn(index, 0);
 		}
 		this._map.sendUnoCommand('.uno:HideColumn');
@@ -153,7 +152,7 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 	},
 
 	showColumn: function(index) {
-		if (this._map._docLayer._selections.getLayers().length === 0) {
+		if (!this.isHighlighted(index)) {
 			this._selectColumn(index, 0);
 		}
 		this._map.sendUnoCommand('.uno:ShowColumn');
@@ -256,7 +255,7 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		}
 		// draw text content
 		ctx.fillStyle = isHighlighted ? this._selectionTextColor : this._textColor;
-		ctx.font = this._font;
+		ctx.font = this._font.getFont();
 		ctx.textAlign = 'center';
 		ctx.textBaseline = 'middle';
 		// The '+ 1' below is a hack - it's currently not possible to measure
@@ -348,7 +347,7 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		ctx.strokeRect(startPar, startOrt, ctrlHeadSize, ctrlHeadSize);
 		// draw level number
 		ctx.fillStyle = this._textColor;
-		ctx.font = this._font;
+		ctx.font = this._font.getFont();
 		ctx.textAlign = 'center';
 		ctx.textBaseline = 'middle';
 		ctx.fillText(level + 1, startPar + (ctrlHeadSize / 2), startOrt + (ctrlHeadSize / 2));
@@ -390,12 +389,22 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		this._setCanvasWidth();
 		this._setCanvasHeight();
 		this._canvasContext.clearRect(0, 0, canvas.width, canvas.height);
+		if (this._map._permission === 'edit') {
+			// Adjust (column) _headerHeight according to zoomlevel. This is used below to call resize()
+			// where column/corner header are resized. Besides the document container and row header container
+			// are moved up or down as required so that there is no gap/overlap below column header.
+			// Limit zoomScale so that the column header is not too small (unreadable) or too big.
+			this._headerHeight = Math.ceil(this._canvasBaseHeight *
+				this.getHeaderZoomScale(/* lowerBound */ 0.74, /* upperBound */ 1.15));
+		}
 
 		// Reset state
 		this._current = -1;
 		this._selection.start = this._selection.end = -1;
 		this._mouseOverEntry = null;
-		this._lastMouseOverIndex = undefined;
+		if (!window.contextMenuWizard) {
+			this._lastMouseOverIndex = undefined;
+		}
 
 		// create data structure for column widths
 		this._tickMap = new L.Control.Header.GapTickMap(this._map, columns);
@@ -430,7 +439,6 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 
 		this.mouseInit(canvas);
 
-		L.DomEvent.on(canvas, 'contextmenu', L.DomEvent.preventDefault);
 		if ($('.spreadsheet-header-columns').length > 0) {
 			$('.spreadsheet-header-columns').contextMenu(this._map._permission === 'edit');
 		}
@@ -649,13 +657,58 @@ L.Control.ColumnHeader = L.Control.Header.extend({
 		var rowHdrTop = parseInt(L.DomUtil.getStyle(rowHeader, 'top')) + deltaTop;
 		var docTop = parseInt(L.DomUtil.getStyle(document, 'top')) + deltaTop;
 		L.DomUtil.setStyle(rowHeader, 'top', rowHdrTop + 'px');
-		L.DomUtil.setStyle(document, 'top', docTop + 'px');
+		// L.DomUtil.setStyle does not seem to affect the attributes when
+		// one of the media queries of document-container element are
+		// active (non-desktop case). Using style.setProperty directly
+		// seems to work as expected for both mobile/desktop.
+		document.style.setProperty('top', docTop + 'px', 'important');
 
 		this._setCanvasHeight(height);
 
 		this._map.fire('updatecornerheader');
-	}
+	},
 
+	_insertColBefore: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.insertColumnBefore.call(this, index);
+		}
+	},
+
+	_insertColAfter: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.insertColumnAfter.call(this, index);
+		}
+	},
+
+	_deleteSelectedCol: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.deleteColumn.call(this, index);
+		}
+	},
+
+	_optimalWidth: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.optimalWidth.call(this, index);
+		}
+	},
+
+	_hideColumn: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.hideColumn.call(this, index);
+		}
+	},
+
+	_showColumn: function() {
+		var index = this._lastMouseOverIndex;
+		if (index) {
+			this.showColumn.call(this, index);
+		}
+	}
 });
 
 L.control.columnHeader = function (options) {
