@@ -71,7 +71,25 @@ L.Control.Header = L.Control.extend({
 		var fontSize = parseInt(L.DomUtil.getStyle(elem, 'font-size'));
 		var fontHeight = parseInt(L.DomUtil.getStyle(elem, 'line-height'));
 		var rate = fontHeight / fontSize;
-		this._font = fontSize + 'px/' + rate + ' ' + fontFamily;
+		this._font = {
+			_hdr: this,
+			_baseFontSize: fontSize,
+			_fontSizeRate: rate,
+			_fontFamily: fontFamily,
+			getFont: function() {
+				// Limit zoomScale to 115%. At 120% the row ids at the bottom eat all
+				// horizontal margins and it looks ugly. Beyond 120% the row ids get
+				// clipped out visibly.
+				var zoomScale = this._hdr.getHeaderZoomScale(
+					/* lowerBound */ 0.5, /* upperBound */ 1.15);
+
+				return Math.floor(this._baseFontSize * zoomScale) +
+					'px/' +
+					this._fontSizeRate +
+					' ' +
+					this._fontFamily;
+			}
+		};
 		this._borderColor = L.DomUtil.getStyle(elem, 'border-top-color');
 		var borderWidth = L.DomUtil.getStyle(elem, 'border-top-width');
 		this._borderWidth = Math.round(parseFloat(borderWidth));
@@ -222,7 +240,7 @@ L.Control.Header = L.Control.extend({
 	// Called whenever the cell cursor is in a cell corresponding to the cursorPos-th
 	// column/row.
 	updateCurrent: function (cursorPos, slim) {
-		if (!this._tickMap) {return}
+		if (!this._tickMap) {return;}
 
 		if (cursorPos < 0) {
 			this.unselect(this._tickMap.getGap(this._current));
@@ -248,7 +266,7 @@ L.Control.Header = L.Control.extend({
 	},
 
 	_onPan: function (event) {
-		if (event.pointerType != 'touch')
+		if (event.pointerType !== 'touch' || this._map._permission !== 'edit')
 			return;
 
 		if (event.type == 'panstart')
@@ -724,6 +742,20 @@ L.Control.Header = L.Control.extend({
 		}
 	},
 
+	getHeaderZoomScale : function(lowerBound, upperBound) {
+		if (typeof lowerBound === 'undefined' || lowerBound < 0)
+			lowerBound = 0.5;
+		if (typeof upperBound === 'undefined' || upperBound < 0)
+			upperBound = 2.0;
+		if (lowerBound > upperBound) {
+			lowerBound = 0.5;
+			upperBound = 2.0;
+		}
+		var zoomScale = this._map.getZoomScale(this._map.getZoom(),
+			this._map.options.defaultZoom);
+		return Math.min(Math.max(zoomScale, lowerBound), upperBound);
+	},
+
 	onDragStart: function () {},
 	onDragMove: function () {},
 	onDragEnd: function () {},
@@ -845,7 +877,7 @@ L.Control.Header.GapTickMap = L.Class.extend({
 			end: end,
 			size: end - start,
 			pos: end,
-		}
+		};
 	},
 
 	// Returns true when the i-th gap has zero size.
