@@ -3,13 +3,13 @@
  * L.Control.PartsPreview
  */
 
-/* global $ */
+/* global $ Hammer w2ui */
 L.Control.PartsPreview = L.Control.extend({
 	options: {
 		fetchThumbnail: true,
 		autoUpdate: true,
-		maxWidth: window.mode.isMobile() ? 60 : 180,
-		maxHeight: window.mode.isMobile() ? 60 : 180
+		maxWidth: (window.mode.isMobile() || window.mode.isTablet()) ? 60 : 180,
+		maxHeight: (window.mode.isMobile() || window.mode.isTablet()) ? 60 : 180
 	},
 	partsFocused: false,
 
@@ -58,7 +58,6 @@ L.Control.PartsPreview = L.Control.extend({
 				}
 			}
 		});
-		$(this._partsPreviewCont).data('preview', this);
 	},
 
 	_updateDisabled: function (e) {
@@ -66,7 +65,7 @@ L.Control.PartsPreview = L.Control.extend({
 		var selectedPart = e.selectedPart;
 		var selectedParts = e.selectedParts;
 		var docType = e.docType;
-		if (docType === 'text') {
+		if (docType === 'text' || isNaN(parts)) {
 			return;
 		}
 
@@ -151,6 +150,27 @@ L.Control.PartsPreview = L.Control.extend({
 						L.DomUtil.addClass(this._previewTiles[j], 'preview-img-selectedpart');
 				}
 			}
+
+			// update portrait / landscape
+			var removePreviewImg = 'preview-img-portrait';
+			var addPreviewImg = 'preview-img-landscape';
+			var removePreviewFrame = 'preview-frame-portrait';
+			var addPreviewFrame = 'preview-frame-landscape';
+			if (L.DomUtil.isPortrait()) {
+				removePreviewImg = 'preview-img-landscape';
+				addPreviewImg = 'preview-img-portrait';
+				removePreviewFrame = 'preview-frame-landscape';
+				addPreviewFrame = 'preview-frame-portrait';
+			}
+
+			for (i = 0; i < parts; i++) {
+				L.DomUtil.removeClass(this._previewTiles[i], removePreviewImg);
+				L.DomUtil.addClass(this._previewTiles[i], addPreviewImg);
+			}
+
+			var previewFrame = $(this._partsPreviewCont).find('.preview-frame');
+			previewFrame.removeClass(removePreviewFrame);
+			previewFrame.addClass(addPreviewFrame);
 		}
 	},
 
@@ -164,14 +184,24 @@ L.Control.PartsPreview = L.Control.extend({
 		img.hash = hashCode;
 		img.src = L.Icon.Default.imagePath + '/preview_placeholder.png';
 		img.fetched = false;
-		L.DomEvent
-			.on(img, 'click', L.DomEvent.stopPropagation)
-			.on(img, 'click', L.DomEvent.stop)
-			.on(img, 'click', this._setPart, this)
-			.on(img, 'click', this._map.focus, this._map)
-			.on(img, 'click', function() {
-				this.partsFocused = true;
-			}, this);
+		if (window.mode.isMobile() || window.mode.isTablet()) {
+			(new Hammer(img, {recognizers: [[Hammer.Press]]}))
+			.on('press', L.bind(function () {
+				if (this._map._permission === 'edit') {
+					setTimeout(function () {
+						w2ui['actionbar'].click('mobile_wizard');
+					}, 0);
+				}
+			}, this));
+		}
+		L.DomEvent.on(img, 'click', function (e) {
+			L.DomEvent.stopPropagation(e);
+			L.DomEvent.stop(e);
+			this._setPart(e);
+			this._map.focus();
+			this.partsFocused = true;
+			document.activeElement.blur();
+		}, this);
 
 		var topBound = this._previewContTop;
 		var previewFrameTop = 0;
@@ -190,7 +220,7 @@ L.Control.PartsPreview = L.Control.extend({
 			previewFrameBottom = previewFrameTop + this._previewFrameHeight;
 
 			if (this._direction === 'x') {
-				L.DomUtil.setStyle(img, 'width', this._previewImgHeight + 'px');
+				L.DomUtil.setStyle(img, 'width', this._previewImgWidth + 'px');
 			} else {
 				L.DomUtil.setStyle(img, 'height', this._previewImgHeight + 'px');
 			}
@@ -357,7 +387,8 @@ L.Control.PartsPreview = L.Control.extend({
 		if (this._map.getDocType() === 'presentation' || this._map.getDocType() === 'drawing') {
 			if (!this._previewInitialized)
 				return;
-			this._previewTiles[e.id].src = e.tile;
+			if (this._previewTiles[e.id])
+				this._previewTiles[e.id].src = e.tile;
 		}
 	},
 
@@ -393,41 +424,43 @@ L.Control.PartsPreview = L.Control.extend({
 	},
 
 	_onScroll: function (e) {
-		var scrollOffset = 0;
-		if (e) {
-			var prevScrollY = this._scrollY;
-			this._scrollY = this._direction === 'x' ? -e.mcs.left : -e.mcs.top;
-			scrollOffset = this._scrollY - prevScrollY;
-		}
+		setTimeout(L.bind(function (e) {
+			var scrollOffset = 0;
+			if (e) {
+				var prevScrollY = this._scrollY;
+				this._scrollY = this._direction === 'x' ? -e.mcs.left : -e.mcs.top;
+				scrollOffset = this._scrollY - prevScrollY;
+			}
 
-		var previewContBB = this._partsPreviewCont.getBoundingClientRect();
-		var extra =  this._direction === 'x' ? previewContBB.width : previewContBB.height;
-		var topBound = this._previewContTop - (scrollOffset < 0 ? extra : extra / 2);
-		var bottomBound = this._previewContTop + extra + (scrollOffset > 0 ? extra : extra / 2);
-		for (var i = 0; i < this._previewTiles.length; ++i) {
-			var img = this._previewTiles[i];
-			if (img && img.parentNode && !img.fetched) {
-				var previewFrameBB = img.parentNode.getBoundingClientRect();
-				if (this._direction === 'x') {
-					if ((previewFrameBB.left >= topBound && previewFrameBB.left <= bottomBound)
-					|| (previewFrameBB.right >= topBound && previewFrameBB.right <= bottomBound)) {
+			var previewContBB = this._partsPreviewCont.getBoundingClientRect();
+			var extra =  this._direction === 'x' ? previewContBB.width : previewContBB.height;
+			var topBound = this._previewContTop - (scrollOffset < 0 ? extra : extra / 2);
+			var bottomBound = this._previewContTop + extra + (scrollOffset > 0 ? extra : extra / 2);
+			for (var i = 0; i < this._previewTiles.length; ++i) {
+				var img = this._previewTiles[i];
+				if (img && img.parentNode && !img.fetched) {
+					var previewFrameBB = img.parentNode.getBoundingClientRect();
+					if (this._direction === 'x') {
+						if ((previewFrameBB.left >= topBound && previewFrameBB.left <= bottomBound)
+						|| (previewFrameBB.right >= topBound && previewFrameBB.right <= bottomBound)) {
+							this._map.getPreview(i, i, this.options.maxWidth, this.options.maxHeight, {autoUpdate: this.options.autoUpdate});
+							img.fetched = true;
+						}
+					} else if ((previewFrameBB.top >= topBound && previewFrameBB.top <= bottomBound)
+						|| (previewFrameBB.bottom >= topBound && previewFrameBB.bottom <= bottomBound)) {
 						this._map.getPreview(i, i, this.options.maxWidth, this.options.maxHeight, {autoUpdate: this.options.autoUpdate});
 						img.fetched = true;
 					}
-				} else if ((previewFrameBB.top >= topBound && previewFrameBB.top <= bottomBound)
-					|| (previewFrameBB.bottom >= topBound && previewFrameBB.bottom <= bottomBound)) {
-					this._map.getPreview(i, i, this.options.maxWidth, this.options.maxHeight, {autoUpdate: this.options.autoUpdate});
-					img.fetched = true;
 				}
 			}
-		}
+		}, this, e), 0);
 	},
 
 	_addDnDHandlers: function (elem) {
 		if (elem) {
 			elem.setAttribute('draggable', true);
 			elem.addEventListener('dragstart', this._handleDragStart, false);
-			elem.addEventListener('dragenter', this._handleDragEnter, false)
+			elem.addEventListener('dragenter', this._handleDragEnter, false);
 			elem.addEventListener('dragover', this._handleDragOver, false);
 			elem.addEventListener('dragleave', this._handleDragLeave, false);
 			elem.addEventListener('drop', this._handleDrop, false);
@@ -439,7 +472,7 @@ L.Control.PartsPreview = L.Control.extend({
 	_handleDragStart: function (e) {
 		// To avoid having to add a new message to move an arbitrary part, let's select the
 		// slide that is being dragged.
-		var part = $(this._partsPreviewCont).find('.mCSB_container .preview-frame').index(e.target.parentNode);
+		var part = $(this.partsPreview._partsPreviewCont).find('.mCSB_container .preview-frame').index(e.target.parentNode);
 		if (part !== null) {
 			var partId = parseInt(part) - 1; // The first part is just a drop-site for reordering.
 			this.partsPreview._map.setPart(partId);
@@ -475,7 +508,7 @@ L.Control.PartsPreview = L.Control.extend({
 			e.stopPropagation();
 		}
 
-		var part = $(this._partsPreviewCont).find('.mCSB_container .preview-frame').index(e.target.parentNode);
+		var part = $(this.partsPreview._partsPreviewCont).find('.mCSB_container .preview-frame').index(e.target.parentNode);
 		if (part !== null) {
 			var partId = parseInt(part) - 1; // First frame is a drop-site for reordering.
 			if (partId < 0)
