@@ -22,6 +22,8 @@ L.Socket = L.Class.extend({
 		console.debug('socket.initialize:');
 		this._map = map;
 		this._msgQueue = [];
+		this._delayedMessages = [];
+		this._handlingDelayedMessages = false;
 	},
 
 	getWebSocketBaseURI: function(map) {
@@ -47,7 +49,7 @@ L.Socket = L.Class.extend({
 			}
 
 			try {
-				this.socket = new WebSocket(this.getWebSocketBaseURI(map) + wopiSrc);
+				this.socket = window.createWebSocket(this.getWebSocketBaseURI(map) + wopiSrc);
 			} catch (e) {
 				// On IE 11 there is a limitation on the number of WebSockets open to a single domain (6 by default and can go to 128).
 				// Detect this and hint the user.
@@ -74,8 +76,11 @@ L.Socket = L.Class.extend({
 		}
 
 		// process messages for early socket connection
-		if (socket && ((socket.readyState === 1 || socket.readyState === 0)) &&
-			window.queueMsg && window.queueMsg.length > 0) {
+		this._emptyQueue();
+	},
+
+	_emptyQueue: function () {
+		if (window.queueMsg && window.queueMsg.length > 0) {
 			for (var it = 0; it < window.queueMsg.length; it++) {
 				this._onMessage({data: window.queueMsg[it]});
 			}
@@ -99,6 +104,11 @@ L.Socket = L.Class.extend({
 		                                            120 * 1000);
 	},
 
+	setUnloading: function() {
+		if (this.socket.setUnloading)
+			this.socket.setUnloading();
+	},
+
 	close: function () {
 		this.socket.onerror = function () {};
 		this.socket.onclose = function () {};
@@ -115,7 +125,7 @@ L.Socket = L.Class.extend({
 		return this.socket && this.socket.readyState === 1;
 	},
 
-	sendMessage: function (msg, coords) {
+	sendMessage: function (msg) {
 		if (this._map._fatal) {
 			// Avoid communicating when we're in fatal state
 			return;
@@ -136,29 +146,18 @@ L.Socket = L.Class.extend({
 		}
 
 		if (socketState === 1) {
-			this.socket.send(msg);
-			// Only attempt to log text frames, not binary ones.
-			if (typeof msg === 'string') {
-				L.Log.log(msg, L.OUTGOING, coords);
-				if (this._map._docLayer && this._map._docLayer._debug) {
-					console.log2(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c '), 'background:#fbb;color:black', 'color:red', 'color:black');
-				}
-			}
+			this._doSend(msg);
 		}
 		else {
 			// push message while trying to connect socket again.
-			this._msgQueue.push({msg: msg, coords: coords});
+			this._msgQueue.push(msg);
 		}
 	},
 
-	_doSend: function(msg, coords) {
+	_doSend: function(msg) {
 		// Only attempt to log text frames, not binary ones.
-		if (typeof msg === 'string') {
-			L.Log.log(msg, L.OUTGOING, coords);
-			if (this._map._docLayer && this._map._docLayer._debug) {
-				console.log2(+new Date() + ' %cOUTGOING%c: ' + msg.concat(' ').replace(' ', '%c '), 'background:#fbb;color:black', 'color:red', 'color:black');
-			}
-		}
+		if (typeof msg === 'string')
+			this._logSocket('OUTGOING', msg);
 
 		this.socket.send(msg);
 	},
@@ -193,15 +192,19 @@ L.Socket = L.Class.extend({
 		if (String.locale) {
 			msg += ' lang=' + String.locale;
 		}
+		if (window.deviceFormFactor) {
+			msg += ' deviceFormFactor=' + window.deviceFormFactor;
+		}
 		if (this._map.options.renderingOptions) {
 			var options = {
 				'rendering': this._map.options.renderingOptions
 			};
 			msg += ' options=' + JSON.stringify(options);
 		}
+
 		this._doSend(msg);
 		for (var i = 0; i < this._msgQueue.length; i++) {
-			this._doSend(this._msgQueue[i].msg, this._msgQueue[i].coords);
+			this._doSend(this._msgQueue[i]);
 		}
 		this._msgQueue = [];
 
@@ -237,10 +240,28 @@ L.Socket = L.Class.extend({
 		return true;
 	},
 
+	_logSocket: function(type, msg) {
+		var fullDebug = this._map._docLayer && this._map._docLayer._debug;
+		if (!window.protocolDebug && !fullDebug)
+			return;
+
+		if (!fullDebug && msg.length > 256) // for reasonable performance.
+			msg = msg.substring(0,256) + '<truncated ' + (msg.length - 256) + 'chars>';
+
+		var status = '';
+		if (!window.fullyLoadedAndReady)
+			status += '[!fullyLoadedAndReady]';
+		if (!window.bundlejsLoaded)
+			status += '[!bundlejsLoaded]';
+
+		var color = type === 'OUTGOING' ? 'color:red' : 'color:blue';
+		console.log2(+new Date() + ' %c' + type + status + '%c: ' + msg.concat(' ').replace(' ', '%c '),
+			     'background:#ddf;color:black', color, 'color:black');
+	},
+
 	_onMessage: function (e) {
 		var imgBytes, index, textMsg, img;
 
-		console.info('onMessage: window.fullyLoadedAndReady: ' + window.fullyLoadedAndReady + ', bundlejsLoaded: ' + window.bundlejsLoaded);
 		if (typeof (e.data) === 'string') {
 			textMsg = e.data;
 		}
@@ -254,9 +275,7 @@ L.Socket = L.Class.extend({
 			textMsg = String.fromCharCode.apply(null, imgBytes.subarray(0, index));
 		}
 
-		if (this._map._docLayer && this._map._docLayer._debug) {
-			console.log2(+new Date() + ' %cINCOMING%c: ' + textMsg.concat(' ').replace(' ', '%c '), 'background:#ddf;color:black', 'color:blue', 'color:black');
-		}
+		this._logSocket('INCOMING', textMsg);
 
 		var command = this.parseServerCmd(textMsg);
 		if (textMsg.startsWith('loolserver ')) {
@@ -264,19 +283,19 @@ L.Socket = L.Class.extend({
 			this.WSDServer = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
 			var h = this.WSDServer.Hash;
 			if (parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
-				if (!window.ThisIsTheiOSApp) {
-					h = '<a target="_blank" href="https://hub.libreoffice.org/git-online/' + h + '">' + h + '</a>';
-				}
+				h = '<a target="_blank" href="javascript:window.open(\'https://hub.libreoffice.org/git-online/' + h + '\');">' + h + '</a>';
 				$('#loolwsd-version').html(this.WSDServer.Version + ' (git hash: ' + h + ')');
 			}
 			else {
 				$('#loolwsd-version').text(this.WSDServer.Version);
 			}
 
-			var idUri = this._map.options.server + this._map.options.serviceRoot + '/hosting/discovery';
-			idUri = idUri.replace(/^ws:/, 'http:');
-			idUri = idUri.replace(/^wss:/, 'https:');
-			$('#loolwsd-id').html(_('Served by:') + ' <a target="_blank" href="' + idUri + '">' + this.WSDServer.Id + '</a>');
+			if (!window.ThisIsAMobileApp) {
+				var idUri = this._map.options.server + this._map.options.serviceRoot + '/hosting/discovery';
+				idUri = idUri.replace(/^ws:/, 'http:');
+				idUri = idUri.replace(/^wss:/, 'https:');
+				$('#loolwsd-id').html(_('Served by:') + ' <a target="_blank" href="' + idUri + '">' + this.WSDServer.Id + '</a>');
+			}
 
 			// TODO: For now we expect perfect match in protocol versions
 			if (this.WSDServer.Protocol !== this.ProtocolVersionNumber) {
@@ -286,16 +305,21 @@ L.Socket = L.Class.extend({
 		else if (textMsg.startsWith('lokitversion ')) {
 			var lokitVersionObj = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
 			h = lokitVersionObj.BuildId.substring(0, 7);
-			if (!window.ThisIsTheiOSApp && parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
-				h = '<a target="_blank" href="https://hub.libreoffice.org/git-core/' + h + '">' + h + '</a>';
+			if (parseInt(h,16).toString(16) === h.toLowerCase().replace(/^0+/, '')) {
+				h = '<a target="_blank" href="javascript:window.open(\'https://hub.libreoffice.org/git-core/' + h + '\');">' + h + '</a>';
 			}
 			$('#lokit-version').html(lokitVersionObj.ProductName + ' ' +
 			                         lokitVersionObj.ProductVersion + lokitVersionObj.ProductExtension.replace('.10.','-') +
 			                         ' (git hash: ' + h + ')');
 		}
+		else if (textMsg.startsWith('osinfo ')) {
+			var osInfo = textMsg.replace('osinfo ', '');
+			document.getElementById('os-info').innerText = osInfo;
+		}
 		else if (textMsg.startsWith('clipboardkey: ')) {
 			var key = textMsg.substring('clipboardkey: '.length);
-			this._map._clip.setKey(key);
+			if (this._map._clip)
+				this._map._clip.setKey(key);
 		}
 		else if (textMsg.startsWith('perm:')) {
 			var perm = textMsg.substring('perm:'.length);
@@ -334,9 +358,16 @@ L.Socket = L.Class.extend({
 		}
 		else if (textMsg.startsWith('commandresult: ')) {
 			var commandresult = JSON.parse(textMsg.substring(textMsg.indexOf('{')));
-			if (commandresult['command'] === 'savetostorage' && commandresult['success']) {
-				// Close any open confirmation dialogs
-				vex.closeAll();
+			if (commandresult['command'] === 'savetostorage' || commandresult['command'] === 'save') {
+				if (commandresult['success']) {
+					// Close any open confirmation dialogs
+					vex.closeAll();
+				}
+
+				var postMessageObj = {
+					success: commandresult['success']
+				};
+				this._map.fire('postMessage', {msgId: 'Action_Save_Resp', args: postMessageObj});
 			}
 			return;
 		}
@@ -436,18 +467,30 @@ L.Socket = L.Class.extend({
 				message = msg;
 			}
 
-			var dialogOpened = vex.dialog.open({
+			var dialogOptions = {
 				message: message,
 				contentClassName: 'loleaflet-user-idle'
-			});
+			};
 
+			var restartConnectionFn;
 			if (textMsg === 'idle' || textMsg === 'oom') {
 				var map = this._map;
-				dialogOpened.contentEl.onclick = function() {
-					console.debug('idleness: reactivating');
-					map._documentIdle = false;
-					return map._activate();
+				restartConnectionFn = function() {
+					if (map._documentIdle)
+					{
+						console.debug('idleness: reactivating');
+						map._documentIdle = false;
+						return map._activate();
+					}
+					return false;
 				};
+				dialogOptions.afterClose = restartConnectionFn;
+			}
+
+			var dialogOpened = vex.dialog.open(dialogOptions);
+
+			if (textMsg === 'idle' || textMsg === 'oom') {
+				dialogOpened.contentEl.onclick = restartConnectionFn;
 				$('.vex-overlay').addClass('loleaflet-user-idle-overlay');
 			}
 
@@ -626,6 +669,7 @@ L.Socket = L.Class.extend({
 							this._map._docPassword = '';
 							this._map.loadDocument();
 						} else {
+							this._map.fire('postMessage', {msgId: 'UI_Cancel_Password'});
 							this._map.hideBusy();
 						}
 					}, this)
@@ -707,18 +751,12 @@ L.Socket = L.Class.extend({
 						}
 					});
 				}
-
-				// Issue the save response to be consistent with normal save.
-				var postMessageObj = {
-					success: true
-				};
-				this._map.fire('postMessage', {msgId: 'Action_Save_Resp', args: postMessageObj});
 			}
 			// var name = command.name; - ignored, we get the new name via the wopi's BaseFileName
 		}
 		else if (textMsg.startsWith('statusindicator:')) {
 			//FIXME: We should get statusindicator when saving too, no?
-			this._map.showBusy(_('Connecting...'), true);
+			this._map.showBusy(window.ThisIsAMobileApp? _('Loading...'): _('Connecting...'), true);
 			if (textMsg.startsWith('statusindicator: ready')) {
 				// We're connected: cancel timer and dialog.
 				this.ReconnectCount = 0;
@@ -726,9 +764,14 @@ L.Socket = L.Class.extend({
 				vex.closeAll();
 			}
 		}
+		else if (window.ThisIsAMobileApp && textMsg.startsWith('mobile:')) {
+			// allow passing some events easily from the mobile app
+			var mobileEvent = textMsg.substring('mobile: '.length);
+			this._map.fire(mobileEvent);
+		}
 		else if (!textMsg.startsWith('tile:') && !textMsg.startsWith('renderfont:') && !textMsg.startsWith('windowpaint:')) {
 			// log the tile msg separately as we need the tile coordinates
-			L.Log.log(textMsg, L.INCOMING);
+			this._logSocket('INCOMING', textMsg);
 
 			if (imgBytes !== undefined) {
 				try {
@@ -803,11 +846,59 @@ L.Socket = L.Class.extend({
 			return;
 		}
 
-		if (this._map._docLayer) {
+		var msgDelayed = false;
+		if (!this._isReady() || this._delayedMessages.length || this._handlingDelayedMessages) {
+			msgDelayed = this._tryToDelayMessage(textMsg);
+		}
+
+		if (this._map._docLayer && !msgDelayed) {
 			this._map._docLayer._onMessage(textMsg, img);
 		}
 	},
 
+	_tryToDelayMessage: function(textMsg) {
+		var delayed = false;
+		if (textMsg.startsWith('window:') ||
+			textMsg.startsWith('celladdress:') ||
+			textMsg.startsWith('statechanged:')) {
+			//console.log('_tryToDelayMessage: textMsg: ' + textMsg);
+			var message = {msg: textMsg};
+			this._delayedMessages.push(message);
+			delayed  = true;
+		}
+
+		if (delayed && !this._delayedMsgHandlerTimeoutId) {
+			this._handleDelayedMessages();
+		}
+		return delayed;
+	},
+
+	_handleDelayedMessages: function() {
+		if (!this._isReady() || this._handlingDelayedMessages) {
+			var that = this;
+			// Retry in a bit.
+			this._delayedMsgHandlerTimeoutId = setTimeout(function() {
+				that._handleDelayedMessages();
+			}, 100);
+			return;
+		}
+		var messages = [];
+		for (var i = 0; i < this._delayedMessages.length; ++i) {
+			var message = this._delayedMessages[i];
+			if (message)
+				messages.push(message.msg);
+		}
+		this._delayedMessages = [];
+		this._delayedMsgHandlerTimeoutId = null;
+		this._handlingDelayedMessages = true;
+		if (this._map._docLayer) {
+			for (var k = 0; k < messages.length; ++k) {
+				this._map._docLayer._onMessage(messages[k]);
+			}
+		}
+		this._handlingDelayedMessages = false;
+	},
+
 	_delayedFitToScreen: function(height) {
 		if (this._map.getSize().y > 0) {
 			// If we have a presentation document and the zoom level has not been set
