@@ -11,7 +11,7 @@ L.GridLayer = L.Layer.extend({
 		tileSize: window.tileSize,
 		opacity: 1,
 
-		updateWhenIdle: L.Browser.mobile,
+		updateWhenIdle: (window.mode.isMobile() || window.mode.isTablet()),
 		updateInterval: 200,
 
 		attribution: null,
@@ -25,7 +25,7 @@ L.GridLayer = L.Layer.extend({
 	initialize: function (options) {
 		L.setOptions(this, options);
 
-		this._resetClientVisArea()
+		this._resetClientVisArea();
 	},
 
 	onAdd: function () {
@@ -211,12 +211,7 @@ L.GridLayer = L.Layer.extend({
 		    maxZoom = this.options.maxZoom;
 
 		for (var z in this._levels) {
-			if (this._levels[z].el.children.length || z === zoom) {
-				this._levels[z].el.style.zIndex = maxZoom - Math.abs(zoom - z);
-			} else {
-				L.DomUtil.remove(this._levels[z].el);
-				delete this._levels[z];
-			}
+			this._levels[z].el.style.zIndex = maxZoom - Math.abs(zoom - z);
 		}
 
 		var level = this._levels[zoom],
@@ -322,6 +317,9 @@ L.GridLayer = L.Layer.extend({
 
 	_viewReset: function (e) {
 		this._reset(this._map.getCenter(), this._map.getZoom(), e && e.hard);
+		if (this._docType === 'spreadsheet' && this._annotations !== 'undefined') {
+			this._map._socket.sendMessage('commandvalues command=.uno:ViewAnnotationsPosition');
+		}
 	},
 
 	_animateZoom: function (e) {
@@ -390,7 +388,7 @@ L.GridLayer = L.Layer.extend({
 		var bottomRight = new L.Point(docPixelLimits.x, docPixelLimits.y);
 		bottomRight = bottomRight.multiplyBy(scale);
 		if (extraSize) {
-			// extraSize is unscalled.
+			// extraSize is unscaled.
 			bottomRight = bottomRight.add(extraSize);
 		}
 		bottomRight = this._map.unproject(bottomRight);
@@ -405,7 +403,7 @@ L.GridLayer = L.Layer.extend({
 			this._docHeightTwips / this._tileHeightTwips);
 		scrollPixelLimits = scrollPixelLimits.multiplyBy(this._tileSize);
 		if (extraSize) {
-			// extraSize is unscalled.
+			// extraSize is unscaled.
 			scrollPixelLimits = scrollPixelLimits.add(extraSize);
 		}
 		this._docPixelSize = {x: scrollPixelLimits.x, y: scrollPixelLimits.y};
@@ -436,13 +434,14 @@ L.GridLayer = L.Layer.extend({
 		}
 		else if (this._map.options._origMaxBounds) {
 			// if after zoomimg the document becomes larger than the viewing area
-			// we need to restore the inital bounds
+			// we need to restore the initial bounds
 			this._map.setMaxBounds(this._map.options._origMaxBounds);
 			this._map.options._origMaxBounds = null;
 		}
 	},
 
 	_updateScrollOffset: function () {
+		if (!this._map) return;
 		var centerPixel = this._map.project(this._map.getCenter());
 		var newScrollPos = centerPixel.subtract(this._map.getSize().divideBy(2));
 		var x = Math.round(newScrollPos.x < 0 ? 0 : newScrollPos.x);
@@ -713,10 +712,12 @@ L.GridLayer = L.Layer.extend({
 			this._map.fire('updatescrolloffset', {x: 0, y: 0, updateHeaders: false});
 			this._map.scrollTop(0);
 			this._map.scrollLeft(0);
-			this._cellCursor = L.LatLngBounds.createDefault();
-			this._prevCellCursor = new L.LatLngBounds(new L.LatLng(0, 0), new L.LatLng(1, 1));
-			this._cellCursorXY = new L.Point(-1, -1);
-			this._prevCellCursorXY = new L.Point(0, 0);
+
+			// 2020-01-14 if no side effect occurs remove this code later
+			// this._cellCursor = L.LatLngBounds.createDefault();
+			// this._prevCellCursor = new L.LatLngBounds(new L.LatLng(0, 0), new L.LatLng(1, 1));
+			// this._cellCursorXY = new L.Point(-1, -1);
+			// this._prevCellCursorXY = new L.Point(0, 0);
 		}
 	},
 
@@ -744,7 +745,7 @@ L.GridLayer = L.Layer.extend({
 			// Visible area is dirty, update it on the server
 			this._map._socket.sendMessage(newClientVisibleArea);
 			if (!this._map._fatal && this._map._active && this._map._socket.connected())
-				this._clientVisibleArea = newClientVisibleArea
+				this._clientVisibleArea = newClientVisibleArea;
 			if (this._debug) {
 				this._debugInfo.clearLayers();
 				for (var key in this._tiles) {
@@ -983,7 +984,7 @@ L.GridLayer = L.Layer.extend({
 				var current = coordsQueue[i];
 
 				// extend the bound vertically if possible (so far it was
-				// continous)
+				// continuous)
 				if (!hasHole && (current.y === bound.y + 1)) {
 					rowLocked = true;
 					++bound.y;
