@@ -3,7 +3,7 @@
  * L.Clipboard is used to abstract our storage and management of
  * local & remote clipboard data.
  */
-/* global _ vex brandProductName */
+/* global _ vex brandProductName isAnyVexDialogActive */
 
 // Get all interesting clipboard related events here, and handle
 // download logic in one place ...
@@ -39,8 +39,8 @@ L.Clipboard = L.Class.extend({
 		this._resetDiv();
 
 		var that = this;
-		var beforeSelect = function(ev) { return that._beforeSelect(ev); }
-		if (window.isInternetExplorer)
+		var beforeSelect = function(ev) { return that._beforeSelect(ev); };
+		if (L.Browser.isInternetExplorer)
 		{
 			document.addEventListener('cut',   function(ev)   { return that.cut(ev); });
 			document.addEventListener('copy',  function(ev)   { return that.copy(ev); });
@@ -61,7 +61,7 @@ L.Clipboard = L.Class.extend({
 	},
 
 	compatRemoveNode: function(node) {
-		if (window.isInternetExplorer)
+		if (L.Browser.isInternetExplorer)
 			node.removeNode(true);
 		else // standard
 			node.parentNode.removeChild(node);
@@ -99,6 +99,10 @@ L.Clipboard = L.Class.extend({
 			'&Tag=' + this._accessKey[idx];
 	},
 
+	getMetaURL: function(idx) {
+		return this.getMetaBase() + this.getMetaPath(idx);
+	},
+
 	// Returns the marker used to identify stub messages.
 	_getHtmlStubMarker: function() {
 		return '<title>Stub HTML Message</title>';
@@ -111,7 +115,7 @@ L.Clipboard = L.Class.extend({
 
 	// wrap some content with our stub magic
 	_originWrapBody: function(body, isStub) {
-		var encodedOrigin = encodeURIComponent(this.getMetaPath());
+		var encodedOrigin = encodeURIComponent(this.getMetaURL());
 		var text =  '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">\n' +
 		            '<html>\n' +
 		            '  <head>\n';
@@ -250,7 +254,7 @@ L.Clipboard = L.Class.extend({
 	// Suck the data from one server to another asynchronously ...
 	_dataTransferDownloadAndPasteAsync: function(src, dest, fallbackHtml) {
 		var that = this;
-		// FIXME: add a timestamp in the links (?) ignroe old / un-responsive servers (?)
+		// FIXME: add a timestamp in the links (?) ignore old / un-responsive servers (?)
 		that._doAsyncDownload(
 			'GET', src, null,
 			function(response) {
@@ -261,7 +265,7 @@ L.Clipboard = L.Class.extend({
 					'POST', dest, formData,
 					function() {
 						console.log('up-load done, now paste');
-						that._map._socket.sendMessage('uno .uno:Paste')
+						that._map._socket.sendMessage('uno .uno:Paste');
 					},
 					function(progress) { return 50 + progress/2; }
 				);
@@ -289,7 +293,7 @@ L.Clipboard = L.Class.extend({
 					'POST', dest, formData,
 					function() {
 						console.log('up-load of fallback done, now paste');
-						that._map._socket.sendMessage('uno .uno:Paste')
+						that._map._socket.sendMessage('uno .uno:Paste');
 					},
 					function(progress) { return 50 + progress/2; },
 					function() {
@@ -340,8 +344,7 @@ L.Clipboard = L.Class.extend({
 		if (meta !== '')
 		{
 			console.log('Transfer between servers\n\t"' + meta + '" vs. \n\t"' + id + '"');
-			var destination = this.getMetaBase() + this.getMetaPath();
-			this._dataTransferDownloadAndPasteAsync(meta, destination, htmlText);
+			this._dataTransferDownloadAndPasteAsync(meta, this.getMetaURL(), htmlText);
 			return;
 		}
 
@@ -381,7 +384,7 @@ L.Clipboard = L.Class.extend({
 					if (files !== null)
 					{
 						for (var f = 0; f < files.length; ++f)
-							this._asyncReadPasteImage(files[f])
+							this._asyncReadPasteImage(files[f]);
 					}
 					else // IE / Edge
 						this._asyncReadPasteImage(dataTransfer.items[t].getAsFile());
@@ -397,14 +400,13 @@ L.Clipboard = L.Class.extend({
 			formData.append('file', content);
 
 			var that = this;
-			var destination = this.getMetaBase() + this.getMetaPath();
-			this._doAsyncDownload('POST', destination, formData,
+			this._doAsyncDownload('POST', this.getMetaURL(), formData,
 							function() {
 								console.log('Posted ' + content.size + ' bytes successfully');
 								if (usePasteKeyEvent) {
 									// paste into dialog
 									var KEY_PASTE = 1299;
-									that._map._clipboardContainer._sendKeyEvent(0, KEY_PASTE);
+									that._map._textInput._sendKeyEvent(0, KEY_PASTE);
 								} else {
 									// paste into document
 									that._map._socket.sendMessage('uno .uno:Paste');
@@ -439,7 +441,7 @@ L.Clipboard = L.Class.extend({
 				text = this._getStubHtml();
 				this._onDownloadOnLargeCopyPaste();
 				this._downloadProgress.setURI( // richer, bigger HTML ...
-					this.getMetaBase() + this.getMetaPath() + '&MimeType=text/html');
+					this.getMetaURL() + '&MimeType=text/html');
 			}
 		} else if (this._selectionType === null) {
 			console.log('Copy/Cut with no selection!');
@@ -455,7 +457,7 @@ L.Clipboard = L.Class.extend({
 	populateClipboard: function(ev) {
 		this._checkSelection();
 
-		if (window.isInternetExplorer)
+		if (L.Browser.isInternetExplorer)
 		{
 			var that = this;
 			setTimeout(function() { that._resetDiv(); }, 0);
@@ -465,11 +467,12 @@ L.Clipboard = L.Class.extend({
 		}
 
 		var text = this._getHtmlForClipboard();
-//		this._stopHideDownload(); - this confuses the borwser ruins copy/cut on iOS
+//		this._stopHideDownload(); - this confuses the browser ruins copy/cut on iOS
 
 		var plainText = this.stripHTML(text);
 		if (ev.clipboardData) { // Standard
-			ev.clipboardData.setData('text/plain', plainText);
+			// if copied content is graphical then plainText is null and it does not work on mobile.
+			ev.clipboardData.setData('text/plain', plainText ? plainText: ' ');
 			ev.clipboardData.setData('text/html', text);
 			console.log('Put "' + text + '" on the clipboard');
 			this._clipboardSerial++;
@@ -493,9 +496,9 @@ L.Clipboard = L.Class.extend({
 	},
 
 	_beforeSelectImpl: function(operation) {
-		if (window.isInternetExplorer && operation != 'paste')
+		if (L.Browser.isInternetExplorer && operation != 'paste')
 			// We need populate our content into the div for
-			// the brower to copy.
+			// the browser to copy.
 			this._dummyDiv.innerHTML = this._getHtmlForClipboard();
 		else
 			// We need some spaces in there ...
@@ -507,7 +510,7 @@ L.Clipboard = L.Class.extend({
 
 		var selected = false;
 		var selectRange;
-		if (window.isInternetExplorer && operation != 'paste')
+		if (L.Browser.isInternetExplorer && operation != 'paste')
 		{
 			this._dummyDiv.focus();
 
@@ -571,7 +574,7 @@ L.Clipboard = L.Class.extend({
 		var serial = this._clipboardSerial;
 
 		// try a direct execCommand.
-		if (window.isInternetExplorer && operation != 'paste')
+		if (L.Browser.isInternetExplorer && operation != 'paste')
 			this._beforeSelectImpl(operation);
 		if (document.execCommand(operation) &&
 		    serial !== this._clipboardSerial) {
@@ -608,15 +611,20 @@ L.Clipboard = L.Class.extend({
 			}
 			else
 			{
-				console.log('help did not arive for ' + operation);
+				console.log('help did not arrive for ' + operation);
 				that._warnCopyPaste();
 			}
-		}, 150 /* ms */)
+		}, 150 /* ms */);
 	},
 
 	// Pull UNO clipboard commands out from menus and normal user input.
 	// We try to massage and re-emit these, to get good security event / credentials.
 	filterExecCopyPaste: function(cmd) {
+		if (window.ThisIsAMobileApp) {
+			// We do native copy/paste in the iOS and Android cases
+			return false;
+		}
+
 		if (cmd === '.uno:Copy') {
 			this._execCopyCutPaste('copy');
 		} else if (cmd === '.uno:Cut') {
@@ -646,11 +654,19 @@ L.Clipboard = L.Class.extend({
 
 	paste: function(ev) {
 		console.log('Paste');
+
+		if (isAnyVexDialogActive() && !this._map.hasFocus())
+			return;
+
+		// If the focus is in the search box, paste there.
+		if (this._map.isSearching())
+			return;
+
 		if (this._map._activeDialog)
 			ev.usePasteKeyEvent = true;
 
 		var that = this;
-		if (window.isInternetExplorer)
+		if (L.Browser.isInternetExplorer)
 		{
 			var active = document.activeElement;
 			// Can't get HTML until it is pasted ... so quick timeout
@@ -661,7 +677,7 @@ L.Clipboard = L.Class.extend({
 					that._map.focus();
 				else
 					active.focus();
-				that._map._clipboardContainer._abortComposition(ev);
+				that._map._textInput._abortComposition(ev);
 				that._clipboardSerial++;
 			}, 0 /* ASAP */);
 			return false;
@@ -675,7 +691,7 @@ L.Clipboard = L.Class.extend({
 			// pass the clipboard data to a different context (async calls, f.e.).
 			var htmlText = ev.clipboardData.getData('text/html');
 			this.dataTransferToDocument(ev.clipboardData, /* preferInternal = */ true, htmlText, usePasteKeyEvent);
-			this._map._clipboardContainer._abortComposition(ev);
+			this._map._textInput._abortComposition(ev);
 			this._clipboardSerial++;
 			this._stopHideDownload();
 		}
@@ -700,6 +716,9 @@ L.Clipboard = L.Class.extend({
 	setTextSelectionHTML: function(html) {
 		this._selectionType = 'text';
 		this._selectionContent = html;
+		if (L.Browser.cypressTest) {
+			this._dummyDiv.innerHTML = html;
+		}
 	},
 
 	// sets the selection to some (cell formula) text)
@@ -761,12 +780,22 @@ L.Clipboard = L.Class.extend({
 	_warnCopyPaste: function() {
 		var self = this;
 		var msg;
-		if (L.Browser.mobile) {
+		if (window.mode.isMobile() || window.mode.isTablet()) {
 			msg = _('<p>Please use the copy/paste buttons on your on-screen keyboard.</p>');
 		} else {
 			msg = _('<p>Your browser has very limited access to the clipboard, so use these keyboard shortcuts:<ul><li><b>Ctrl+C</b>: For copying.</li><li><b>Ctrl+X</b>: For cutting.</li><li><b>Ctrl+V</b>: For pasting.</li></ul></p>');
 			if (navigator.appVersion.indexOf('Mac') != -1 || navigator.userAgent.indexOf('Mac') != -1) {
-				msg = msg.replace(/Ctrl/g, '⌘');
+				var ctrl = /Ctrl/g;
+				if (String.locale.startsWith('de') || String.locale.startsWith('dsb') || String.locale.startsWith('hsb')) {
+					ctrl = /Strg/g;
+				}
+				if (String.locale.startsWith('lt')) {
+					ctrl = /Vald/g;
+				}
+				if (String.locale.startsWith('sl')) {
+					ctrl = /Krmilka/g;
+				}
+				msg = msg.replace(ctrl, '⌘');
 			}
 		}
 		vex.dialog.alert({
@@ -814,5 +843,7 @@ L.Clipboard = L.Class.extend({
 });
 
 L.clipboard = function(map) {
+	if (window.ThisIsAMobileApp)
+		console.log('======> Assertion failed!? No L.Clipboard object should be needed in a mobile app');
 	return new L.Clipboard(map);
 };
