@@ -39,7 +39,7 @@ var unoCommandsArray = {
 	ContinueNumbering:{text:{menu:_('Continue previous numbering'),},},
 	ControlCodes:{text:{menu:_('For~matting Marks'),},},
 	Copy:{global:{menu:_('Cop~y'),},},
-	CopyHyperlinkLocation:{global:{menu:_('Copy Hyperlink Location'),},},
+	CopyHyperlinkLocation:{text:{menu:_('Copy Hyperlink Location'),},},
 	CustomAnimation:{presentation:{menu:_('Animation'),},},
 	Cut:{global:{menu:_('~Cut'),},},
 	DataBarFormatDialog:{spreadsheet:{menu:_('Data Bar...'),},},
@@ -49,6 +49,7 @@ var unoCommandsArray = {
 	DataFilterSpecialFilter:{spreadsheet:{menu:_('~Advanced Filter...'),},},
 	DataFilterStandardFilter:{spreadsheet:{menu:_('~Standard Filter...'),},},
 	DataMenu:{global:{menu:_('~Data'),},},
+	DataRanges:{global:{menu:_('~Data Ranges...'),},},
 	DataSort:{spreadsheet:{menu:_('~Sort...'),},},
 	DecrementIndent:{global:{context:_('Decrease Indent'),menu:_('Decrease Indent'),},},
 	DecrementLevel:{text:{menu:_('Demote One Level'),},},
@@ -56,16 +57,24 @@ var unoCommandsArray = {
 	DefaultBullet:{global:{context:_('Toggle Bulleted List'),menu:_('~Bulleted List'),},},
 	DefaultNumbering:{global:{context:_('Toggle Numbered List'),menu:_('~Numbered List'),},},
 	DelBreakMenu:{spreadsheet:{menu:_('Delete Page ~Break'),},},
+	Delete:{global:{menu:_('Delete C~ontents...'),},spreadsheet:{context:_('Cl~ear Contents...'),menu:_('Cle~ar Cells...'),},},
 	DeleteAllNotes:{global:{menu:_('Delete All Comments'),},spreadsheet:{menu:_('Delete All Comments'),},},
 	DeleteAuthor:{global:{menu:_('Delete All Comments by This Author'),},},
-	DeleteColumnbreak:{spreadsheet:{menu:_('~Column Break'),},},
+	DeleteAxis:{global:{menu:_('Delete Axis'),},},
+	DeleteColumnbreak:{spreadsheet:{menu:_('Remove ~Column Break'),},},
 	DeleteColumns:{presentation:{menu:_('Delete Column'),},spreadsheet:{menu:_('Delete Columns'),},text:{menu:_('~Columns'),},},
 	DeleteComment:{global:{menu:_('Delete Comment'),},},
+	DeleteDataLabel:{global:{menu:_('Delete Single Data Label'),},},
+	DeleteLegend:{global:{menu:_('Delete Legend'),},},
+	DeleteMajorGrid:{global:{menu:_('Delete Major Grid'),},},
+	DeleteMinorGrid:{global:{menu:_('Delete Minor Grid'),},},
 	DeleteNote:{spreadsheet:{menu:_('Delete Comment'),},},
-	DeleteRowbreak:{spreadsheet:{menu:_('~Row Break'),},},
+	DeleteRowbreak:{spreadsheet:{menu:_('Remove ~Row Break'),},},
 	DeleteRows:{presentation:{menu:_('Delete Row'),},spreadsheet:{menu:_('Delete Rows'),},text:{menu:_('~Rows'),},},
 	DeleteSlide:{presentation:{menu:_('~Delete Slide'),},},
 	DeleteTable:{text:{menu:_('~Table'),},},
+	DiagramData:{global:{menu:_('~Data Table...'),},},
+	DiagramType:{global:{menu:_('Chart T~ype...'),},},
 	DownSearch:{global:{menu:_('Find Next'),},},
 	DuplicateSlide:{presentation:{menu:_('Duplicate ~Slide'),},},
 	EditHeaderAndFooter:{spreadsheet:{menu:_('~Headers and Footers...'),},},
@@ -79,17 +88,36 @@ var unoCommandsArray = {
 	FontColor:{global:{menu:_('Font Color'),},text:{menu:_('Font Color'),},},
 	FontDialog:{global:{menu:_('C~haracter...'),},},
 	FormatArea:{global:{menu:_('A~rea...'),},},
+	FormatAxis:{global:{menu:_('Format Axis...'),},},
 	FormatBulletsMenu:{global:{menu:_('Lis~ts'),},},
 	FormatCellDialog:{spreadsheet:{context:_('~Format Cells...'),menu:_('Ce~lls...'),},},
+	FormatChartArea:{global:{menu:_('Format Chart Area...'),},},
 	FormatColumns:{text:{menu:_('Co~lumns...'),},},
+	FormatDataLabel:{global:{menu:_('Format Single Data Label...'),},},
+	FormatDataLabels:{global:{menu:_('Format Data Labels...'),},},
+	FormatDataPoint:{global:{menu:_('Format Data Point...'),},},
+	FormatDataSeries:{global:{menu:_('Format Data Series...'),},},
+	FormatFloor:{global:{menu:_('Format Floor...'),},},
+	FormatLegend:{global:{menu:_('Format Legend...'),},},
 	FormatLine:{global:{menu:_('L~ine...'),},},
+	FormatMajorGrid:{global:{menu:_('Format Major Grid...'),},},
+	FormatMeanValue:{global:{menu:_('Format Mean Value Line...'),},},
 	FormatMenu:{global:{menu:_('F~ormat'),},},
+	FormatMinorGrid:{global:{menu:_('Format Minor Grid...'),},},
 	FormatPaintbrush:{global:{context:_('Clone Formatting (double click for multi-selection)'),menu:_('Clone'),},text:{context:_('Clone Formatting (double click and Ctrl or Cmd to alter behavior)'),menu:_('Clone'),},},
+	FormatSelection:{global:{menu:_('Format Selection...'),},},
 	FormatSpacingMenu:{global:{menu:_('~Spacing'),},},
+	FormatStockGain:{global:{menu:_('Format Stock Gain...'),},},
+	FormatStockLoss:{global:{menu:_('Format Stock Loss...'),},},
 	FormatTextMenu:{global:{menu:_('Te~xt'),},},
+	FormatTitle:{global:{menu:_('Format Title...'),},},
+	FormatTrendline:{global:{menu:_('Format Trend Line...'),},},
+	FormatTrendlineEquation:{global:{menu:_('Format Trend Line Equation...'),},},
+	FormatWall:{global:{menu:_('Format Wall...'),},},
+	FormatXErrorBars:{global:{menu:_('Format X Error Bars...'),},},
+	FormatYErrorBars:{global:{menu:_('Format Y Error Bars...'),},},
 	FormattingMarkMenu:{global:{menu:_('Formatting Mark'),},},
 	FullScreen:{global:{menu:_('F~ull Screen'),},},
-	FunctionDialog:{spreadsheet:{menu:_('~Function...'),},},
 	GoalSeekDialog:{spreadsheet:{menu:_('~Goal Seek...'),},},
 	Group:{global:{menu:_('~Group...'),},},
 	GroupOutlineMenu:{spreadsheet:{menu:_('~Group and Outline'),},},
@@ -109,49 +137,71 @@ var unoCommandsArray = {
 	InsertAnnotation:{global:{context:_('Insert Comment'),menu:_('Comme~nt'),},presentation:{menu:_('Comme~nt'),},spreadsheet:{context:_('Insert Co~mment'),menu:_('Comm~ent'),},},
 	InsertAuthorField:{presentation:{menu:_('~Author'),},text:{menu:_('First ~Author'),},},
 	InsertAuthoritiesEntry:{text:{menu:_('~Bibliography Entry...'),},},
+	InsertAxis:{global:{menu:_('Insert Axis'),},},
+	InsertAxisTitle:{global:{menu:_('Insert Axis Title'),},},
 	InsertBreakMenu:{spreadsheet:{menu:_('Insert Page ~Break'),},},
 	InsertColumnBreak:{spreadsheet:{menu:_('~Column Break'),},text:{menu:_('Insert Column Break'),},},
 	InsertColumnsAfter:{presentation:{menu:_('Insert Column After'),},spreadsheet:{context:_('Insert Columns ~After'),menu:_('Columns ~After'),},text:{menu:_('Columns ~After'),},},
 	InsertColumnsBefore:{presentation:{menu:_('Insert Column Before'),},spreadsheet:{context:_('Insert Columns ~Before'),menu:_('Columns ~Before'),},text:{menu:_('Columns ~Before'),},},
 	InsertColumnsMenu:{spreadsheet:{menu:_('Insert Co~lumns'),},},
+	InsertCurrentDate:{spreadsheet:{menu:_('~Date'),},},
+	InsertCurrentTime:{spreadsheet:{menu:_('~Time'),},},
+	InsertDataLabel:{global:{menu:_('Insert Single Data Label'),},},
 	InsertDateField:{text:{menu:_('~Date'),},},
+	InsertDateFieldFix:{presentation:{menu:_('~Date (fixed)'),},},
+	InsertDateFieldVar:{presentation:{menu:_('Dat~e (variable)'),},},
 	InsertEndnote:{text:{menu:_('~Endnote'),},},
 	InsertField:{text:{menu:_('~More Fields...'),},},
 	InsertFootnote:{text:{menu:_('~Footnote'),},},
 	InsertGraphic:{global:{context:_('Insert Image...'),menu:_('~Image...'),},},
-	InsertHardHyphen:{global:{menu:_('Insert non-br~eaking hyphen'),},},
+	InsertHardHyphen:{global:{menu:_('Non-br~eaking hyphen'),},},
 	InsertHeaderFooterMenu:{text:{menu:_('He~ader and Footer'),},},
 	InsertIndexesEntry:{text:{menu:_('~Index Entry...'),},},
-	InsertLRM:{global:{menu:_('~Left-to-right Mark'),},},
+	InsertLRM:{global:{menu:_('~Left-to-right mark'),},},
+	InsertMajorGrid:{global:{menu:_('Insert Major Grid'),},},
+	InsertMeanValue:{global:{menu:_('Insert Mean ~Value Line'),},},
 	InsertMenu:{global:{menu:_('~Insert'),},},
+	InsertMinorGrid:{global:{menu:_('Insert Minor Grid'),},},
 	InsertMultiIndex:{text:{menu:_('Table of Contents, ~Index or Bibliography...'),},},
 	InsertNeutralParagraph:{text:{menu:_('Insert Unnumbered Entry'),},},
-	InsertNonBreakingSpace:{global:{menu:_('Insert ~non-breaking space'),},},
+	InsertNonBreakingSpace:{global:{menu:_('~Non-breaking space'),},},
 	InsertObjectChart:{global:{context:_('Insert Chart'),menu:_('~Chart...'),},},
 	InsertPageCountField:{text:{menu:_('Page ~Count'),},},
 	InsertPageFooter:{text:{menu:_('Foote~r'),},},
 	InsertPageHeader:{text:{menu:_('He~ader'),},},
 	InsertPageNumberField:{global:{menu:_('~Page Numbers...'),},text:{menu:_('~Page Number'),},},
 	InsertPagebreak:{text:{menu:_('~Page Break'),},},
-	InsertRLM:{global:{menu:_('~Right-to-left Mark'),},},
+	InsertRLM:{global:{menu:_('~Right-to-left mark'),},},
+	InsertRemoveAxes:{global:{menu:_('Insert/Delete Axes...'),},},
 	InsertRowBreak:{spreadsheet:{menu:_('~Row Break'),},},
 	InsertRowsAfter:{presentation:{menu:_('Insert Row Below'),},spreadsheet:{context:_('Insert Rows ~Below'),menu:_('Rows ~Below'),},text:{menu:_('Rows ~Below'),},},
 	InsertRowsBefore:{presentation:{menu:_('Insert Row Above'),},spreadsheet:{context:_('Insert Rows ~Above'),menu:_('Rows ~Above'),},text:{menu:_('Rows ~Above'),},},
 	InsertRowsMenu:{spreadsheet:{menu:_('Insert ~Rows'),},},
 	InsertSection:{text:{menu:_('Se~ction...'),},},
 	InsertSlide:{presentation:{menu:_('~New Slide'),},},
-	InsertSoftHyphen:{global:{menu:_('Insert s~oft Hyphen'),},},
+	InsertSlideField:{presentation:{menu:_('~Slide Number'),},},
+	InsertSlideTitleField:{presentation:{menu:_('Slide Tit~le'),},},
+	InsertSlidesField:{presentation:{menu:_('Slide ~Count'),},},
+	InsertSoftHyphen:{global:{menu:_('S~oft hyphen'),},},
 	InsertSymbol:{global:{context:_('Insert Special Character'),menu:_('S~pecial Character...'),},},
+	InsertTable:{presentation:{menu:_('~Table...'),},text:{menu:_('Insert ~Table...'),},},
 	InsertTimeField:{global:{menu:_('Time Field'),},text:{menu:_('~Time'),},},
+	InsertTimeFieldFix:{presentation:{menu:_('~Time (fixed)'),},},
+	InsertTimeFieldVar:{presentation:{menu:_('T~ime (variable)'),},},
 	InsertTitleField:{text:{menu:_('T~itle'),},},
+	InsertTitles:{global:{menu:_('Insert Titles...'),},},
 	InsertTopicField:{text:{menu:_('~Subject'),},},
-	InsertZWNBSP:{global:{menu:_('No-width No ~Break'),},},
-	InsertZWSP:{global:{menu:_('No-~width Optional Break'),},},
+	InsertTrendline:{global:{menu:_('Insert Tre~nd Line...'),},},
+	InsertXErrorBars:{global:{menu:_('Insert X Error ~Bars...'),},},
+	InsertYErrorBars:{global:{menu:_('Insert Y Error ~Bars...'),},},
+	InsertZWNBSP:{global:{menu:_('No-width no ~break'),},},
+	InsertZWSP:{global:{menu:_('No-~width optional break'),},},
 	Italic:{global:{menu:_('Italic'),},},
 	JumpDownThisLevel:{text:{menu:_('To Next Paragraph in Level'),},},
 	JumpUpThisLevel:{text:{menu:_('To Previous Paragraph in Level'),},},
 	JustifyPara:{global:{menu:_('Justified'),},},
 	LanguageMenu:{global:{menu:_('Language'),},},
+	LanguageStatus:{global:{menu:_('Language Status'),},},
 	LayoutStatus:{presentation:{menu:_('Layout'),},},
 	LeftPara:{global:{context:_('Align Left'),menu:_('Left'),},},
 	MasterSlidesPanel:{presentation:{menu:_('Master Slides'),},},
@@ -174,7 +224,7 @@ var unoCommandsArray = {
 	OutlineBullet:{global:{menu:_('~Bullets and Numbering...'),},},
 	OutlineFont:{global:{menu:_('Outline'),},},
 	Overline:{global:{menu:_('Overline'),},},
-	PageDialog:{global:{menu:_('~Page...'),},text:{menu:_('~Page Style...'),},},
+	PageDialog:{global:{menu:_('~Page...'),},text:{menu:_('~Page...'),},},
 	PageDown:{text:{menu:_('Next Page'),},},
 	PageFormatDialog:{spreadsheet:{menu:_('~Page...'),},},
 	PageSetup:{presentation:{menu:_('Properties...'),},},
@@ -192,11 +242,13 @@ var unoCommandsArray = {
 	Redo:{global:{menu:_('~Redo'),},},
 	RejectAllTrackedChanges:{text:{menu:_('Reject All'),},},
 	Remove:{spreadsheet:{menu:_('~Delete Sheet...'),},},
-	RemoveHyperlink:{global:{menu:_('~Remove Hyperlink'),},},
+	RemoveHyperlink:{text:{menu:_('Remove Hyperlink'),},},
 	RemoveTableOf:{text:{menu:_('Delete index'),},},
-	RenameTable:{spreadsheet:{menu:_('Rename S~heet...'),},},
+	RenameTable:{spreadsheet:{menu:_('~Rename Sheet...'),},},
 	ReplyComment:{global:{menu:_('Reply Comment'),},},
+	ResetAllDataPoints:{global:{menu:_('Reset all Data Points'),},},
 	ResetAttributes:{global:{menu:_('~Clear Direct Formatting'),},spreadsheet:{context:_('Clear Direct Formatting'),menu:_('Clear ~Direct Formatting'),},text:{context:_('Clear Direct Formatting'),menu:_('Clear ~Direct Formatting'),},},
+	ResetDataPoint:{global:{menu:_('Reset Data Point'),},},
 	RightPara:{global:{context:_('Align Right'),menu:_('Right'),},},
 	RotateLeft:{text:{menu:_('Rotate 90° ~Left'),},},
 	RotateMenu:{global:{menu:_('Rot~ate'),},},
@@ -209,16 +261,22 @@ var unoCommandsArray = {
 	SelectTable:{presentation:{menu:_('~Select...'),},text:{menu:_('~Table'),},},
 	SendToBack:{global:{menu:_('~Send to Back'),},},
 	SetAnchorAtChar:{text:{menu:_('To ~Character'),},},
+	SetAnchorToCell:{spreadsheet:{menu:_('To ~Cell'),},},
+	SetAnchorToCellResize:{spreadsheet:{menu:_('To Cell (~resize with cell)'),},},
 	SetAnchorToChar:{text:{menu:_('As C~haracter'),},},
 	SetAnchorToFrame:{text:{menu:_('To ~Frame'),},},
 	SetAnchorToPage:{spreadsheet:{menu:_('To P~age'),},text:{menu:_('To P~age'),},},
 	SetAnchorToPara:{text:{menu:_('To ~Paragraph'),},},
+	SetDefault:{global:{menu:_('Clear ~Direct Formatting'),},},
 	SetLanguageAllTextMenu:{global:{menu:_('For All Text'),},},
 	SetLanguageParagraphMenu:{global:{menu:_('For Paragraph'),},},
 	SetLanguageSelectionMenu:{global:{menu:_('For Selection'),},},
+	SetObjectToBackground:{global:{menu:_('To Background'),},},
+	SetObjectToForeground:{global:{menu:_('To Foreground'),},},
 	SetOptimalColumnWidth:{presentation:{menu:_('Optimal Column Width'),},spreadsheet:{menu:_('~Optimal Width...'),},text:{menu:_('Optimal Column Width'),},},
 	SetOptimalRowHeight:{presentation:{menu:_('Optimal Row Height'),},spreadsheet:{menu:_('~Optimal Height...'),},text:{menu:_('Optimal Row Height'),},},
 	Shadowed:{global:{menu:_('Shadow'),},},
+	ShapesMenu:{global:{menu:_('~Shape'),},},
 	SheetMenu:{spreadsheet:{menu:_('~Sheet'),},},
 	Show:{spreadsheet:{menu:_('~Show Sheet...'),},},
 	ShowColumn:{spreadsheet:{context:_('S~how Columns'),menu:_('~Show'),},},
@@ -232,13 +290,15 @@ var unoCommandsArray = {
 	SlideChangeWindow:{presentation:{menu:_('Slide Transition'),},},
 	SlideMasterPage:{presentation:{menu:_('~Master Slide'),},},
 	SlideMenu:{presentation:{menu:_('S~lide'),},},
-	SlideShowMenu:{presentation:{menu:_('~Slide Show'),},},
 	SmallCaps:{global:{menu:_('Small capitals'),},},
 	SortAscending:{spreadsheet:{menu:_('Sort Ascending'),},},
 	SortDescending:{spreadsheet:{menu:_('Sort Descending'),},},
 	SpacePara1:{global:{menu:_('Line Spacing: 1'),},},
 	SpacePara15:{global:{menu:_('Line Spacing: 1.5'),},},
 	SpacePara2:{global:{menu:_('Line Spacing: 2'),},},
+	SpellCheckApplySuggestion:{global:{menu:_('Apply Suggestion'),},},
+	SpellCheckIgnore:{global:{menu:_('Ignore'),},},
+	SpellCheckIgnoreAll:{global:{menu:_('IgnoreAll'),},},
 	SpellDialog:{global:{menu:_('~Spelling...'),},},
 	SpellOnline:{global:{menu:_('~Automatic Spell Checking'),},},
 	SpellingAndGrammarDialog:{global:{menu:_('~Spelling...'),},},
@@ -266,6 +326,7 @@ var unoCommandsArray = {
 	UpSearch:{global:{menu:_('Find Previous'),},},
 	UpdateCurIndex:{text:{context:_('Update index'),menu:_('Current ~Index'),},},
 	Validation:{spreadsheet:{menu:_('~Validity...'),},},
+	View3D:{global:{menu:_('~3D View...'),},},
 	ViewMenu:{global:{menu:_('~View'),},},
 	Watermark:{text:{menu:_('Watermark...'),},},
 	WordCountDialog:{text:{menu:_('~Word Count...'),},},
@@ -279,7 +340,7 @@ var unoCommandsArray = {
 	WrapRight:{text:{menu:_('Wrap Right'),},},
 	WrapText:{spreadsheet:{menu:_('Wrap Text'),},},
 	WrapThrough:{text:{menu:_('~Wrap Through'),},},
-	WrapThroughTransparent:{text:{menu:_('In ~Background'),},},
+	WrapThroughTransparencyToggle:{text:{menu:_('In ~Background'),},},
 	ZoomMinus:{global:{menu:_('Zoom Out'),},},
 	ZoomPlus:{global:{menu:_('Zoom In'),},},
 };
@@ -309,6 +370,10 @@ window._UNO = function(string, component, isContext) {
 		}
 	}
 
+	return this.removeAccessKey(text);
+}
+
+window.removeAccessKey = function(text) {
 	// Remove access key markers from translated strings
 	// 1. access key in parenthesis in case of non-latin scripts
 	text = text.replace(/\(~[A-Za-z]\)/, '');
