@@ -35,7 +35,7 @@
 #endif
 #include "WebSocketHandler.hpp"
 
-int SocketPoll::DefaultPollTimeoutMs = 5000;
+int SocketPoll::DefaultPollTimeoutMicroS = 5000 * 1000;
 std::atomic<bool> SocketPoll::InhibitThreadChecks(false);
 std::atomic<bool> Socket::InhibitThreadChecks(false);
 
@@ -194,6 +194,135 @@ void SocketPoll::pollingThreadEntry()
     LOG_INF("Finished polling thread [" << _name << "].");
 }
 
+int SocketPoll::poll(int64_t timeoutMaxMicroS)
+{
+    if (_runOnClientThread)
+        checkAndReThread();
+    else
+        assertCorrectThread();
+
+    std::chrono::steady_clock::time_point now =
+        std::chrono::steady_clock::now();
+
+    // The events to poll on change each spin of the loop.
+    setupPollFds(now, timeoutMaxMicroS);
+    const size_t size = _pollSockets.size();
+
+    int rc;
+    do
+    {
+#if !MOBILEAPP
+#  if HAVE_PPOLL
+        LOG_TRC("ppoll start, timeoutMicroS: " << timeoutMaxMicroS << " size " << size);
+        timeoutMaxMicroS = std::max(timeoutMaxMicroS, (int64_t)0);
+        struct timespec timeout;
+        timeout.tv_sec = timeoutMaxMicroS / (1000 * 1000);
+        timeout.tv_nsec = (timeoutMaxMicroS % (1000 * 1000)) * 1000;
+        rc = ::ppoll(&_pollFds[0], size + 1, &timeout, nullptr);
+#  else
+        int timeoutMaxMs = (timeoutMaxMicroS + 9999) / 1000;
+        LOG_TRC("Legacy Poll start, timeoutMs: " << timeoutMaxMs);
+        rc = ::poll(&_pollFds[0], size + 1, std::max(timeoutMaxMs,0));
+#  endif
+#else
+        LOG_TRC("SocketPoll Poll");
+        int timeoutMaxMs = (timeoutMaxMicroS + 9999) / 1000;
+        rc = fakeSocketPoll(&_pollFds[0], size + 1, std::max(timeoutMaxMs,0));
+#endif
+    }
+    while (rc < 0 && errno == EINTR);
+    LOG_TRC("Poll completed with " << rc << " live polls max (" <<
+            timeoutMaxMicroS << "us)" << ((rc==0) ? "(timedout)" : ""));
+
+    // First process the wakeup pipe (always the last entry).
+    if (_pollFds[size].revents)
+    {
+        std::vector<CallbackFn> invoke;
+        {
+            std::lock_guard<std::mutex> lock(_mutex);
+
+            // Clear the data.
+#if !MOBILEAPP
+            int dump = ::read(_wakeup[0], &dump, sizeof(dump));
+#else
+            LOG_TRC("Wakeup pipe read");
+            int dump = fakeSocketRead(_wakeup[0], &dump, sizeof(dump));
+#endif
+            // Copy the new sockets over and clear.
+            _pollSockets.insert(_pollSockets.end(),
+                                _newSockets.begin(), _newSockets.end());
+
+            // Update thread ownership.
+            for (auto &i : _newSockets)
+                i->setThreadOwner(std::this_thread::get_id());
+
+            _newSockets.clear();
+
+            // Extract list of callbacks to process
+            std::swap(_newCallbacks, invoke);
+        }
+
+        for (const auto& callback : invoke)
+        {
+            try
+            {
+                callback();
+            }
+            catch (const std::exception& exc)
+            {
+                LOG_ERR("Exception while invoking poll [" << _name <<
+                        "] callback: " << exc.what());
+            }
+        }
+
+        try
+        {
+            wakeupHook();
+        }
+        catch (const std::exception& exc)
+        {
+            LOG_ERR("Exception while invoking poll [" << _name <<
+                    "] wakeup hook: " << exc.what());
+        }
+    }
+
+    // This should only happen when we're stopping.
+    if (_pollSockets.size() != size)
+        return rc;
+
+    // Fire the poll callbacks and remove dead fds.
+    std::chrono::steady_clock::time_point newNow =
+        std::chrono::steady_clock::now();
+
+    for (int i = static_cast<int>(size) - 1; i >= 0; --i)
+    {
+        SocketDisposition disposition(_pollSockets[i]);
+        try
+        {
+            _pollSockets[i]->handlePoll(disposition, newNow,
+                                        _pollFds[i].revents);
+        }
+        catch (const std::exception& exc)
+        {
+            LOG_ERR("Error while handling poll for socket #" <<
+                    _pollFds[i].fd << " in " << _name << ": " << exc.what());
+            disposition.setClosed();
+            rc = -1;
+        }
+
+        if (disposition.isMove() || disposition.isClosed())
+        {
+            LOG_DBG("Removing socket #" << _pollFds[i].fd << " (of " <<
+                    _pollSockets.size() << ") from " << _name);
+            _pollSockets.erase(_pollSockets.begin() + i);
+        }
+
+        disposition.execute();
+    }
+
+    return rc;
+}
+
 void SocketPoll::wakeupWorld()
 {
     for (const auto& fd : getWakeupsArray())
@@ -204,7 +333,7 @@ void SocketPoll::wakeupWorld()
 
 void SocketPoll::insertNewWebSocketSync(
     const Poco::URI &uri,
-    const std::shared_ptr<SocketHandlerInterface>& websocketHandler)
+    const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler)
 {
     LOG_INF("Connecting to " << uri.getHost() << " : " << uri.getPort() << " : " << uri.getPath());
 
@@ -277,8 +406,8 @@ void SocketPoll::insertNewWebSocketSync(
 
 // should this be a static method in the WebsocketHandler(?)
 void SocketPoll::clientRequestWebsocketUpgrade(const std::shared_ptr<StreamSocket>& socket,
-                                               const std::shared_ptr<SocketHandlerInterface>& websocketHandler,
-                                               const std::string &pathAndQuery)
+                                               const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler,
+                                               const std::string &pathAndQuery, const int shareFD)
 {
     // cf. WebSocketHandler::upgradeToWebSocket (?)
     // send Sec-WebSocket-Key: <hmm> ... Sec-WebSocket-Protocol: chat, Sec-WebSocket-Version: 13
@@ -297,14 +426,21 @@ void SocketPoll::clientRequestWebsocketUpgrade(const std::shared_ptr<StreamSocke
         "Sec-WebSocket-Version:13\r\n"
         "User-Agent: " WOPI_AGENT_STRING "\r\n"
         "\r\n";
-    socket->send(oss.str());
+    if (shareFD == -1)
+        socket->send(oss.str());
+    else
+    {
+        std::string request = oss.str();
+        socket->sendFD(request.c_str(), request.size(), shareFD);
+    }
     websocketHandler->onConnect(socket);
 }
 
 void SocketPoll::insertNewUnixSocket(
     const std::string &location,
     const std::string &pathAndQuery,
-    const std::shared_ptr<SocketHandlerInterface>& websocketHandler)
+    const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler,
+    const int shareFD)
 {
     int fd = socket(AF_UNIX, SOCK_STREAM | SOCK_NONBLOCK, 0);
 
@@ -327,7 +463,7 @@ void SocketPoll::insertNewUnixSocket(
         if (socket)
         {
             LOG_DBG("Connected to local UDS " << location << " #" << socket->getFD());
-            clientRequestWebsocketUpgrade(socket, websocketHandler, pathAndQuery);
+            clientRequestWebsocketUpgrade(socket, websocketHandler, pathAndQuery, shareFD);
             insertNewSocket(socket);
         }
     }
@@ -337,7 +473,7 @@ void SocketPoll::insertNewUnixSocket(
 
 void SocketPoll::insertNewFakeSocket(
     int peerSocket,
-    const std::shared_ptr<SocketHandlerInterface>& websocketHandler)
+    const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler)
 {
     LOG_INF("Connecting to " << peerSocket);
     int fd = fakeSocketSocket();
@@ -384,8 +520,8 @@ void SocketDisposition::execute()
     _socketMove = nullptr;
 }
 
-const int WebSocketHandler::InitialPingDelayMs = 25;
-const int WebSocketHandler::PingFrequencyMs = 18 * 1000;
+const int WebSocketHandler::InitialPingDelayMicroS = 25 * 1000;
+const int WebSocketHandler::PingFrequencyMicroS = 18 * 1000 * 1000;
 
 void WebSocketHandler::dumpState(std::ostream& os)
 {
@@ -394,12 +530,14 @@ void WebSocketHandler::dumpState(std::ostream& os)
     if (_wsPayload.size() > 0)
         Util::dumpHex(os, "\t\tws queued payload:\n", "\t\t", _wsPayload);
     os << "\n";
+    if (_msgHandler)
+        _msgHandler->dumpState(os);
 }
 
 void StreamSocket::dumpState(std::ostream& os)
 {
-    int timeoutMaxMs = SocketPoll::DefaultPollTimeoutMs;
-    int events = getPollEvents(std::chrono::steady_clock::now(), timeoutMaxMs);
+    int64_t timeoutMaxMicroS = SocketPoll::DefaultPollTimeoutMicroS;
+    int events = getPollEvents(std::chrono::steady_clock::now(), timeoutMaxMicroS);
     os << "\t" << getFD() << "\t" << events << "\t"
        << _inBuffer.size() << "\t" << _outBuffer.size() << "\t"
        << " r: " << _bytesRecvd << "\t w: " << _bytesSent << "\t"
@@ -556,6 +694,18 @@ int Socket::getPid() const
     return creds.pid;
 }
 
+// Does this socket come from the localhost ?
+bool Socket::isLocal() const
+{
+    if (_clientAddress.size() < 1)
+        return false;
+    if (_clientAddress[0] == '/') // Unix socket
+        return true;
+    if (_clientAddress == "::1")
+        return true;
+    return  _clientAddress.rfind("127.0.0.", 0);
+}
+
 std::shared_ptr<Socket> LocalServerSocket::accept()
 {
     const int rc = ::accept4(getFD(), nullptr, nullptr, SOCK_NONBLOCK);
@@ -778,14 +928,14 @@ bool StreamSocket::parseHeader(const char *clientName,
     }
     catch (const Poco::Exception& exc)
     {
-        LOG_DBG("parseHeader exception caught: " << exc.displayText());
+        LOG_DBG("parseHeader exception caught with " << _inBuffer.size() << " bytes: " << exc.displayText());
         // Probably don't have enough data just yet.
         // TODO: timeout if we never get enough.
         return false;
     }
     catch (const std::exception& exc)
     {
-        LOG_DBG("parseHeader exception caught: " << exc.what());
+        LOG_DBG("parseHeader std::exception caught with " << _inBuffer.size() << " bytes: " << exc.what());
         // Probably don't have enough data just yet.
         // TODO: timeout if we never get enough.
         return false;
