@@ -7,8 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#ifndef INCLUDED_SOCKET_HPP
-#define INCLUDED_SOCKET_HPP
+#pragma once
 
 #include <poll.h>
 #include <unistd.h>
@@ -86,6 +85,10 @@ public:
     {
         _disposition = Type::CLOSED;
     }
+    std::shared_ptr<Socket> getSocket() const
+    {
+        return _socket;
+    }
     bool isMove() { return _disposition == Type::MOVE; }
     bool isClosed() { return _disposition == Type::CLOSED; }
 
@@ -155,7 +158,7 @@ public:
     /// for timeouts, based on current time @now.
     /// @returns POLLIN and POLLOUT if output is expected.
     virtual int getPollEvents(std::chrono::steady_clock::time_point now,
-                              int &timeoutMaxMs) = 0;
+                              int64_t &timeoutMaxMicroS) = 0;
 
     /// Handle results of events returned from poll
     virtual void handlePoll(SocketDisposition &disposition,
@@ -271,6 +274,9 @@ public:
     }
 #endif
 
+    // Does this socket come from the localhost ?
+    bool isLocal() const;
+
     virtual void dumpState(std::ostream&) {}
 
     /// Set the thread-id we're bound to
@@ -344,17 +350,30 @@ private:
 };
 
 class StreamSocket;
+class MessageHandlerInterface;
 
-/// Interface that handles the actual incoming message.
-class SocketHandlerInterface
+/// Interface that decodes the actual incoming message.
+class ProtocolHandlerInterface :
+    public std::enable_shared_from_this<ProtocolHandlerInterface>
 {
+protected:
+    /// We own a message handler, after decoding the socket data we pass it on as messages.
+    std::shared_ptr<MessageHandlerInterface> _msgHandler;
 public:
-    virtual ~SocketHandlerInterface() {}
+    // ------------------------------------------------------------------
+    // Interface for implementing low level socket goodness from streams.
+    // ------------------------------------------------------------------
+    virtual ~ProtocolHandlerInterface() { }
+
     /// Called when the socket is newly created to
     /// set the socket associated with this ResponseClient.
     /// Will be called exactly once.
     virtual void onConnect(const std::shared_ptr<StreamSocket>& socket) = 0;
 
+    /// Enable/disable processing of incoming data at socket level.
+    virtual void enableProcessInput(bool /*enable*/){};
+    virtual bool processInputEnabled(){ return true; };
+
     /// Called after successful socket reads.
     virtual void handleIncomingMessage(SocketDisposition &disposition) = 0;
 
@@ -362,7 +381,7 @@ public:
     /// for timeouts, based on current time @now.
     /// @returns POLLIN and POLLOUT if output is expected.
     virtual int getPollEvents(std::chrono::steady_clock::time_point now,
-                              int &timeoutMaxMs) = 0;
+                              int64_t &timeoutMaxMicroS) = 0;
 
     /// Do we need to handle a timeout ?
     virtual void checkTimeout(std::chrono::steady_clock::time_point /* now */) {}
@@ -370,14 +389,116 @@ public:
     /// Do some of the queued writing.
     virtual void performWrites() = 0;
 
-    /// Called when the is disconnected and will be destroyed.
+    /// Called when the socket is disconnected and will be destroyed.
     /// Will be called exactly once.
     virtual void onDisconnect() {}
 
+    // -----------------------------------------------------------------
+    //            Interface for external MessageHandlers
+    // -----------------------------------------------------------------
+public:
+    void setMessageHandler(const std::shared_ptr<MessageHandlerInterface> &msgHandler)
+    {
+        _msgHandler = msgHandler;
+    }
+
+    /// Clear all external references
+    virtual void dispose() { _msgHandler.reset(); }
+
+    virtual int sendTextMessage(const char* msg, const size_t len, bool flush = false) const = 0;
+    virtual int sendBinaryMessage(const char *data, const size_t len, bool flush = false) const = 0;
+    virtual void shutdown(bool goingAway = false, const std::string &statusMessage = "") = 0;
+
+    virtual void getIOStats(uint64_t &sent, uint64_t &recv) = 0;
+
     /// Append pretty printed internal state to a line
     virtual void dumpState(std::ostream& os) { os << "\n"; }
 };
 
+/// A ProtocolHandlerInterface with dummy sending API.
+class SimpleSocketHandler : public ProtocolHandlerInterface
+{
+public:
+    SimpleSocketHandler() {}
+    int sendTextMessage(const char*, const size_t, bool) const override { return 0; }
+    int sendBinaryMessage(const char *, const size_t , bool ) const override     { return 0; }
+    void shutdown(bool, const std::string &) override {}
+    void getIOStats(uint64_t &, uint64_t &) override {}
+};
+
+/// Interface that receives and sends incoming messages.
+class MessageHandlerInterface :
+    public std::enable_shared_from_this<MessageHandlerInterface>
+{
+protected:
+    std::shared_ptr<ProtocolHandlerInterface> _protocol;
+    MessageHandlerInterface(const std::shared_ptr<ProtocolHandlerInterface> &protocol) :
+        _protocol(protocol)
+    {
+    }
+    virtual ~MessageHandlerInterface() {}
+
+public:
+    /// Setup, after construction for shared_from_this
+    void initialize()
+    {
+        if (_protocol)
+            _protocol->setMessageHandler(shared_from_this());
+    }
+
+    /// Clear all external references
+    virtual void dispose()
+    {
+        if (_protocol)
+        {
+            _protocol->dispose();
+            _protocol.reset();
+        }
+    }
+
+    std::shared_ptr<ProtocolHandlerInterface> getProtocol() const
+    {
+        return _protocol;
+    }
+
+    /// Do we have something to send ?
+    virtual bool hasQueuedMessages() const = 0;
+    /// Please send them to me then.
+    virtual void writeQueuedMessages() = 0;
+    /// We just got a message - here it is
+    virtual void handleMessage(const std::vector<char> &data) = 0;
+    /// Get notified that the underlying transports disconnected
+    virtual void onDisconnect() = 0;
+    /// Append pretty printed internal state to a line
+    virtual void dumpState(std::ostream& os) = 0;
+};
+
+class InputProcessingManager
+{
+public:
+    InputProcessingManager(const std::shared_ptr<ProtocolHandlerInterface> &protocol, bool inputProcess)
+    : _protocol(protocol)
+    {
+        if (_protocol)
+        {
+            // Save previous state to be restored in destructor
+            _prevInputProcess = _protocol->processInputEnabled();
+            protocol->enableProcessInput(inputProcess);
+        }
+    }
+
+    ~InputProcessingManager()
+    {
+        // Restore previous state
+        if (_protocol)
+            _protocol->enableProcessInput(_prevInputProcess);
+    }
+
+private:
+    std::shared_ptr<ProtocolHandlerInterface> _protocol;
+    bool _prevInputProcess;
+};
+
 /// Handles non-blocking socket event polling.
 /// Only polls on N-Sockets and invokes callback and
 /// doesn't manage buffers or client data.
@@ -396,7 +517,7 @@ public:
     ~SocketPoll();
 
     /// Default poll time - useful to increase for debugging.
-    static int DefaultPollTimeoutMs;
+    static int DefaultPollTimeoutMicroS;
     static std::atomic<bool> InhibitThreadChecks;
 
     /// Stop the polling thread.
@@ -453,10 +574,15 @@ public:
     {
         while (continuePolling())
         {
-            poll(DefaultPollTimeoutMs);
+            poll(DefaultPollTimeoutMicroS);
         }
     }
 
+    const std::thread::id &getThreadOwner()
+    {
+        return _owner;
+    }
+
     /// Are we running in either shutdown, or the polling thread.
     /// Asserts in the debug builds, otherwise just logs.
     void assertCorrectThread() const
@@ -478,10 +604,10 @@ public:
     {
         if (InhibitThreadChecks)
             return;
-        std::thread::id us = std::thread::id();
+        std::thread::id us = std::this_thread::get_id();
         if (_owner == us)
             return; // all well
-        LOG_DBG("Ununusual - SocketPoll used from a new thread");
+        LOG_DBG("Unusual - SocketPoll used from a new thread");
         _owner = us;
         for (const auto& it : _pollSockets)
             it->setThreadOwner(us);
@@ -491,122 +617,7 @@ public:
     /// Poll the sockets for available data to read or buffer to write.
     /// Returns the return-value of poll(2): 0 on timeout,
     /// -1 for error, and otherwise the number of events signalled.
-    int poll(int timeoutMaxMs)
-    {
-        if (_runOnClientThread)
-            checkAndReThread();
-        else
-            assertCorrectThread();
-
-        std::chrono::steady_clock::time_point now =
-            std::chrono::steady_clock::now();
-
-        // The events to poll on change each spin of the loop.
-        setupPollFds(now, timeoutMaxMs);
-        const size_t size = _pollSockets.size();
-
-        int rc;
-        do
-        {
-            LOG_TRC("Poll start");
-#if !MOBILEAPP
-            rc = ::poll(&_pollFds[0], size + 1, std::max(timeoutMaxMs,0));
-#else
-            LOG_TRC("SocketPoll Poll");
-            rc = fakeSocketPoll(&_pollFds[0], size + 1, std::max(timeoutMaxMs,0));
-#endif
-        }
-        while (rc < 0 && errno == EINTR);
-        LOG_TRC("Poll completed with " << rc << " live polls max (" <<
-                timeoutMaxMs << "ms)" << ((rc==0) ? "(timedout)" : ""));
-
-        // First process the wakeup pipe (always the last entry).
-        if (_pollFds[size].revents)
-        {
-            std::vector<CallbackFn> invoke;
-            {
-                std::lock_guard<std::mutex> lock(_mutex);
-
-                // Clear the data.
-#if !MOBILEAPP
-                int dump = ::read(_wakeup[0], &dump, sizeof(dump));
-#else
-                LOG_TRC("Wakeup pipe read");
-                int dump = fakeSocketRead(_wakeup[0], &dump, sizeof(dump));
-#endif
-                // Copy the new sockets over and clear.
-                _pollSockets.insert(_pollSockets.end(),
-                                    _newSockets.begin(), _newSockets.end());
-
-                // Update thread ownership.
-                for (auto &i : _newSockets)
-                    i->setThreadOwner(std::this_thread::get_id());
-
-                _newSockets.clear();
-
-                // Extract list of callbacks to process
-                std::swap(_newCallbacks, invoke);
-            }
-
-            for (const auto& callback : invoke)
-            {
-                try
-                {
-                    callback();
-                }
-                catch (const std::exception& exc)
-                {
-                    LOG_ERR("Exception while invoking poll [" << _name <<
-                            "] callback: " << exc.what());
-                }
-            }
-
-            try
-            {
-                wakeupHook();
-            }
-            catch (const std::exception& exc)
-            {
-                LOG_ERR("Exception while invoking poll [" << _name <<
-                        "] wakeup hook: " << exc.what());
-            }
-        }
-
-        // This should only happen when we're stopping.
-        if (_pollSockets.size() != size)
-            return rc;
-
-        // Fire the poll callbacks and remove dead fds.
-        std::chrono::steady_clock::time_point newNow =
-            std::chrono::steady_clock::now();
-
-        for (int i = static_cast<int>(size) - 1; i >= 0; --i)
-        {
-            SocketDisposition disposition(_pollSockets[i]);
-            try
-            {
-                _pollSockets[i]->handlePoll(disposition, newNow,
-                                            _pollFds[i].revents);
-            }
-            catch (const std::exception& exc)
-            {
-                LOG_ERR("Error while handling poll for socket #" <<
-                        _pollFds[i].fd << " in " << _name << ": " << exc.what());
-                disposition.setClosed();
-            }
-
-            if (disposition.isMove() || disposition.isClosed())
-            {
-                LOG_DBG("Removing socket #" << _pollFds[i].fd << " (of " <<
-                        _pollSockets.size() << ") from " << _name);
-                _pollSockets.erase(_pollSockets.begin() + i);
-            }
-
-            disposition.execute();
-        }
-
-        return rc;
-    }
+    int poll(int64_t timeoutMaxMicroS);
 
     /// Write to a wakeup descriptor
     static void wakeup (int fd)
@@ -617,21 +628,6 @@ public:
 #if !MOBILEAPP
             rc = ::write(fd, "w", 1);
 #else
-#if 0
-            // Our fake sockets are record-oriented with a single record buffer, so as we write one
-            // byte at a time to the wakeup pipe, we can't write another one before the previous one
-            // has been read.
-            //
-            // FIXME: Still, explicitly waiting here with fakeSocketPoll() for it to be writable
-            // doesn't seem to be any improvement. On the contrary, with this fakeSocketPoll(), we
-            // hang every time we run the app. Without it, we only hang occasionally.
-
-            LOG_TRC("Wakeup pipe write");
-            struct pollfd p;
-            p.fd = fd;
-            p.events = POLLOUT;
-            fakeSocketPoll(&p, 1, -1);
-#endif
             rc = fakeSocketWrite(fd, "w", 1);
 #endif
         } while (rc == -1 && errno == EINTR);
@@ -672,16 +668,17 @@ public:
     /// Inserts a new remote websocket to be polled.
     /// NOTE: The DNS lookup is synchronous.
     void insertNewWebSocketSync(const Poco::URI &uri,
-                                const std::shared_ptr<SocketHandlerInterface>& websocketHandler);
+                                const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler);
 
     void insertNewUnixSocket(
         const std::string &location,
         const std::string &pathAndQuery,
-        const std::shared_ptr<SocketHandlerInterface>& websocketHandler);
+        const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler,
+        const int shareFD = -1);
 #else
     void insertNewFakeSocket(
         int peerSocket,
-        const std::shared_ptr<SocketHandlerInterface>& websocketHandler);
+        const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler);
 #endif
 
     typedef std::function<void()> CallbackFn;
@@ -696,22 +693,6 @@ public:
 
     virtual void dumpState(std::ostream& os);
 
-    /// Removes a socket from this poller.
-    /// NB. this must be called from the socket poll that
-    /// owns the socket.
-    void releaseSocket(const std::shared_ptr<Socket>& socket)
-    {
-        assert(socket);
-        assertCorrectThread();
-        socket->assertCorrectThread();
-        auto it = std::find(_pollSockets.begin(), _pollSockets.end(), socket);
-        assert(it != _pollSockets.end());
-
-        _pollSockets.erase(it);
-        LOG_DBG("Removing socket #" << socket->getFD() << " (of " <<
-                _pollSockets.size() << ") from " << _name);
-    }
-
     size_t getSocketCount() const
     {
         assertCorrectThread();
@@ -751,13 +732,14 @@ protected:
 
 private:
     /// Generate the request to connect & upgrade this socket to a given path
+    /// and sends a file descriptor along request if is != -1.
     void clientRequestWebsocketUpgrade(const std::shared_ptr<StreamSocket>& socket,
-                                       const std::shared_ptr<SocketHandlerInterface>& websocketHandler,
-                                       const std::string &pathAndQuery);
+                                       const std::shared_ptr<ProtocolHandlerInterface>& websocketHandler,
+                                       const std::string &pathAndQuery, const int shareFD = -1);
 
     /// Initialize the poll fds array with the right events
     void setupPollFds(std::chrono::steady_clock::time_point now,
-                      int &timeoutMaxMs)
+                      int64_t &timeoutMaxMicroS)
     {
         const size_t size = _pollSockets.size();
 
@@ -765,7 +747,7 @@ private:
 
         for (size_t i = 0; i < size; ++i)
         {
-            int events = _pollSockets[i]->getPollEvents(now, timeoutMaxMs);
+            int events = _pollSockets[i]->getPollEvents(now, timeoutMaxMicroS);
             assert(events >= 0); // Or > 0 even?
             _pollFds[i].fd = _pollSockets[i]->getFD();
             _pollFds[i].events = events;
@@ -807,12 +789,20 @@ private:
 };
 
 /// A plain, non-blocking, data streaming socket.
-class StreamSocket : public Socket, public std::enable_shared_from_this<StreamSocket>
+class StreamSocket : public Socket,
+                     public std::enable_shared_from_this<StreamSocket>
 {
 public:
+    enum ReadType
+    {
+        NormalRead,
+        UseRecvmsgExpectFD
+    };
+
     /// Create a StreamSocket from native FD.
     StreamSocket(const int fd, bool /* isClient */,
-                 std::shared_ptr<SocketHandlerInterface> socketHandler) :
+                 std::shared_ptr<ProtocolHandlerInterface> socketHandler,
+                 ReadType readType = NormalRead) :
         Socket(fd),
         _socketHandler(std::move(socketHandler)),
         _bytesSent(0),
@@ -820,7 +810,10 @@ public:
         _wsState(WSState::HTTP),
         _closed(false),
         _sentHTTPContinue(false),
-        _shutdownSignalled(false)
+        _shutdownSignalled(false),
+        _incomingFD(-1),
+        _readType(readType),
+        _inputProcessingEnabled(true)
     {
         LOG_DBG("StreamSocket ctor #" << fd);
 
@@ -865,11 +858,11 @@ public:
     }
 
     int getPollEvents(std::chrono::steady_clock::time_point now,
-                      int &timeoutMaxMs) override
+                      int64_t &timeoutMaxMicroS) override
     {
         // cf. SslSocket::getPollEvents
         assertCorrectThread();
-        int events = _socketHandler->getPollEvents(now, timeoutMaxMs);
+        int events = _socketHandler->getPollEvents(now, timeoutMaxMicroS);
         if (!_outBuffer.empty() || _shutdownSignalled)
             events |= POLLOUT;
         return events;
@@ -897,6 +890,45 @@ public:
     /// Adds Date and User-Agent.
     void send(Poco::Net::HTTPResponse& response);
 
+    /// Sends data with file descriptor as control data.
+    /// Can be used only with Unix sockets.
+    void sendFD(const char* data, const uint64_t len, int fd)
+    {
+        assertCorrectThread();
+
+        // Flush existing non-ancillary data
+        // so that our non-ancillary data will
+        // match ancillary data.
+        if (getOutBuffer().size() > 0)
+        {
+            writeOutgoingData();
+        }
+
+        msghdr msg;
+        iovec iov[1];
+
+        iov[0].iov_base = const_cast<char*>(data);
+        iov[0].iov_len = len;
+
+        msg.msg_name = nullptr;
+        msg.msg_namelen = 0;
+        msg.msg_iov = &iov[0];
+        msg.msg_iovlen = 1;
+
+        char adata[CMSG_SPACE(sizeof(int))];
+        cmsghdr *cmsg = (cmsghdr*)adata;
+        cmsg->cmsg_type = SCM_RIGHTS;
+        cmsg->cmsg_level = SOL_SOCKET;
+        cmsg->cmsg_len = CMSG_LEN(sizeof(int));
+        *(int *)CMSG_DATA(cmsg) = fd;
+
+        msg.msg_control = const_cast<char*>(adata);
+        msg.msg_controllen = CMSG_LEN(sizeof(int));
+        msg.msg_flags = 0;
+
+        sendmsg(getFD(), &msg, 0);
+    }
+
     /// Reads data by invoking readData() and buffering.
     /// Return false iff the socket is closed.
     virtual bool readIncomingData()
@@ -949,7 +981,7 @@ public:
     }
 
     /// Replace the existing SocketHandler with a new one.
-    void setHandler(std::shared_ptr<SocketHandlerInterface> handler)
+    void setHandler(std::shared_ptr<ProtocolHandlerInterface> handler)
     {
         _socketHandler = std::move(handler);
         _socketHandler->onConnect(shared_from_this());
@@ -960,10 +992,12 @@ public:
     /// but we can't have a shared_ptr in the ctor.
     template <typename TSocket>
     static
-    std::shared_ptr<TSocket> create(const int fd, bool isClient, std::shared_ptr<SocketHandlerInterface> handler)
+    std::shared_ptr<TSocket> create(const int fd, bool isClient,
+                                    std::shared_ptr<ProtocolHandlerInterface> handler,
+                                    ReadType readType = NormalRead)
     {
-        SocketHandlerInterface* pHandler = handler.get();
-        auto socket = std::make_shared<TSocket>(fd, isClient, std::move(handler));
+        ProtocolHandlerInterface* pHandler = handler.get();
+        auto socket = std::make_shared<TSocket>(fd, isClient, std::move(handler), readType);
         pHandler->onConnect(socket);
         return socket;
     }
@@ -979,6 +1013,12 @@ public:
         std::vector<std::pair<size_t, size_t>> _spans;
     };
 
+    /// remove all queued input bytes
+    void clearInput()
+    {
+        _inBuffer.clear();
+    }
+
     /// Remove the first @count bytes from input buffer
     void eraseFirstInputBytes(const MessageMap &map)
     {
@@ -1018,6 +1058,14 @@ public:
         return _outBuffer;
     }
 
+    int getIncomingFD()
+    {
+        return _incomingFD;
+    }
+
+    bool processInputEnabled(){ return _inputProcessingEnabled; }
+    void enableProcessInput(bool enable = true){ _inputProcessingEnabled = enable; }
+
 protected:
 
     std::vector<std::pair<size_t, size_t>> findChunks(Poco::Net::HTTPRequest &request);
@@ -1032,7 +1080,7 @@ protected:
 
         _socketHandler->checkTimeout(now);
 
-        if (!events)
+        if (!events && _inBuffer.empty())
             return;
 
         // FIXME: need to close input, but not output (?)
@@ -1041,16 +1089,18 @@ protected:
         // Always try to read.
         closed = !readIncomingData() || closed;
 
+        LOG_TRC("#" << getFD() << ": Incoming data buffer " << _inBuffer.size() <<
+                " bytes, closeSocket? " << closed);
+
+#ifdef LOG_SOCKET_DATA
         auto& log = Log::logger();
-        if (log.trace()) {
-            LOG_TRC("#" << getFD() << ": Incoming data buffer " << _inBuffer.size() <<
-                    " bytes, closeSocket? " << closed);
-            // log.dump("", &_inBuffer[0], _inBuffer.size());
-        }
+        if (log.trace() && _inBuffer.size() > 0)
+            log.dump("", &_inBuffer[0], _inBuffer.size());
+#endif
 
         // If we have data, allow the app to consume.
         size_t oldSize = 0;
-        while (!_inBuffer.empty() && oldSize != _inBuffer.size())
+        while (!_inBuffer.empty() && oldSize != _inBuffer.size() && processInputEnabled())
         {
             oldSize = _inBuffer.size();
             _socketHandler->handleIncomingMessage(disposition);
@@ -1105,16 +1155,18 @@ public:
             ssize_t len;
             do
             {
-                // Writing more than we can absorb in the kernel causes SSL wasteage.
+                // Writing more than we can absorb in the kernel causes SSL wastage.
                 len = writeData(&_outBuffer[0], std::min((int)_outBuffer.size(),
                                                          getSendBufferSize()));
 
+                LOG_TRC("#" << getFD() << ": Wrote outgoing data " << len << " bytes of "
+                            << _outBuffer.size() << " bytes buffered.");
+
+#ifdef LOG_SOCKET_DATA
                 auto& log = Log::logger();
-                if (log.trace() && len > 0) {
-                    LOG_TRC("#" << getFD() << ": Wrote outgoing data " << len <<
-                            " bytes of " << _outBuffer.size() << " bytes buffered.");
-                    // log.dump("", &_outBuffer[0], len);
-                }
+                if (log.trace() && len > 0)
+                    log.dump("", &_outBuffer[0], len);
+#endif
 
                 if (len <= 0 && errno != EAGAIN && errno != EWOULDBLOCK)
                     LOG_SYS("#" << getFD() << ": Socket write returned " << len);
@@ -1138,12 +1190,55 @@ public:
     /// Does it look like we have some TLS / SSL where we don't expect it ?
     bool sniffSSL() const;
 
+    void dumpState(std::ostream& os) override;
+
 protected:
+    /// Reads data with file descriptor as control data if received.
+    /// Can be used only with Unix sockets.
+    int readFD(char* buf, int len, int& fd)
+    {
+        msghdr msg;
+        iovec iov[1];
+        /// We don't expect more than one FD
+        char ctrl[CMSG_SPACE(sizeof(int))];
+        int ctrlLen = sizeof(ctrl);
+
+        iov[0].iov_base = buf;
+        iov[0].iov_len = len;
+
+        msg.msg_name = nullptr;
+        msg.msg_namelen = 0;
+        msg.msg_iov = &iov[0];
+        msg.msg_iovlen = 1;
+        msg.msg_control = ctrl;
+        msg.msg_controllen = ctrlLen;
+        msg.msg_flags = 0;
+
+        int ret = recvmsg(getFD(), &msg, 0);
+        if (ret > 0 && msg.msg_controllen)
+        {
+            cmsghdr *cmsg = CMSG_FIRSTHDR(&msg);
+            if (cmsg && cmsg->cmsg_type == SCM_RIGHTS && cmsg->cmsg_len == CMSG_LEN(sizeof(int)))
+            {
+                fd = *(int*)CMSG_DATA(cmsg);
+                if (_readType == UseRecvmsgExpectFD)
+                {
+                    _readType = NormalRead;
+                }
+            }
+        }
+
+        return ret;
+    }
+
     /// Override to handle reading of socket data differently.
     virtual int readData(char* buf, int len)
     {
         assertCorrectThread();
 #if !MOBILEAPP
+        if (_readType == UseRecvmsgExpectFD)
+            return readFD(buf, len, _incomingFD);
+
         return ::read(getFD(), buf, len);
 #else
         return fakeSocketRead(getFD(), buf, len);
@@ -1161,11 +1256,9 @@ protected:
 #endif
     }
 
-    void dumpState(std::ostream& os) override;
-
-    void setShutdownSignalled(bool shutdownSignalled)
+    void setShutdownSignalled()
     {
-        _shutdownSignalled = shutdownSignalled;
+        _shutdownSignalled = true;
     }
 
     bool isShutdownSignalled() const
@@ -1173,14 +1266,14 @@ protected:
         return _shutdownSignalled;
     }
 
-    const std::shared_ptr<SocketHandlerInterface>& getSocketHandler() const
+    const std::shared_ptr<ProtocolHandlerInterface>& getSocketHandler() const
     {
         return _socketHandler;
     }
 
   private:
     /// Client handling the actual data.
-    std::shared_ptr<SocketHandlerInterface> _socketHandler;
+    std::shared_ptr<ProtocolHandlerInterface> _socketHandler;
 
     std::vector<char> _inBuffer;
     std::vector<char> _outBuffer;
@@ -1198,6 +1291,9 @@ protected:
 
     /// True when shutdown was requested via shutdown().
     bool _shutdownSignalled;
+    int _incomingFD;
+    ReadType _readType;
+    std::atomic_bool _inputProcessingEnabled;
 };
 
 enum class WSOpCode : unsigned char {
@@ -1222,6 +1318,5 @@ namespace HttpHelper
                   Poco::Net::HTTPResponse& response, bool noCache = false, bool deflate = false,
                   const bool headerOnly = false);
 }
-#endif
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
