@@ -7,8 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#ifndef INCLUDED_WEBSOCKETHANDLER_HPP
-#define INCLUDED_WEBSOCKETHANDLER_HPP
+#pragma once
 
 #include <chrono>
 #include <memory>
@@ -24,7 +23,7 @@
 #include <Poco/Net/HTTPResponse.h>
 #include <Poco/Net/WebSocket.h>
 
-class WebSocketHandler : public SocketHandlerInterface
+class WebSocketHandler : public ProtocolHandlerInterface
 {
 private:
     /// The socket that owns us (we can't own it).
@@ -39,7 +38,6 @@ private:
 #if !MOBILEAPP
     bool _isMasking;
     bool _inFragmentBlock;
-    bool _isManualDefrag;
 #endif
 
 protected:
@@ -49,8 +47,8 @@ protected:
         static const unsigned char Mask = 0x80;
     };
 
-    static const int InitialPingDelayMs;
-    static const int PingFrequencyMs;
+    static const int InitialPingDelayMicroS;
+    static const int PingFrequencyMicroS;
 
 public:
     /// Perform upgrade ourselves, or select a client web socket.
@@ -62,7 +60,7 @@ public:
     ///                 defragmentation should be handled inside message handler (true) or the message handler
     ///                 should be called after all fragments of a message were received and the message
     ///                 was defragmented (false).
-    WebSocketHandler(bool isClient = false, bool isMasking = true, bool isManualDefrag = false)
+    WebSocketHandler(bool isClient = false, bool isMasking = true)
         : _lastPingSentTime(std::chrono::steady_clock::now())
         , _pingTimeUs(0)
         , _shuttingDown(false)
@@ -70,7 +68,6 @@ public:
 #if !MOBILEAPP
         , _isMasking(isClient && isMasking)
         , _inFragmentBlock(false)
-        , _isManualDefrag(isManualDefrag)
 #endif
     {
     }
@@ -83,21 +80,20 @@ public:
                      const Poco::Net::HTTPRequest& request)
         : _socket(socket)
         , _lastPingSentTime(std::chrono::steady_clock::now() -
-                            std::chrono::milliseconds(PingFrequencyMs) -
-                            std::chrono::milliseconds(InitialPingDelayMs))
+                            std::chrono::microseconds(PingFrequencyMicroS) -
+                            std::chrono::microseconds(InitialPingDelayMicroS))
         , _pingTimeUs(0)
         , _shuttingDown(false)
         , _isClient(false)
 #if !MOBILEAPP
         , _isMasking(false)
         , _inFragmentBlock(false)
-        , _isManualDefrag(false)
 #endif
     {
         upgradeToWebSocket(request);
     }
 
-    /// Implementation of the SocketHandlerInterface.
+    /// Implementation of the ProtocolHandlerInterface.
     void onConnect(const std::shared_ptr<StreamSocket>& socket) override
     {
         _socket = socket;
@@ -126,7 +122,7 @@ public:
     void sendCloseFrame(const StatusCodes statusCode = StatusCodes::NORMAL_CLOSE, const std::string& statusMessage = "")
     {
         std::shared_ptr<StreamSocket> socket = _socket.lock();
-        if (socket == nullptr)
+        if (!socket)
         {
             LOG_ERR("No socket associated with WebSocketHandler " << this);
             return;
@@ -149,6 +145,24 @@ public:
 #endif
     }
 
+    void shutdown(bool goingAway, const std::string &statusMessage) override
+    {
+        shutdown(goingAway ? WebSocketHandler::StatusCodes::ENDPOINT_GOING_AWAY :
+                 WebSocketHandler::StatusCodes::NORMAL_CLOSE, statusMessage);
+    }
+
+    void getIOStats(uint64_t &sent, uint64_t &recv) override
+    {
+        std::shared_ptr<StreamSocket> socket = getSocket().lock();
+        if (socket)
+            socket->getIOStats(sent, recv);
+        else
+        {
+            sent = 0;
+            recv = 0;
+        }
+    }
+
     void shutdown(const StatusCodes statusCode = StatusCodes::NORMAL_CLOSE, const std::string& statusMessage = "")
     {
         if (!_shuttingDown)
@@ -367,27 +381,18 @@ public:
 
         if (fin)
         {
-            //If is final fragment then process the accumulated message.
-            handleMessage(fin, code, _wsPayload);
+            // If is final fragment then process the accumulated message.
+            handleMessage(_wsPayload);
             _inFragmentBlock = false;
         }
         else
         {
-            if (_isManualDefrag)
-            {
-                //If the user wants to process defragmentation on its own then let him process it.
-                handleMessage(fin, code, _wsPayload);
-                _inFragmentBlock = true;
-            }
-            else
-            {
-                _inFragmentBlock = true;
-                //If is not final fragment then wait for next fragment.
-                return false;
-            }
+            _inFragmentBlock = true;
+            // If is not final fragment then wait for next fragment.
+            return false;
         }
 #else
-        handleMessage(true, WSOpCode::Binary, _wsPayload);
+        handleMessage(_wsPayload);
 
 #endif
 
@@ -396,26 +401,24 @@ public:
         return true;
     }
 
-    /// Implementation of the SocketHandlerInterface.
+    /// Implementation of the ProtocolHandlerInterface.
     virtual void handleIncomingMessage(SocketDisposition&) override
     {
-        // LOG_TRC("***** WebSocketHandler::handleIncomingMessage()");
-
         std::shared_ptr<StreamSocket> socket = _socket.lock();
 
 #if MOBILEAPP
         // No separate "upgrade" is going on
-        if (socket != nullptr && !socket->isWebSocket())
+        if (socket && !socket->isWebSocket())
             socket->setWebSocket();
 #endif
 
-        if (socket == nullptr)
+        if (!socket)
         {
             LOG_ERR("No socket associated with WebSocketHandler " << this);
         }
 #if !MOBILEAPP
         else if (_isClient && !socket->isWebSocket())
-            handleClientUpgrade();
+            handleClientUpgrade(socket);
 #endif
         else
         {
@@ -425,18 +428,22 @@ public:
     }
 
     int getPollEvents(std::chrono::steady_clock::time_point now,
-                      int & timeoutMaxMs) override
+                      int64_t & timeoutMaxMicroS) override
     {
         if (!_isClient)
         {
-            const int timeSincePingMs =
-                std::chrono::duration_cast<std::chrono::milliseconds>(now - _lastPingSentTime).count();
-            timeoutMaxMs = std::min(timeoutMaxMs, PingFrequencyMs - timeSincePingMs);
+            const int64_t timeSincePingMicroS =
+                std::chrono::duration_cast<std::chrono::microseconds>(now - _lastPingSentTime).count();
+            timeoutMaxMicroS = std::min(timeoutMaxMicroS, PingFrequencyMicroS - timeSincePingMicroS);
         }
-        return POLLIN;
+        int events = POLLIN;
+        if (_msgHandler && _msgHandler->hasQueuedMessages())
+            events |= POLLOUT;
+        return events;
     }
 
 #if !MOBILEAPP
+private:
     /// Send a ping message
     void sendPingOrPong(std::chrono::steady_clock::time_point now,
                         const char* data, const size_t len,
@@ -483,9 +490,9 @@ public:
         if (_isClient)
             return;
 
-        const int timeSincePingMs =
-            std::chrono::duration_cast<std::chrono::milliseconds>(now - _lastPingSentTime).count();
-        if (timeSincePingMs >= PingFrequencyMs)
+        const int64_t timeSincePingMicroS =
+            std::chrono::duration_cast<std::chrono::microseconds>(now - _lastPingSentTime).count();
+        if (timeSincePingMicroS >= PingFrequencyMicroS)
         {
             const std::shared_ptr<StreamSocket> socket = _socket.lock();
             if (socket)
@@ -493,14 +500,35 @@ public:
         }
 #endif
     }
+public:
+    void performWrites() override
+    {
+        if (_msgHandler)
+            _msgHandler->writeQueuedMessages();
+    }
 
-    /// By default rely on the socket buffer.
-    void performWrites() override {}
+    void onDisconnect() override
+    {
+        if (_msgHandler)
+            _msgHandler->onDisconnect();
+    }
 
     /// Sends a WebSocket Text message.
     int sendMessage(const std::string& msg) const
     {
-        return sendMessage(msg.data(), msg.size(), WSOpCode::Text);
+        return sendTextMessage(msg.c_str(), msg.size());
+    }
+
+    /// Implementation of the ProtocolHandlerInterface.
+    int sendTextMessage(const char* msg, const size_t len, bool flush = false) const override
+    {
+        return sendMessage(msg, len, WSOpCode::Text, flush);
+    }
+
+    /// Implementation of the ProtocolHandlerInterface.
+    int sendBinaryMessage(const char *data, const size_t len, bool flush = false) const override
+    {
+        return sendMessage(data, len, WSOpCode::Binary, flush);
     }
 
     /// Sends a WebSocket message of WPOpCode type.
@@ -509,7 +537,7 @@ public:
     int sendMessage(const char* data, const size_t len, const WSOpCode code, const bool flush = true) const
     {
         int unitReturn = -1;
-        if (UnitBase::get().filterSendMessage(data, len, code, flush, unitReturn))
+        if (!Util::isFuzzing() && UnitBase::get().filterSendMessage(data, len, code, flush, unitReturn))
             return unitReturn;
 
         //TODO: Support fragmented messages.
@@ -518,27 +546,13 @@ public:
         return sendFrame(socket, data, len, WSFrameMask::Fin | static_cast<unsigned char>(code), flush);
     }
 
-private:
-
-    /// Sends a WebSocket frame given the data, length, and flags.
-    /// Returns the number of bytes written (including frame overhead) on success,
-    /// 0 for closed/invalid socket, and -1 for other errors.
-    int sendFrame(const std::shared_ptr<StreamSocket>& socket,
-                  const char* data, const uint64_t len,
-                  unsigned char flags, const bool flush = true) const
-    {
-        if (!socket || data == nullptr || len == 0)
-            return -1;
-
-        if (socket->isClosed())
-            return 0;
-
-        socket->assertCorrectThread();
-        std::vector<char>& out = socket->getOutBuffer();
+protected:
 
 #if !MOBILEAPP
-        const size_t oldSize = out.size();
-
+    /// Builds a websocket frame based on data and flags received as parameters.
+    /// The frame is output in 'out' parameter
+    void buildFrame(const char* data, const uint64_t len, unsigned char flags, std::vector<char>& out) const
+    {
         out.push_back(flags);
 
         int maskFlag = _isMasking ? 0x80 : 0;
@@ -586,23 +600,50 @@ private:
             // Copy the data.
             out.insert(out.end(), data, data + len);
         }
-        const size_t size = out.size() - oldSize;
-#else
-        LOG_TRC("WebSocketHandle::sendFrame: Writing to #" << socket->getFD() << " " << len << " bytes");
-        assert(flush);
-        assert(out.size() == 0);
-
-        out.insert(out.end(), data, data + len);
-        const size_t size = out.size();
+    }
 #endif
+
+    /// Sends a WebSocket frame given the data, length, and flags.
+    /// Returns the number of bytes written (including frame overhead) on success,
+    /// 0 for closed/invalid socket, and -1 for other errors.
+    int sendFrame(const std::shared_ptr<StreamSocket>& socket,
+                  const char* data, const uint64_t len,
+                  unsigned char flags, const bool flush = true) const
+    {
+        if (!socket || data == nullptr || len == 0)
+            return -1;
+
+        if (socket->isClosed())
+            return 0;
+
+        socket->assertCorrectThread();
+        std::vector<char>& out = socket->getOutBuffer();
+
+#if !MOBILEAPP
+        const size_t oldSize = out.size();
+
+        buildFrame(data, len, flags, out);
+
+        const size_t size = out.size() - oldSize;
+
         if (flush)
             socket->writeOutgoingData();
+#else
+        LOG_TRC("WebSocketHandle::sendFrame: Writing to #" << socket->getFD() << " " << len << " bytes");
+
+        // We ignore the flush parameter and always flush in the MOBILEAPP case because there is no
+        // WebSocket framing, we put the messages as such into the FakeSocket queue.
+
+        (void) flush;
+        out.insert(out.end(), data, data + len);
+        const size_t size = out.size();
+
+        socket->writeOutgoingData();
+#endif
 
         return size;
     }
 
-protected:
-
     bool isControlFrame(WSOpCode code){ return code >= WSOpCode::Close; }
 
     void readPayload(unsigned char *data, size_t dataLen, unsigned char* mask, std::vector<char>& payload)
@@ -626,11 +667,13 @@ protected:
     }
 
     /// To be overriden to handle the websocket messages the way you need.
-    virtual void handleMessage(bool /*fin*/, WSOpCode /*code*/, std::vector<char> &/*data*/)
+    virtual void handleMessage(const std::vector<char> &data)
     {
+        if (_msgHandler)
+            _msgHandler->handleMessage(data);
     }
 
-    std::weak_ptr<StreamSocket>& getSocket()
+    const std::weak_ptr<StreamSocket>& getSocket() const
     {
         return _socket;
     }
@@ -640,6 +683,7 @@ protected:
         _socket = socket;
     }
 
+    /// Implementation of the ProtocolHandlerInterface.
     void dumpState(std::ostream& os) override;
 
 private:
@@ -658,7 +702,7 @@ protected:
     void upgradeToWebSocket(const Poco::Net::HTTPRequest& req)
     {
         std::shared_ptr<StreamSocket> socket = _socket.lock();
-        if (socket == nullptr)
+        if (!socket)
             throw std::runtime_error("Invalid socket while upgrading to WebSocket. Request: " + req.getURI());
 
         LOG_TRC("#" << socket->getFD() << ": Upgrading to WebSocket.");
@@ -694,9 +738,9 @@ protected:
 
 #if !MOBILEAPP
     // Handle incoming upgrade to full socket as client WS.
-    void handleClientUpgrade()
+    void handleClientUpgrade(const std::shared_ptr<StreamSocket>& socket)
     {
-        std::shared_ptr<StreamSocket> socket = _socket.lock();
+        assert(socket && "socket must be valid");
 
         LOG_TRC("Incoming client websocket upgrade response: "
                 << std::string(&socket->getInBuffer()[0], socket->getInBuffer().size()));
@@ -764,14 +808,29 @@ protected:
     void setWebSocket()
     {
         std::shared_ptr<StreamSocket> socket = _socket.lock();
-        socket->setWebSocket();
+        if (socket)
+            socket->setWebSocket();
 
         // No need to ping right upon connection/upgrade,
         // but do reset the time to avoid pinging immediately after.
         _lastPingSentTime = std::chrono::steady_clock::now();
     }
+
+    virtual void enableProcessInput(bool enable = true) override
+    {
+        std::shared_ptr<StreamSocket> socket = _socket.lock();
+        if (socket)
+            socket->enableProcessInput(enable);
+    }
+
+    virtual bool processInputEnabled() override
+    {
+        std::shared_ptr<StreamSocket> socket = _socket.lock();
+        if (socket)
+            return socket->processInputEnabled();
+
+        return false;
+    }
 };
 
-#endif
-
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
