@@ -7,8 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#ifndef INCLUDED_TEST_HELPERS_HPP
-#define INCLUDED_TEST_HELPERS_HPP
+#pragma once
 
 #include <iterator>
 
@@ -27,7 +26,7 @@
 #include <Poco/Path.h>
 #include <Poco/URI.h>
 
-#include <cppunit/extensions/HelperMacros.h>
+#include <test/lokassert.hpp>
 
 #include <Common.hpp>
 #include "common/FileUtil.hpp"
@@ -40,22 +39,71 @@
 
 // Oh dear std::cerr and/or its re-direction is not
 // necessarily thread safe on Linux
+// This is the canonical test log function.
+inline void writeTestLog(const char* const p)
+{
+    fputs(p, stderr);
+    fflush(stderr);
+}
+
+inline void writeTestLog(const std::string& s) { writeTestLog(s.c_str()); }
+
 #ifdef TST_LOG_REDIRECT
-  void tstLog(const std::ostringstream &stream);
+void tstLog(const std::ostringstream& stream);
 #else
-  inline void tstLog(const std::ostringstream &stream)
-  {
-      fprintf(stderr, "%s", stream.str().c_str());
-  }
+inline void tstLog(const std::ostringstream& stream) { writeTestLog(stream.str()); }
 #endif
-#define TST_LOG_NAME_BEGIN(NAME, X) do { \
-                        char t[64]; Poco::DateTime time; snprintf(t, sizeof(t), "%.2u:%.2u:%.2u.%.6u (@%lums) ", \
-                        time.hour(), time.minute(), time.second(), time.millisecond() * 1000 + time.microsecond(), helpers::timeSinceTestStartMs()); \
-                        std::ostringstream str; str << NAME << t << X; tstLog(str); } while (false)
-#define TST_LOG_BEGIN(X) TST_LOG_NAME_BEGIN(testname, X)
-#define TST_LOG_APPEND(X) do { std::ostringstream str; str << X; tstLog(str); } while (false)
-#define TST_LOG_END do { std::ostringstream str; str << "| " << __FILE__ << ':' << __LINE__ << std::endl; tstLog(str); } while (false)
-#define TST_LOG_NAME(NAME, X) TST_LOG_NAME_BEGIN(NAME, X); TST_LOG_END
+
+#define TST_LOG_NAME_BEGIN(OSS, NAME, X, FLUSH)                                                    \
+    do                                                                                             \
+    {                                                                                              \
+        char t[64];                                                                                \
+        Poco::DateTime time;                                                                       \
+        snprintf(t, sizeof(t), "%.2u:%.2u:%.2u.%.6u (@%zums) ", time.hour(), time.minute(),        \
+                 time.second(), time.millisecond() * 1000 + time.microsecond(),                    \
+                 helpers::timeSinceTestStartMs());                                                 \
+        OSS << NAME << t << X;                                                                     \
+        if (FLUSH)                                                                                 \
+            tstLog(OSS);                                                                           \
+    } while (false)
+
+#define TST_LOG_BEGIN(X)                                                                           \
+    do                                                                                             \
+    {                                                                                              \
+        std::ostringstream oss;                                                                    \
+        TST_LOG_NAME_BEGIN(oss, testname, X, true);                                                \
+    } while (false)
+
+#define TST_LOG_APPEND(X)                                                                          \
+    do                                                                                             \
+    {                                                                                              \
+        std::ostringstream str;                                                                    \
+        str << X;                                                                                  \
+        tstLog(str);                                                                               \
+    } while (false)
+
+#define TST_LOG_END_X(OSS)                                                                         \
+    do                                                                                             \
+    {                                                                                              \
+        OSS << "| " __FILE__ ":" << __LINE__ << '\n';                                              \
+        tstLog(OSS);                                                                               \
+    } while (false)
+
+#define TST_LOG_END                                                                                \
+    do                                                                                             \
+    {                                                                                              \
+        std::ostringstream oss_log_end;                                                            \
+        TST_LOG_END_X(oss_log_end);                                                                \
+    } while (false)
+
+#define TST_LOG_NAME(NAME, X)                                                                      \
+    do                                                                                             \
+    {                                                                                              \
+        std::ostringstream oss_log_name;                                                           \
+        TST_LOG_NAME_BEGIN(oss_log_name, NAME, X, false);                                          \
+        TST_LOG_END_X(oss_log_name);                                                               \
+    } while (false)
+
 #define TST_LOG(X) TST_LOG_NAME(testname, X)
 
 // Sometimes we need to retry some commands as they can (due to timing or load) soft-fail.
@@ -181,18 +229,12 @@ inline
 Poco::Net::HTTPClientSession* createSession(const Poco::URI& uri)
 {
 #if ENABLE_SSL
-    if (uri.getScheme() == "https")
+    if (uri.getScheme() == "https" || uri.getScheme() == "wss")
         return new Poco::Net::HTTPSClientSession(uri.getHost(), uri.getPort());
 #endif
     return new Poco::Net::HTTPClientSession(uri.getHost(), uri.getPort());
 }
 
-inline int getClientPort()
-{
-    static const char* clientPort = std::getenv("LOOL_TEST_CLIENT_PORT");
-    return clientPort? atoi(clientPort) : DEFAULT_CLIENT_PORT_NUMBER;
-}
-
 inline std::shared_ptr<Poco::Net::StreamSocket> createRawSocket()
 {
     return
@@ -201,7 +243,7 @@ inline std::shared_ptr<Poco::Net::StreamSocket> createRawSocket()
 #else
         std::make_shared<Poco::Net::StreamSocket>
 #endif
-            (Poco::Net::SocketAddress("127.0.0.1", getClientPort()));
+        (Poco::Net::SocketAddress("127.0.0.1", ClientPortNumber));
 }
 
 inline
@@ -213,7 +255,7 @@ std::string const & getTestServerURI()
 #else
             "http://127.0.0.1:"
 #endif
-            + std::to_string(getClientPort()));
+            + std::to_string(ClientPortNumber));
 
     return serverURI;
 }
@@ -231,8 +273,8 @@ int getErrorCode(LOOLWebSocket& ws, std::string& message, const std::string& tes
     do
     {
         bytes = ws.receiveFrame(buffer.begin(), READ_BUFFER_SIZE, flags);
-        TST_LOG("Got " << LOOLProtocol::getAbbreviatedFrameDump(buffer.begin(), bytes, flags));
-        std::this_thread::sleep_for(std::chrono::milliseconds(POLL_TIMEOUT_MS));
+        TST_LOG("Got " << LOOLWebSocket::getAbbreviatedFrameDump(buffer.begin(), bytes, flags));
+        std::this_thread::sleep_for(std::chrono::microseconds(POLL_TIMEOUT_MICRO_S));
     }
     while (bytes > 0 && (flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE);
 
@@ -258,22 +300,22 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
     try
     {
         int flags = 0;
-        int retries = timeoutMs / 500;
-        const Poco::Timespan waitTime(retries ? timeoutMs * 1000 / retries : timeoutMs * 1000);
         std::vector<char> response;
 
-        bool timedout = false;
+        auto endTime = std::chrono::steady_clock::now() + std::chrono::milliseconds(timeoutMs);
+
         ws.setReceiveTimeout(0);
         do
         {
-            if (ws.poll(waitTime, Poco::Net::Socket::SELECT_READ))
+            auto now = std::chrono::steady_clock::now();
+            if (now > endTime) // timedout
+            {
+                TST_LOG("Timeout after " << timeoutMs << " ms.");
+                break;
+            }
+            long waitTimeUs = std::chrono::duration_cast<std::chrono::microseconds>(endTime - now).count();
+            if (ws.poll(Poco::Timespan(waitTimeUs), Poco::Net::Socket::SELECT_READ))
             {
-                if (timedout)
-                {
-                    TST_LOG_END;
-                    timedout = false;
-                }
-
                 response.resize(READ_BUFFER_SIZE * 8);
                 const int bytes = ws.receiveFrame(response.data(), response.size(), flags);
                 response.resize(std::max(bytes, 0));
@@ -283,7 +325,7 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
                     if (LOOLProtocol::matchPrefix(prefix, message))
                     {
                         TST_LOG("[" << prefix <<  "] Matched " <<
-                                LOOLProtocol::getAbbreviatedFrameDump(response.data(), bytes, flags));
+                                LOOLWebSocket::getAbbreviatedFrameDump(response.data(), bytes, flags));
                         return response;
                     }
                 }
@@ -294,7 +336,8 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
 
                 if (bytes <= 0)
                 {
-                    break;
+                    // Try again, timeout will be handled above.
+                    continue;
                 }
 
                 if ((flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE)
@@ -306,34 +349,14 @@ std::vector<char> getResponseMessage(LOOLWebSocket& ws, const std::string& prefi
                     }
 
                     TST_LOG("[" << prefix <<  "] Ignored " <<
-                            LOOLProtocol::getAbbreviatedFrameDump(response.data(), bytes, flags));
+                            LOOLWebSocket::getAbbreviatedFrameDump(response.data(), bytes, flags));
                 }
             }
-            else
-            {
-                if (!timedout)
-                {
-                    TST_LOG_BEGIN("Timeout (" << (retries > 1 ? "retrying" : "giving up") << ") ");
-                }
-                else
-                {
-                    TST_LOG_APPEND(retries << ' ');
-                }
-
-                --retries;
-                timedout = true;
-            }
-        }
-        while (retries > 0 && (flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE);
-
-        if (timedout)
-        {
-            TST_LOG_END;
         }
+        while ((flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE);
     }
     catch (const Poco::Net::WebSocketException& exc)
     {
-        TST_LOG_END;
         TST_LOG(exc.message());
     }
 
@@ -354,10 +377,10 @@ std::string getResponseString(T& ws, const std::string& prefix, const std::strin
 }
 
 template <typename T>
-std::string assertResponseString(T& ws, const std::string& prefix, const std::string& testname)
+std::string assertResponseString(T& ws, const std::string& prefix, const std::string& testname, const size_t timeoutMs = 10000)
 {
-    const auto res = getResponseString(ws, prefix, testname);
-    CPPUNIT_ASSERT_EQUAL(prefix, res.substr(0, prefix.length()));
+    const auto res = getResponseString(ws, prefix, testname, timeoutMs);
+    LOK_ASSERT_EQUAL(prefix, res.substr(0, prefix.length()));
     return res;
 }
 
@@ -366,7 +389,7 @@ template <typename T>
 std::string assertNotInResponse(T& ws, const std::string& prefix, const std::string& testname)
 {
     const auto res = getResponseString(ws, prefix, testname, 1000);
-    CPPUNIT_ASSERT_MESSAGE(testname + "Did not expect getting message [" + res + "].", res.empty());
+    LOK_ASSERT_MESSAGE(testname + "Did not expect getting message [" + res + "].", res.empty());
     return res;
 }
 
@@ -390,8 +413,12 @@ inline
 bool isDocumentLoaded(LOOLWebSocket& ws, const std::string& testname, bool isView = true)
 {
     const std::string prefix = isView ? "status:" : "statusindicatorfinish:";
-    const auto message = getResponseString(ws, prefix, testname);
-    return LOOLProtocol::matchPrefix(prefix, message);
+    // Allow 20 secs to load
+    const auto message = getResponseString(ws, prefix, testname, 30 * 1000);
+    bool success = LOOLProtocol::matchPrefix(prefix, message);
+    if (!success)
+        TST_LOG("ERR: Timed out loading document");
+    return success;
 }
 
 inline
@@ -435,7 +462,7 @@ connectLOKit(const Poco::URI& uri,
             TST_LOG("Connection problem: " << ex.what());
         }
 
-        std::this_thread::sleep_for(std::chrono::milliseconds(POLL_TIMEOUT_MS));
+        std::this_thread::sleep_for(std::chrono::microseconds(POLL_TIMEOUT_MICRO_S));
     }
     while (retries--);
 
@@ -459,14 +486,14 @@ std::shared_ptr<LOOLWebSocket> loadDocAndGetSocket(const Poco::URI& uri, const s
         {
             return nullptr;
         }
-        CPPUNIT_ASSERT_MESSAGE("cannot load the document " + documentURL, isLoaded);
+        LOK_ASSERT_MESSAGE("cannot load the document " + documentURL, isLoaded);
 
         TST_LOG("Loaded document [" << documentURL << "].");
         return socket;
     }
     catch (const Poco::Exception& exc)
     {
-        CPPUNIT_FAIL(exc.displayText());
+        LOK_ASSERT_FAIL(exc.displayText());
     }
 
     // Really couldn't reach here, but the compiler doesn't know any better.
@@ -484,7 +511,7 @@ std::shared_ptr<LOOLWebSocket> loadDocAndGetSocket(const std::string& docFilenam
     }
     catch (const Poco::Exception& exc)
     {
-        CPPUNIT_FAIL(exc.displayText());
+        LOK_ASSERT_FAIL(exc.displayText());
     }
 
     // Really couldn't reach here, but the compiler doesn't know any better.
@@ -512,7 +539,7 @@ void SocketProcessor(const std::string& testname,
         }
 
         n = socket->receiveFrame(buffer, sizeof(buffer), flags);
-        TST_LOG("Got " << LOOLProtocol::getAbbreviatedFrameDump(buffer, n, flags));
+        TST_LOG("Got " << LOOLWebSocket::getAbbreviatedFrameDump(buffer, n, flags));
         if (n > 0 && (flags & Poco::Net::WebSocket::FRAME_OP_BITMASK) != Poco::Net::WebSocket::FRAME_OP_CLOSE)
         {
             if (!handler(std::string(buffer, n)))
@@ -528,7 +555,7 @@ inline
 void parseDocSize(const std::string& message, const std::string& type,
                   int& part, int& parts, int& width, int& height, int& viewid)
 {
-    std::vector<std::string> tokens(LOOLProtocol::tokenize(message, ' '));
+    StringVector tokens(LOOLProtocol::tokenize(message, ' '));
 
     // Expected format is something like 'type= parts= current= width= height='.
     const std::string text = tokens[0].substr(std::string("type=").size());
@@ -537,7 +564,7 @@ void parseDocSize(const std::string& message, const std::string& type,
     width = std::stoi(tokens[3].substr(std::string("width=").size()));
     height = std::stoi(tokens[4].substr(std::string("height=").size()));
     viewid = std::stoi(tokens[5].substr(std::string("viewid=").size()));
-    CPPUNIT_ASSERT_EQUAL(type, text);
+    LOK_ASSERT_EQUAL(type, text);
     CPPUNIT_ASSERT(parts > 0);
     CPPUNIT_ASSERT(part >= 0);
     CPPUNIT_ASSERT(width > 0);
@@ -557,15 +584,15 @@ std::vector<char> assertTileMessage(LOOLWebSocket& ws, const std::string& testna
     const std::vector<char> response = getTileMessage(ws, testname);
 
     const std::string firstLine = LOOLProtocol::getFirstLine(response);
-    std::vector<std::string> tileTokens(LOOLProtocol::tokenize(firstLine, ' '));
-    CPPUNIT_ASSERT_EQUAL(std::string("tile:"), tileTokens[0]);
-    CPPUNIT_ASSERT_EQUAL(std::string("part="), tileTokens[1].substr(0, std::string("part=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("width="), tileTokens[2].substr(0, std::string("width=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("height="), tileTokens[3].substr(0, std::string("height=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("tileposx="), tileTokens[4].substr(0, std::string("tileposx=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("tileposy="), tileTokens[5].substr(0, std::string("tileposy=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("tilewidth="), tileTokens[6].substr(0, std::string("tilewidth=").size()));
-    CPPUNIT_ASSERT_EQUAL(std::string("tileheight="), tileTokens[7].substr(0, std::string("tileheight=").size()));
+    StringVector tileTokens(LOOLProtocol::tokenize(firstLine, ' '));
+    LOK_ASSERT_EQUAL(std::string("tile:"), tileTokens[0]);
+    LOK_ASSERT_EQUAL(std::string("part="), tileTokens[1].substr(0, std::string("part=").size()));
+    LOK_ASSERT_EQUAL(std::string("width="), tileTokens[2].substr(0, std::string("width=").size()));
+    LOK_ASSERT_EQUAL(std::string("height="), tileTokens[3].substr(0, std::string("height=").size()));
+    LOK_ASSERT_EQUAL(std::string("tileposx="), tileTokens[4].substr(0, std::string("tileposx=").size()));
+    LOK_ASSERT_EQUAL(std::string("tileposy="), tileTokens[5].substr(0, std::string("tileposy=").size()));
+    LOK_ASSERT_EQUAL(std::string("tilewidth="), tileTokens[6].substr(0, std::string("tilewidth=").size()));
+    LOK_ASSERT_EQUAL(std::string("tileheight="), tileTokens[7].substr(0, std::string("tileheight=").size()));
 
     return response;
 }
@@ -774,6 +801,5 @@ inline std::string getAllText(const std::shared_ptr<LOOLWebSocket>& socket,
 }
 
 }
-#endif
 
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
