@@ -7,8 +7,7 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/.
  */
 
-#ifndef INCLUDED_TEST_HPP
-#define INCLUDED_TEST_HPP
+#pragma once
 
 #include <vector>
 
@@ -31,9 +30,10 @@ std::vector<int> getDocKitPids();
 /// Get the PID of the forkit
 std::vector<int> getForKitPids();
 
+/// Which port should we connect to get to WSD.
+int getClientPort();
+
 /// How many live loolkit processes do we have ?
 int getLoolKitProcessCount();
 
-#endif
-
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
