@@ -23,8 +23,6 @@
 #include "Auth.hpp"
 #include <Common.hpp>
 #include "FileServer.hpp"
-#include <IoUtil.hpp>
-#include "LOOLWSD.hpp"
 #include <Log.hpp>
 #include <Protocol.hpp>
 #include "Storage.hpp"
@@ -45,15 +43,15 @@ using Poco::Net::HTTPResponse;
 using Poco::Util::Application;
 
 const int Admin::MinStatsIntervalMs = 50;
-const int Admin::DefStatsIntervalMs = 2500;
+const int Admin::DefStatsIntervalMs = 1000;
+
 
 /// Process incoming websocket messages
-void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
-                                       std::vector<char> &payload)
+void AdminSocketHandler::handleMessage(const std::vector<char> &payload)
 {
     // FIXME: check fin, code etc.
     const std::string firstLine = getFirstLine(payload.data(), payload.size());
-    std::vector<std::string> tokens(LOOLProtocol::tokenize(firstLine, ' '));
+    StringVector tokens(LOOLProtocol::tokenize(firstLine, ' '));
     LOG_TRC("Recv: " << firstLine << " tokens " << tokens.size());
 
     if (tokens.empty())
@@ -64,7 +62,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
 
     AdminModel& model = _admin->getModel();
 
-    if (tokens[0] == "auth")
+    if (tokens.equals(0, "auth"))
     {
         if (tokens.size() < 2)
         {
@@ -101,72 +99,85 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
         shutdown();
         return;
     }
-    else if (tokens[0] == "documents" ||
-             tokens[0] == "active_users_count" ||
-             tokens[0] == "active_docs_count" ||
-             tokens[0] == "mem_stats" ||
-             tokens[0] == "cpu_stats" ||
-             tokens[0] == "sent_activity" ||
-             tokens[0] == "recv_activity")
+    else if (tokens.equals(0, "documents") ||
+             tokens.equals(0, "active_users_count") ||
+             tokens.equals(0, "active_docs_count") ||
+             tokens.equals(0, "mem_stats") ||
+             tokens.equals(0, "cpu_stats") ||
+             tokens.equals(0, "sent_activity") ||
+             tokens.equals(0, "recv_activity"))
     {
         const std::string result = model.query(tokens[0]);
         if (!result.empty())
             sendTextFrame(tokens[0] + ' ' + result);
     }
-    else if (tokens[0] == "history")
+    else if (tokens.equals(0, "history"))
     {
         sendTextFrame("{ \"History\": " + model.getAllHistory() + "}");
     }
-    else if (tokens[0] == "version")
+    else if (tokens.equals(0, "version"))
     {
         // Send LOOL version information
         sendTextFrame("loolserver " + LOOLWSD::getVersionJSON());
         // Send LOKit version information
         sendTextFrame("lokitversion " + LOOLWSD::LOKitVersion);
     }
-    else if (tokens[0] == "subscribe" && tokens.size() > 1)
+    else if (tokens.equals(0, "subscribe") && tokens.size() > 1)
     {
         for (std::size_t i = 0; i < tokens.size() - 1; i++)
         {
             model.subscribe(_sessionId, tokens[i + 1]);
         }
     }
-    else if (tokens[0] == "unsubscribe" && tokens.size() > 1)
+    else if (tokens.equals(0, "unsubscribe") && tokens.size() > 1)
     {
         for (std::size_t i = 0; i < tokens.size() - 1; i++)
         {
             model.unsubscribe(_sessionId, tokens[i + 1]);
         }
     }
-    else if (tokens[0] == "mem_consumed")
+    else if (tokens.equals(0, "mem_consumed"))
         sendTextFrame("mem_consumed " + std::to_string(_admin->getTotalMemoryUsage()));
 
-    else if (tokens[0] == "total_avail_mem")
+    else if (tokens.equals(0, "total_avail_mem"))
         sendTextFrame("total_avail_mem " + std::to_string(_admin->getTotalAvailableMemory()));
 
-    else if (tokens[0] == "sent_bytes")
+    else if (tokens.equals(0, "sent_bytes"))
         sendTextFrame("sent_bytes " + std::to_string(model.getSentBytesTotal() / 1024));
 
-    else if (tokens[0] == "recv_bytes")
+    else if (tokens.equals(0, "recv_bytes"))
         sendTextFrame("recv_bytes " + std::to_string(model.getRecvBytesTotal() / 1024));
 
-    else if (tokens[0] == "uptime")
+    else if (tokens.equals(0, "uptime"))
         sendTextFrame("uptime " + std::to_string(model.getServerUptime()));
 
-    else if (tokens[0] == "kill" && tokens.size() == 2)
+    else if (tokens.equals(0, "kill") && tokens.size() == 2)
     {
         try
         {
             const int pid = std::stoi(tokens[1]);
             LOG_INF("Admin request to kill PID: " << pid);
-            SigUtil::killChild(pid);
+
+            std::set<pid_t> pids = model.getDocumentPids();
+            if (pids.find(pid) != pids.end())
+            {
+                SigUtil::killChild(pid);
+            }
+            else
+            {
+                LOG_WRN("Invalid PID to kill (not a document pid)");
+            }
         }
         catch (std::invalid_argument& exc)
         {
-            LOG_WRN("Invalid PID to kill: " << tokens[1]);
+            LOG_WRN("Invalid PID to kill (invalid argument): " << tokens[1]);
+        }
+        catch (std::out_of_range& exc)
+        {
+            LOG_WRN("Invalid PID to kill (out of range): " << tokens[1]);
         }
     }
-    else if (tokens[0] == "settings")
+    else if (tokens.equals(0, "settings"))
     {
         // for now, we have only these settings
         std::ostringstream oss;
@@ -186,17 +197,17 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
 
         sendTextFrame(oss.str());
     }
-    else if (tokens[0] == "shutdown")
+    else if (tokens.equals(0, "shutdown"))
     {
         LOG_INF("Shutdown requested by admin.");
         SigUtil::requestShutdown();
         return;
     }
-    else if (tokens[0] == "set" && tokens.size() > 1)
+    else if (tokens.equals(0, "set") && tokens.size() > 1)
     {
         for (size_t i = 1; i < tokens.size(); i++)
         {
-            std::vector<std::string> setting(LOOLProtocol::tokenize(tokens[i], '='));
+            StringVector setting(LOOLProtocol::tokenize(tokens[i], '='));
             int settingVal = 0;
             try
             {
@@ -212,7 +223,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
             const std::string settingName = setting[0];
             if (settingName == "mem_stats_size")
             {
-                if (settingVal != std::stoi(model.query(settingName)))
+                if (settingVal != std::stol(model.query(settingName)))
                 {
                     model.setMemStatsSize(settingVal);
                 }
@@ -228,7 +239,7 @@ void AdminSocketHandler::handleMessage(bool /* fin */, WSOpCode /* code */,
             }
             else if (settingName == "cpu_stats_size")
             {
-                if (settingVal != std::stoi(model.query(settingName)))
+                if (settingVal != std::stol(model.query(settingName)))
                 {
                     model.setCpuStatsSize(settingVal);
                 }
@@ -284,7 +295,11 @@ AdminSocketHandler::AdminSocketHandler(Admin* adminManager)
 
 void AdminSocketHandler::sendTextFrame(const std::string& message)
 {
-    UnitWSD::get().onAdminQueryMessage(message);
+    if (!Util::isFuzzing())
+    {
+        UnitWSD::get().onAdminQueryMessage(message);
+    }
+
     if (_isAuthenticated)
     {
         LOG_TRC("send admin text frame '" << message << "'");
@@ -316,9 +331,14 @@ bool AdminSocketHandler::handleInitialRequest(
     }
 
     std::shared_ptr<StreamSocket> socket = socketWeak.lock();
+    if (!socket)
+    {
+        LOG_ERR("Invalid socket while reading initial request.");
+        return false;
+    }
 
     const std::string& requestURI = request.getURI();
-    std::vector<std::string> pathTokens(LOOLProtocol::tokenize(requestURI, '/'));
+    StringVector pathTokens(LOOLProtocol::tokenize(requestURI, '/'));
 
     if (request.find("Upgrade") != request.end() && Poco::icompare(request["Upgrade"], "websocket") == 0)
     {
@@ -343,16 +363,15 @@ bool AdminSocketHandler::handleInitialRequest(
 /// An admin command processor.
 Admin::Admin() :
     SocketPoll("admin"),
-    _model(AdminModel()),
     _forKitPid(-1),
-    _forKitWritePipe(-1),
     _lastTotalMemory(0),
     _lastJiffies(0),
     _lastSentCount(0),
     _lastRecvCount(0),
     _cpuStatsTaskIntervalMs(DefStatsIntervalMs),
     _memStatsTaskIntervalMs(DefStatsIntervalMs * 2),
-    _netStatsTaskIntervalMs(DefStatsIntervalMs * 2)
+    _netStatsTaskIntervalMs(DefStatsIntervalMs * 2),
+    _cleanupIntervalMs(DefStatsIntervalMs * 10)
 {
     LOG_INF("Admin ctor.");
 
@@ -378,13 +397,14 @@ Admin::~Admin()
 
 void Admin::pollingThread()
 {
-    std::chrono::steady_clock::time_point lastCPU, lastMem, lastNet;
+    std::chrono::steady_clock::time_point lastCPU, lastMem, lastNet, lastCleanup;
 
     _model.setThreadOwner(std::this_thread::get_id());
 
     lastCPU = std::chrono::steady_clock::now();
     lastMem = lastCPU;
     lastNet = lastCPU;
+    lastCleanup = lastCPU;
 
     while (!isStop() && !SigUtil::getTerminationFlag() && !SigUtil::getShutdownRequestFlag())
     {
@@ -406,6 +426,8 @@ void Admin::pollingThread()
             std::chrono::duration_cast<std::chrono::milliseconds>(now - lastMem).count();
         if (memWait <= MinStatsIntervalMs / 2) // Close enough
         {
+            _model.UpdateMemoryDirty();
+
             const size_t totalMem = getTotalMemoryUsage();
             _model.addMemStats(totalMem);
 
@@ -417,6 +439,8 @@ void Admin::pollingThread()
                 _lastTotalMemory = totalMem;
             }
 
+            notifyDocsMemDirtyChanged();
+
             memWait += _memStatsTaskIntervalMs;
             lastMem = now;
         }
@@ -442,6 +466,19 @@ void Admin::pollingThread()
             lastNet = now;
         }
 
+        int cleanupWait = _cleanupIntervalMs -
+                std::chrono::duration_cast<std::chrono::milliseconds>(now - lastCleanup).count();
+        if (cleanupWait <= MinStatsIntervalMs / 2) // Close enough
+        {
+            if (_defDocProcSettings.getCleanupSettings().getEnable())
+            {
+                cleanupResourceConsumingDocs();
+                
+                cleanupWait += _cleanupIntervalMs;
+                lastCleanup = now;
+            }
+        }
+
         // (re)-connect (with sync. DNS - urk) to one monitor at a time
         if (_pendingConnects.size())
         {
@@ -454,20 +491,21 @@ void Admin::pollingThread()
         }
 
         // Handle websockets & other work.
-        const int timeout = capAndRoundInterval(std::min(std::min(cpuWait, memWait), netWait));
+        const int timeout = capAndRoundInterval(std::min(std::min(std::min(cpuWait, memWait), netWait), cleanupWait));
         LOG_TRC("Admin poll for " << timeout << "ms.");
-        poll(timeout);
+        poll(timeout * 1000); // continue with ms for admin, settings etc.
     }
 }
 
-void Admin::modificationAlert(const std::string& dockey, Poco::Process::PID pid, bool value){
+void Admin::modificationAlert(const std::string& dockey, pid_t pid, bool value){
     addCallback([=] { _model.modificationAlert(dockey, pid, value); });
 }
 
-void Admin::addDoc(const std::string& docKey, Poco::Process::PID pid, const std::string& filename,
-        const std::string& sessionId, const std::string& userName, const std::string& userId)
+void Admin::addDoc(const std::string& docKey, pid_t pid, const std::string& filename,
+                   const std::string& sessionId, const std::string& userName, const std::string& userId,
+                   const int smapsFD)
 {
-    addCallback([=] { _model.addDocument(docKey, pid, filename, sessionId, userName, userId); });
+    addCallback([=] { _model.addDocument(docKey, pid, filename, sessionId, userName, userId, smapsFD); });
 }
 
 void Admin::rmDoc(const std::string& docKey, const std::string& sessionId)
@@ -504,7 +542,7 @@ size_t Admin::getTotalMemoryUsage()
     // memory to the forkit; and then count only dirty pages in the clients
     // since we know that they share everything else with the forkit.
     const size_t forkitRssKb = Util::getMemoryUsageRSS(_forKitPid);
-    const size_t wsdPssKb = Util::getMemoryUsagePSS(Poco::Process::id());
+    const size_t wsdPssKb = Util::getMemoryUsagePSS(getpid());
     const size_t kitsDirtyKb = _model.getKitsMemoryUsage();
     const size_t totalMem = wsdPssKb + forkitRssKb + kitsDirtyKb;
 
@@ -514,7 +552,7 @@ size_t Admin::getTotalMemoryUsage()
 size_t Admin::getTotalCpuUsage()
 {
     const size_t forkitJ = Util::getCpuUsage(_forKitPid);
-    const size_t wsdJ = Util::getCpuUsage(Poco::Process::id());
+    const size_t wsdJ = Util::getCpuUsage(getpid());
     const size_t kitsJ = _model.getKitsJiffies();
 
     if (_lastJiffies == 0)
@@ -554,16 +592,32 @@ void Admin::updateLastActivityTime(const std::string& docKey)
     addCallback([=]{ _model.updateLastActivityTime(docKey); });
 }
 
-void Admin::updateMemoryDirty(const std::string& docKey, int dirty)
-{
-    addCallback([=] { _model.updateMemoryDirty(docKey, dirty); });
-}
 
 void Admin::addBytes(const std::string& docKey, uint64_t sent, uint64_t recv)
 {
     addCallback([=] { _model.addBytes(docKey, sent, recv); });
 }
 
+void Admin::setViewLoadDuration(const std::string& docKey, const std::string& sessionId, std::chrono::milliseconds viewLoadDuration)
+{
+    addCallback([=]{ _model.setViewLoadDuration(docKey, sessionId, viewLoadDuration); });
+}
+
+void Admin::setDocWopiDownloadDuration(const std::string& docKey, std::chrono::milliseconds wopiDownloadDuration)
+{
+    addCallback([=]{ _model.setDocWopiDownloadDuration(docKey, wopiDownloadDuration); });
+}
+
+void Admin::setDocWopiUploadDuration(const std::string& docKey, const std::chrono::milliseconds uploadDuration)
+{
+    addCallback([=]{ _model.setDocWopiUploadDuration(docKey, uploadDuration); });
+}
+
+void Admin::addSegFaultCount(unsigned segFaultCount)
+{
+    addCallback([=]{ _model.addSegFaultCount(segFaultCount); });
+}
+
 void Admin::notifyForkit()
 {
     std::ostringstream oss;
@@ -572,10 +626,7 @@ void Admin::notifyForkit()
         << "setconfig limit_file_size_mb " << _defDocProcSettings.getLimitFileSizeMb() << '\n'
         << "setconfig limit_num_open_files " << _defDocProcSettings.getLimitNumberOpenFiles() << '\n';
 
-    if (_forKitWritePipe != -1)
-        IoUtil::writeToPipe(_forKitWritePipe, oss.str());
-    else
-        LOG_INF("Forkit write pipe not set (yet).");
+    LOOLWSD::sendMessageToForKit(oss.str());
 }
 
 void Admin::triggerMemoryCleanup(const size_t totalMem)
@@ -626,6 +677,16 @@ void Admin::triggerMemoryCleanup(const size_t totalMem)
     }
 }
 
+void Admin::notifyDocsMemDirtyChanged()
+{
+    _model.notifyDocsMemDirtyChanged();
+}
+
+void Admin::cleanupResourceConsumingDocs()
+{
+    _model.cleanupResourceConsumingDocs();
+}
+
 void Admin::dumpState(std::ostream& os)
 {
     // FIXME: be more helpful ...
@@ -645,7 +706,7 @@ public:
     {
     }
     int getPollEvents(std::chrono::steady_clock::time_point now,
-                      int &timeoutMaxMs) override
+                      int64_t &timeoutMaxMicroS) override
     {
         if (_connecting)
         {
@@ -653,7 +714,7 @@ public:
             return POLLOUT;
         }
         else
-            return AdminSocketHandler::getPollEvents(now, timeoutMaxMs);
+            return AdminSocketHandler::getPollEvents(now, timeoutMaxMicroS);
     }
 
     void performWrites() override
@@ -688,6 +749,34 @@ void Admin::scheduleMonitorConnect(const std::string &uri, std::chrono::steady_c
     _pendingConnects.push_back(todo);
 }
 
+void Admin::getMetrics(std::ostringstream &metrics)
+{
+    size_t memAvail =  getTotalAvailableMemory();
+    size_t memUsed = getTotalMemoryUsage();
+
+    metrics << "global_host_system_memory_bytes " << _totalSysMemKb * 1024 << std::endl;
+    metrics << "global_memory_available_bytes " << memAvail * 1024 << std::endl;
+    metrics << "global_memory_used_bytes " << memUsed * 1024 << std::endl;
+    metrics << "global_memory_free_bytes " << (memAvail - memUsed) * 1024 << std::endl;
+    metrics << std::endl;
+
+    _model.getMetrics(metrics);
+}
+
+void Admin::sendMetrics(const std::shared_ptr<StreamSocket>& socket, const std::shared_ptr<Poco::Net::HTTPResponse>& response)
+{
+    std::ostringstream oss;
+    response->write(oss);
+    getMetrics(oss);
+    socket->send(oss.str());
+    socket->shutdown();
+}
+
+void Admin::sendMetricsAsync(const std::shared_ptr<StreamSocket>& socket, const std::shared_ptr<Poco::Net::HTTPResponse>& response)
+{
+    addCallback([this, socket, response]{ sendMetrics(socket, response); });
+}
+
 void Admin::start()
 {
     bool haveMonitors = false;
