@@ -35,15 +35,22 @@ using namespace LOOLProtocol;
 
 using Poco::Path;
 
-static std::mutex SessionMapMutex;
-static std::unordered_map<std::string, std::weak_ptr<ClientSession>> SessionMap;
+// rotates regularly
+const int ClipboardTokenLengthBytes = 16;
+// home-use, disabled by default.
+const int ProxyAccessTokenLengthBytes = 32;
 
-ClientSession::ClientSession(const std::string& id,
-                             const std::shared_ptr<DocumentBroker>& docBroker,
-                             const Poco::URI& uriPublic,
-                             const bool readOnly,
-                             const std::string& hostNoTrust) :
-    Session("ToClient-" + id, id, readOnly),
+static std::mutex GlobalSessionMapMutex;
+static std::unordered_map<std::string, std::weak_ptr<ClientSession>> GlobalSessionMap;
+
+ClientSession::ClientSession(
+    const std::shared_ptr<ProtocolHandlerInterface>& ws,
+    const std::string& id,
+    const std::shared_ptr<DocumentBroker>& docBroker,
+    const Poco::URI& uriPublic,
+    const bool readOnly,
+    const RequestDetails &requestDetails) :
+    Session(ws, "ToClient-" + id, id, readOnly),
     _docBroker(docBroker),
     _uriPublic(uriPublic),
     _isDocumentOwner(false),
@@ -56,11 +63,23 @@ ClientSession::ClientSession(const std::string& id,
     _tileWidthTwips(0),
     _tileHeightTwips(0),
     _kitViewId(-1),
-    _hostNoTrust(hostNoTrust),
+    _serverURL(requestDetails),
     _isTextDocument(false)
 {
     const size_t curConnections = ++LOOLWSD::NumConnections;
-    LOG_INF("ClientSession ctor [" << getName() << "], current number of connections: " << curConnections);
+    LOG_INF("ClientSession ctor [" << getName() << "] for URI: [" << _uriPublic.toString()
+                                   << "], current number of connections: " << curConnections);
+
+    for (const auto& param : _uriPublic.getQueryParameters())
+    {
+        if (param.first == "reuse_cookies")
+        {
+            // Cache the cookies to avoid re-parsing the URI again.
+            _cookies = param.second;
+            LOG_INF("ClientSession [" << getName() << "] has cookies: " << _cookies);
+            break;
+        }
+    }
 
     // populate with random values.
     for (auto it : _clipboardKeys)
@@ -73,8 +92,9 @@ ClientSession::ClientSession(const std::string& id,
 // Can't take a reference in the constructor.
 void ClientSession::construct()
 {
-    std::unique_lock<std::mutex> lock(SessionMapMutex);
-    SessionMap[getId()] = shared_from_this();
+    std::unique_lock<std::mutex> lock(GlobalSessionMapMutex);
+    MessageHandlerInterface::initialize();
+    GlobalSessionMap[getId()] = client_from_this();
 }
 
 ClientSession::~ClientSession()
@@ -82,8 +102,8 @@ ClientSession::~ClientSession()
     const size_t curConnections = --LOOLWSD::NumConnections;
     LOG_INF("~ClientSession dtor [" << getName() << "], current number of connections: " << curConnections);
 
-    std::unique_lock<std::mutex> lock(SessionMapMutex);
-    SessionMap.erase(getId());
+    std::unique_lock<std::mutex> lock(GlobalSessionMapMutex);
+    GlobalSessionMap.erase(getId());
 }
 
 static const char *stateToString(ClientSession::SessionState s)
@@ -142,10 +162,11 @@ bool ClientSession::disconnectFromKit()
     {
         setState(SessionState::WAIT_DISCONNECT);
 
+#ifndef IOS
         LOG_TRC("request/rescue clipboard on disconnect for " << getId());
         // rescue clipboard before shutdown.
         docBroker->forwardToChild(getId(), "getclipboard");
-
+#endif
         // handshake nicely; so wait for 'disconnected'
         docBroker->forwardToChild(getId(), "disconnect");
 
@@ -172,7 +193,8 @@ void ClientSession::rotateClipboardKey(bool notifyClient)
         return;
 
     _clipboardKeys[1] = _clipboardKeys[0];
-    _clipboardKeys[0] = Util::rng::getHardRandomHexString(16);
+    _clipboardKeys[0] = Util::rng::getHardRandomHexString(
+        ClipboardTokenLengthBytes);
     LOG_TRC("Clipboard key on [" << getId() << "] set to " << _clipboardKeys[0] <<
             " last was " << _clipboardKeys[1]);
     if (notifyClient)
@@ -191,12 +213,11 @@ std::string ClientSession::getClipboardURI(bool encode)
     std::string encodeChars = ",/?:@&=+$#"; // match JS encodeURIComponent
     Poco::URI::encode(wopiSrc.toString(), encodeChars, encodedFrom);
 
-    std::string proto = (LOOLWSD::isSSLEnabled() || LOOLWSD::isSSLTermination()) ? "https://" : "http://";
-    std::string meta = proto + _hostNoTrust +
+    std::string meta = _serverURL.getSubURLForEndpoint(
         "/lool/clipboard?WOPISrc=" + encodedFrom +
         "&ServerId=" + LOOLWSD::HostIdentifier +
         "&ViewId=" + std::to_string(getKitViewId()) +
-        "&Tag=" + _clipboardKeys[0];
+        "&Tag=" + _clipboardKeys[0]);
 
     if (!encode)
         return meta;
@@ -297,21 +318,11 @@ void ClientSession::handleClipboardRequest(DocumentBroker::ClipboardRequest
     }
 }
 
-void ClientSession::handleIncomingMessage(SocketDisposition &disposition)
-{
-    // LOG_TRC("***** ClientSession::handleIncomingMessage()");
-    if (UnitWSD::get().filterHandleRequest(
-            UnitWSD::TestRequest::Client, disposition, *this))
-        return;
-
-    Session::handleIncomingMessage(disposition);
-}
-
 bool ClientSession::_handleInput(const char *buffer, int length)
 {
     LOG_TRC(getName() << ": handling incoming [" << getAbbreviatedMessage(buffer, length) << "].");
     const std::string firstLine = getFirstLine(buffer, length);
-    const std::vector<std::string> tokens = LOOLProtocol::tokenize(firstLine.data(), firstLine.size());
+    const StringVector tokens = LOOLProtocol::tokenize(firstLine.data(), firstLine.size());
 
     std::shared_ptr<DocumentBroker> docBroker = getDocumentBroker();
     if (!docBroker || docBroker->isMarkedToDestroy())
@@ -322,7 +333,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 
     if (tokens.size() < 1)
     {
-        sendTextFrame("error: cmd=empty kind=unknown");
+        sendTextFrameAndLogError("error: cmd=empty kind=unknown");
         return false;
     }
 
@@ -334,11 +345,11 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         updateLastActivityTime();
         docBroker->updateLastActivityTime();
     }
-    if (tokens[0] == "loolclient")
+    if (tokens.equals(0, "loolclient"))
     {
         if (tokens.size() < 2)
         {
-            sendTextFrame("error: cmd=loolclient kind=badprotocolversion");
+            sendTextFrameAndLogError("error: cmd=loolclient kind=badprotocolversion");
             return false;
         }
 
@@ -346,7 +357,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         if (std::get<0>(versionTuple) != ProtocolMajorVersionNumber ||
             std::get<1>(versionTuple) != ProtocolMinorVersionNumber)
         {
-            sendTextFrame("error: cmd=loolclient kind=badprotocolversion");
+            sendTextFrameAndLogError("error: cmd=loolclient kind=badprotocolversion");
             return false;
         }
 
@@ -354,22 +365,28 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         sendTextFrame("loolserver " + LOOLWSD::getVersionJSON());
         // Send LOKit version information
         sendTextFrame("lokitversion " + LOOLWSD::LOKitVersion);
+
+        #if !MOBILEAPP
+            // If it is not mobile, it must be Linux (for now).
+            sendTextFrame(std::string("osinfo ") + Util::getLinuxVersion());
+        #endif
+
         // Send clipboard key
         rotateClipboardKey(true);
 
         return true;
     }
 
-    if (tokens[0] == "jserror")
+    if (tokens.equals(0, "jserror"))
     {
         LOG_ERR(std::string(buffer, length));
         return true;
     }
-    else if (tokens[0] == "load")
+    else if (tokens.equals(0, "load"))
     {
         if (getDocURL() != "")
         {
-            sendTextFrame("error: cmd=load kind=docalreadyloaded");
+            sendTextFrameAndLogError("error: cmd=load kind=docalreadyloaded");
             return false;
         }
 
@@ -404,6 +421,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
              tokens[0] != "savetostorage" &&
              tokens[0] != "selectgraphic" &&
              tokens[0] != "selecttext" &&
+             tokens[0] != "windowselecttext" &&
              tokens[0] != "setclientpart" &&
              tokens[0] != "selectclientpart" &&
              tokens[0] != "moveselectedclientparts" &&
@@ -426,27 +444,29 @@ bool ClientSession::_handleInput(const char *buffer, int length)
              tokens[0] != "renamefile" &&
              tokens[0] != "resizewindow" &&
              tokens[0] != "removetextcontext" &&
-             tokens[0] != "dialogevent")
+             tokens[0] != "dialogevent" &&
+             tokens[0] != "completefunction" &&
+             tokens[0] != "formfieldevent")
     {
         LOG_ERR("Session [" << getId() << "] got unknown command [" << tokens[0] << "].");
-        sendTextFrame("error: cmd=" + tokens[0] + " kind=unknown");
+        sendTextFrameAndLogError("error: cmd=" + tokens[0] + " kind=unknown");
         return false;
     }
     else if (getDocURL() == "")
     {
-        sendTextFrame("error: cmd=" + tokens[0] + " kind=nodocloaded");
+        sendTextFrameAndLogError("error: cmd=" + tokens[0] + " kind=nodocloaded");
         return false;
     }
-    else if (tokens[0] == "canceltiles")
+    else if (tokens.equals(0, "canceltiles"))
     {
-        docBroker->cancelTileRequests(shared_from_this());
+        docBroker->cancelTileRequests(client_from_this());
         return true;
     }
-    else if (tokens[0] == "commandvalues")
+    else if (tokens.equals(0, "commandvalues"))
     {
         return getCommandValues(buffer, length, tokens, docBroker);
     }
-    else if (tokens[0] == "closedocument")
+    else if (tokens.equals(0, "closedocument"))
     {
         // If this session is the owner of the file & 'EnableOwnerTermination' feature
         // is turned on by WOPI, let it close all sessions
@@ -465,46 +485,46 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 
         return true;
     }
-    else if (tokens[0] == "versionrestore")
+    else if (tokens.equals(0, "versionrestore"))
     {
-        if (tokens.size() > 1 && tokens[1] == "prerestore")
+        if (tokens.size() > 1 && tokens.equals(1, "prerestore"))
         {
             // green signal to WOPI host to restore the version *after* saving
             // any unsaved changes, if any, to the storage
             docBroker->closeDocument("versionrestore: prerestore_ack");
         }
     }
-    else if (tokens[0] == "partpagerectangles")
+    else if (tokens.equals(0, "partpagerectangles"))
     {
         // We don't support partpagerectangles any more, will be removed in the
         // next version
         sendTextFrame("partpagerectangles: ");
         return true;
     }
-    else if (tokens[0] == "ping")
+    else if (tokens.equals(0, "ping"))
     {
         std::string count = std::to_string(docBroker->getRenderedTileCount());
         sendTextFrame("pong rendercount=" + count);
         return true;
     }
-    else if (tokens[0] == "renderfont")
+    else if (tokens.equals(0, "renderfont"))
     {
         return sendFontRendering(buffer, length, tokens, docBroker);
     }
-    else if (tokens[0] == "status" || tokens[0] == "statusupdate")
+    else if (tokens.equals(0, "status") || tokens.equals(0, "statusupdate"))
     {
         assert(firstLine.size() == static_cast<size_t>(length));
         return forwardToChild(firstLine, docBroker);
     }
-    else if (tokens[0] == "tile")
+    else if (tokens.equals(0, "tile"))
     {
         return sendTile(buffer, length, tokens, docBroker);
     }
-    else if (tokens[0] == "tilecombine")
+    else if (tokens.equals(0, "tilecombine"))
     {
         return sendCombinedTiles(buffer, length, tokens, docBroker);
     }
-    else if (tokens[0] == "save")
+    else if (tokens.equals(0, "save"))
     {
         if (isReadOnly())
         {
@@ -523,26 +543,34 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 
             std::string extendedData;
             if (tokens.size() > 3)
+            {
                 getTokenString(tokens[3], "extendedData", extendedData);
+                std::string decoded;
+                Poco::URI::decode(extendedData, decoded);
+                extendedData = decoded;
+            }
 
             constexpr bool isAutosave = false;
             constexpr bool isExitSave = false;
-            docBroker->sendUnoSave(getId(), dontTerminateEdit != 0, dontSaveIfUnmodified != 0,
+            bool result = docBroker->sendUnoSave(getId(), dontTerminateEdit != 0, dontSaveIfUnmodified != 0,
                                     isAutosave, isExitSave, extendedData);
+            std::string resultstr = result ? "true" : "false";
+            std::string msg = "commandresult: { \"command\": \"save\", \"success\": " + resultstr + " }";
+            docBroker->broadcastMessage(msg);
         }
     }
-    else if (tokens[0] == "savetostorage")
+    else if (tokens.equals(0, "savetostorage"))
     {
         int force = 0;
         if (tokens.size() > 1)
             getTokenInteger(tokens[1], "force", force);
 
-        if (docBroker->saveToStorage(getId(), true, "" /* This is irrelevant when success is true*/, true))
-        {
-            docBroker->broadcastMessage("commandresult: { \"command\": \"savetostorage\", \"success\": true }");
-        }
+        bool result = docBroker->saveToStorage(getId(), true, "" /* This is irrelevant when success is true*/, true);
+        std::string resultstr = result ? "true" : "false";
+        std::string msg = "commandresult: { \"command\": \"savetostorage\", \"success\": " + resultstr + " }";
+        docBroker->broadcastMessage(msg);
     }
-    else if (tokens[0] == "clientvisiblearea")
+    else if (tokens.equals(0, "clientvisiblearea"))
     {
         int x;
         int y;
@@ -555,7 +583,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             !getTokenInteger(tokens[4], "height", height))
         {
             // Be forgiving and log instead of disconnecting.
-            // sendTextFrame("error: cmd=clientvisiblearea kind=syntax");
+            // sendTextFrameAndLogError("error: cmd=clientvisiblearea kind=syntax");
             LOG_WRN("Invalid syntax for '" << tokens[0] << "' message: [" << firstLine << "].");
             return true;
         }
@@ -566,7 +594,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             return forwardToChild(std::string(buffer, length), docBroker);
         }
     }
-    else if (tokens[0] == "setclientpart")
+    else if (tokens.equals(0, "setclientpart"))
     {
         if(!_isTextDocument)
         {
@@ -574,7 +602,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             if (tokens.size() != 2 ||
                 !getTokenInteger(tokens[1], "part", temp))
             {
-                sendTextFrame("error: cmd=setclientpart kind=syntax");
+                sendTextFrameAndLogError("error: cmd=setclientpart kind=syntax");
                 return false;
             }
             else
@@ -585,7 +613,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             }
         }
     }
-    else if (tokens[0] == "selectclientpart")
+    else if (tokens.equals(0, "selectclientpart"))
     {
         if(!_isTextDocument)
         {
@@ -595,7 +623,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
                 !getTokenInteger(tokens[1], "part", part) ||
                 !getTokenInteger(tokens[2], "how", how))
             {
-                sendTextFrame("error: cmd=selectclientpart kind=syntax");
+                sendTextFrameAndLogError("error: cmd=selectclientpart kind=syntax");
                 return false;
             }
             else
@@ -604,7 +632,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             }
         }
     }
-    else if (tokens[0] == "moveselectedclientparts")
+    else if (tokens.equals(0, "moveselectedclientparts"))
     {
         if(!_isTextDocument)
         {
@@ -612,7 +640,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             if (tokens.size() != 2 ||
                 !getTokenInteger(tokens[1], "position", nPosition))
             {
-                sendTextFrame("error: cmd=moveselectedclientparts kind=syntax");
+                sendTextFrameAndLogError("error: cmd=moveselectedclientparts kind=syntax");
                 return false;
             }
             else
@@ -621,7 +649,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             }
         }
     }
-    else if (tokens[0] == "clientzoom")
+    else if (tokens.equals(0, "clientzoom"))
     {
         int tilePixelWidth, tilePixelHeight, tileTwipWidth, tileTwipHeight;
         if (tokens.size() != 5 ||
@@ -631,7 +659,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             !getTokenInteger(tokens[4], "tiletwipheight", tileTwipHeight))
         {
             // Be forgiving and log instead of disconnecting.
-            // sendTextFrame("error: cmd=clientzoom kind=syntax");
+            // sendTextFrameAndLogError("error: cmd=clientzoom kind=syntax");
             LOG_WRN("Invalid syntax for '" << tokens[0] << "' message: [" << firstLine << "].");
             return true;
         }
@@ -645,14 +673,14 @@ bool ClientSession::_handleInput(const char *buffer, int length)
             return forwardToChild(std::string(buffer, length), docBroker);
         }
     }
-    else if (tokens[0] == "tileprocessed")
+    else if (tokens.equals(0, "tileprocessed"))
     {
         std::string tileID;
         if (tokens.size() != 2 ||
             !getTokenString(tokens[1], "tile", tileID))
         {
             // Be forgiving and log instead of disconnecting.
-            // sendTextFrame("error: cmd=tileprocessed kind=syntax");
+            // sendTextFrameAndLogError("error: cmd=tileprocessed kind=syntax");
             LOG_WRN("Invalid syntax for '" << tokens[0] << "' message: [" << firstLine << "].");
             return true;
         }
@@ -668,10 +696,10 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         else
             LOG_INF("Tileprocessed message with an unknown tile ID");
 
-        docBroker->sendRequestedTiles(shared_from_this());
+        docBroker->sendRequestedTiles(client_from_this());
         return true;
     }
-    else if (tokens[0] == "removesession") {
+    else if (tokens.equals(0, "removesession")) {
         if (tokens.size() > 1 && (_isDocumentOwner || !isReadOnly()))
         {
             std::string sessionId = Util::encodeId(std::stoi(tokens[1]), 4);
@@ -681,13 +709,13 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         else
             LOG_WRN("Readonly session '" << getId() << "' trying to kill another view");
     }
-    else if (tokens[0] == "renamefile")
+    else if (tokens.equals(0, "renamefile"))
     {
         std::string encodedWopiFilename;
         if (tokens.size() < 2 || !getTokenString(tokens[1], "filename", encodedWopiFilename))
         {
             LOG_ERR("Bad syntax for: " << firstLine);
-            sendTextFrame("error: cmd=renamefile kind=syntax");
+            sendTextFrameAndLogError("error: cmd=renamefile kind=syntax");
             return false;
         }
         std::string wopiFilename;
@@ -695,13 +723,21 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         docBroker->saveAsToStorage(getId(), "", wopiFilename, true);
         return true;
     }
-    else if (tokens[0] == "dialogevent")
+    else if (tokens.equals(0, "dialogevent"))
+    {
+        return forwardToChild(firstLine, docBroker);
+    }
+    else if (tokens.equals(0, "completefunction"))
+    {
+        return forwardToChild(std::string(buffer, length), docBroker);
+    }
+    else if (tokens.equals(0, "formfieldevent"))
     {
         return forwardToChild(firstLine, docBroker);
     }
     else
     {
-        if (tokens[0] == "key")
+        if (tokens.equals(0, "key"))
             _keyEvents++;
 
         if (!filterMessage(firstLine))
@@ -715,7 +751,7 @@ bool ClientSession::_handleInput(const char *buffer, int length)
         }
         else
         {
-            assert(tokens[0] == "requestloksession");
+            assert(tokens.equals(0, "requestloksession"));
             return true;
         }
     }
@@ -724,16 +760,17 @@ bool ClientSession::_handleInput(const char *buffer, int length)
 }
 
 bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
-                                 const std::vector<std::string>& tokens,
+                                 const StringVector& tokens,
                                  const std::shared_ptr<DocumentBroker>& docBroker)
 {
     if (tokens.size() < 2)
     {
         // Failed loading ends connection.
-        sendTextFrame("error: cmd=load kind=syntax");
+        sendTextFrameAndLogError("error: cmd=load kind=syntax");
         return false;
     }
 
+    _viewLoadStart = std::chrono::steady_clock::now();
     LOG_INF("Requesting document load from child.");
     try
     {
@@ -786,6 +823,11 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
             oss << " lang=" << getLang();
         }
 
+        if (!getDeviceFormFactor().empty())
+        {
+            oss << " deviceFormFactor=" << getDeviceFormFactor();
+        }
+
         if (!getWatermarkText().empty())
         {
             std::string encodedWatermarkText;
@@ -808,18 +850,18 @@ bool ClientSession::loadDocument(const char* /*buffer*/, int /*length*/,
     }
     catch (const Poco::SyntaxException&)
     {
-        sendTextFrame("error: cmd=load kind=uriinvalid");
+        sendTextFrameAndLogError("error: cmd=load kind=uriinvalid");
     }
 
     return false;
 }
 
-bool ClientSession::getCommandValues(const char *buffer, int length, const std::vector<std::string>& tokens,
+bool ClientSession::getCommandValues(const char *buffer, int length, const StringVector& tokens,
                                      const std::shared_ptr<DocumentBroker>& docBroker)
 {
     std::string command;
     if (tokens.size() != 2 || !getTokenString(tokens[1], "command", command))
-        return sendTextFrame("error: cmd=commandvalues kind=syntax");
+        return sendTextFrameAndLogError("error: cmd=commandvalues kind=syntax");
 
     std::string cmdValues;
     if (docBroker->tileCache().getTextStream(TileCache::StreamType::CmdValues, command, cmdValues))
@@ -828,14 +870,14 @@ bool ClientSession::getCommandValues(const char *buffer, int length, const std::
     return forwardToChild(std::string(buffer, length), docBroker);
 }
 
-bool ClientSession::sendFontRendering(const char *buffer, int length, const std::vector<std::string>& tokens,
+bool ClientSession::sendFontRendering(const char *buffer, int length, const StringVector& tokens,
                                       const std::shared_ptr<DocumentBroker>& docBroker)
 {
     std::string font, text;
     if (tokens.size() < 2 ||
         !getTokenString(tokens[1], "font", font))
     {
-        return sendTextFrame("error: cmd=renderfont kind=syntax");
+        return sendTextFrameAndLogError("error: cmd=renderfont kind=syntax");
     }
 
     getTokenString(tokens[2], "char", text);
@@ -843,45 +885,45 @@ bool ClientSession::sendFontRendering(const char *buffer, int length, const std:
     TileCache::Tile cachedTile = docBroker->tileCache().lookupCachedStream(TileCache::StreamType::Font, font+text);
     if (cachedTile)
     {
-        const std::string response = "renderfont: " + Poco::cat(std::string(" "), tokens.begin() + 1, tokens.end()) + "\n";
+        const std::string response = "renderfont: " + tokens.cat(std::string(" "), 1) + "\n";
         return sendTile(response, cachedTile);
     }
 
     return forwardToChild(std::string(buffer, length), docBroker);
 }
 
-bool ClientSession::sendTile(const char * /*buffer*/, int /*length*/, const std::vector<std::string>& tokens,
+bool ClientSession::sendTile(const char * /*buffer*/, int /*length*/, const StringVector& tokens,
                              const std::shared_ptr<DocumentBroker>& docBroker)
 {
     try
     {
         TileDesc tileDesc = TileDesc::parse(tokens);
-        tileDesc.setNormalizedViewId(getHash());
-        docBroker->handleTileRequest(tileDesc, shared_from_this());
+        tileDesc.setNormalizedViewId(getCanonicalViewId());
+        docBroker->handleTileRequest(tileDesc, client_from_this());
     }
     catch (const std::exception& exc)
     {
         LOG_ERR("Failed to process tile command: " << exc.what());
-        return sendTextFrame("error: cmd=tile kind=invalid");
+        return sendTextFrameAndLogError("error: cmd=tile kind=invalid");
     }
 
     return true;
 }
 
-bool ClientSession::sendCombinedTiles(const char* /*buffer*/, int /*length*/, const std::vector<std::string>& tokens,
+bool ClientSession::sendCombinedTiles(const char* /*buffer*/, int /*length*/, const StringVector& tokens,
                                       const std::shared_ptr<DocumentBroker>& docBroker)
 {
     try
     {
         TileCombined tileCombined = TileCombined::parse(tokens);
-        tileCombined.setNormalizedViewId(getHash());
-        docBroker->handleTileCombinedRequest(tileCombined, shared_from_this());
+        tileCombined.setNormalizedViewId(getCanonicalViewId());
+        docBroker->handleTileCombinedRequest(tileCombined, client_from_this());
     }
     catch (const std::exception& exc)
     {
         LOG_ERR("Failed to process tilecombine command: " << exc.what());
         // Be forgiving and log instead of disconnecting.
-        // return sendTextFrame("error: cmd=tile kind=invalid");
+        // return sendTextFrameAndLogError("error: cmd=tile kind=invalid");
     }
 
     return true;
@@ -896,13 +938,13 @@ bool ClientSession::forwardToChild(const std::string& message,
 bool ClientSession::filterMessage(const std::string& message) const
 {
     bool allowed = true;
-    std::vector<std::string> tokens(LOOLProtocol::tokenize(message, ' '));
+    StringVector tokens(LOOLProtocol::tokenize(message, ' '));
 
     // Set allowed flag to false depending on if particular WOPI properties are set
-    if (tokens[0] == "downloadas")
+    if (tokens.equals(0, "downloadas"))
     {
         std::string id;
-        if (getTokenString(tokens[2], "id", id))
+        if (tokens.size() >= 3 && getTokenString(tokens[2], "id", id))
         {
             if (id == "print" && _wopiFileInfo && _wopiFileInfo->getDisablePrint())
             {
@@ -921,7 +963,7 @@ bool ClientSession::filterMessage(const std::string& message) const
             LOG_WRN("No value of id in downloadas message");
         }
     }
-    else if (tokens[0] == "gettextselection" || tokens[0] == ".uno:Copy")
+    else if (tokens.equals(0, "gettextselection") || tokens.equals(0, ".uno:Copy"))
     {
         if (_wopiFileInfo && _wopiFileInfo->getDisableCopy())
         {
@@ -933,13 +975,13 @@ bool ClientSession::filterMessage(const std::string& message) const
     {
         // By default, don't allow anything
         allowed = false;
-        if (tokens[0] == "userinactive" || tokens[0] == "useractive" || tokens[0] == "saveas")
+        if (tokens.equals(0, "userinactive") || tokens.equals(0, "useractive") || tokens.equals(0, "saveas"))
         {
             allowed = true;
         }
-        else if (tokens[0] == "uno")
+        else if (tokens.equals(0, "uno"))
         {
-            if (tokens[1] == ".uno:ExecuteSearch")
+            if (tokens.size() > 1 && tokens.equals(1, ".uno:ExecuteSearch"))
             {
                 allowed = true;
             }
@@ -956,17 +998,13 @@ void ClientSession::setReadOnly()
     sendTextFrame("perm: readonly");
 }
 
-int ClientSession::getPollEvents(std::chrono::steady_clock::time_point /* now */,
-                                 int & /* timeoutMaxMs */)
+bool ClientSession::hasQueuedMessages() const
 {
-    LOG_TRC(getName() << " ClientSession has " << _senderQueue.size() << " write message(s) queued.");
-    int events = POLLIN;
-    if (_senderQueue.size())
-        events |= POLLOUT;
-    return events;
+    return _senderQueue.size() > 0;
 }
 
-void ClientSession::performWrites()
+    /// Please send them to me then.
+void ClientSession::writeQueuedMessages()
 {
     LOG_TRC(getName() << " ClientSession: performing writes.");
 
@@ -995,6 +1033,7 @@ void ClientSession::performWrites()
     LOG_TRC(getName() << " ClientSession: performed write.");
 }
 
+// NB. also see loleaflet/src/map/Clipboard.js that does this in JS for stubs.
 void ClientSession::postProcessCopyPayload(std::shared_ptr<Message> payload)
 {
     // Insert our meta origin if we can
@@ -1008,6 +1047,7 @@ void ClientSession::postProcessCopyPayload(std::shared_ptr<Message> payload)
             if (pos != std::string::npos) // assume text/html
             {
                 const std::string meta = getClipboardURI();
+                LOG_TRC("Inject clipboard meta origin of '" << meta << "'");
                 const std::string origin = "<meta name=\"origin\" content=\"" + meta + "\"/>\n";
                 data.insert(data.begin() + pos, origin.begin(), origin.end());
                 return true;
@@ -1156,7 +1196,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
             // the session
             if (!isConvertTo)
             {
-                sendTextFrame("error: cmd=saveas kind=syntax");
+                sendTextFrameAndLogError("error: cmd=saveas kind=syntax");
                 return false;
             }
         }
@@ -1165,7 +1205,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         if (!isConvertTo && !getTokenString(tokens[2], "filename", encodedWopiFilename))
         {
             LOG_ERR("Bad syntax for: " << firstLine);
-            sendTextFrame("error: cmd=saveas kind=syntax");
+            sendTextFrameAndLogError("error: cmd=saveas kind=syntax");
             return false;
         }
 
@@ -1175,13 +1215,15 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
 
         // URI constructor implicitly decodes when it gets std::string as param
         Poco::URI resultURL(encodedURL);
-        if (resultURL.getScheme() == "file")
+
+        // Prepend the jail path in the normal (non-nocaps) case
+        if (resultURL.getScheme() == "file" && !LOOLWSD::NoCapsForKit)
         {
             std::string relative(resultURL.getPath());
             if (relative.size() > 0 && relative[0] == '/')
                 relative = relative.substr(1);
 
-            // Rewrite file:// URLs, as they are visible to the outside world.
+            // Rewrite file:// URLs to be visible to the outside world.
             const Path path(docBroker->getJailRoot(), relative);
             if (Poco::File(path).exists())
             {
@@ -1211,7 +1253,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
                 docBroker->saveAsToStorage(getId(), resultURL.getPath(), wopiFilename, false);
             }
             else
-                sendTextFrame("error: cmd=storage kind=savefailed");
+                sendTextFrameAndLogError("error: cmd=storage kind=savefailed");
         }
         else
         {
@@ -1245,12 +1287,12 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         return true;
     }
 #endif
-    else if (tokens.size() == 2 && tokens[0] == "statechanged:")
+    else if (tokens.size() == 2 && tokens.equals(0, "statechanged:"))
     {
-        std::vector<std::string> stateTokens(LOOLProtocol::tokenize(tokens[1], '='));
-        if (stateTokens.size() == 2 && stateTokens[0] == ".uno:ModifiedStatus")
+        StringVector stateTokens(LOOLProtocol::tokenize(tokens[1], '='));
+        if (stateTokens.size() == 2 && stateTokens.equals(0, ".uno:ModifiedStatus"))
         {
-            docBroker->setModified(stateTokens[1] == "true");
+            docBroker->setModified(stateTokens.equals(1, "true"));
         }
         else
         {
@@ -1328,12 +1370,16 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
                 << "X-Content-Type-Options: nosniff\r\n"
                 << "\r\n";
             auto socket = it.lock();
+            if (!socket)
+                continue;
+
             if (!empty)
             {
                 oss.write(&payload->data()[header], payload->size() - header);
                 socket->setSocketBufferSize(std::min(payload->size() + 256,
                                                      size_t(Socket::MaximumSendBufferSize)));
             }
+
             socket->send(oss.str());
             socket->shutdown();
             LOG_INF("Queued " << (empty?"empty":"clipboard") << " response for send.");
@@ -1358,6 +1404,11 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
         {
             setState(ClientSession::SessionState::LIVE);
             docBroker->setLoaded();
+
+#if !MOBILEAPP
+            Admin::instance().setViewLoadDuration(docBroker->getDocKey(), getId(), std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() - _viewLoadStart));
+#endif
+
             // Wopi post load actions
             if (_wopiFileInfo && !_wopiFileInfo->getTemplateSource().empty())
             {
@@ -1370,7 +1421,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
             {
                 // Need to get the initial part id from status message
                 int part = -1;
-                if(getTokenInteger(token, "current", part))
+                if(getTokenInteger(tokens.getParam(token), "current", part))
                 {
                     _clientSelectedPart = part;
                     resetWireIdMap();
@@ -1378,14 +1429,14 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
 
                 // Get document type too
                 std::string docType;
-                if(getTokenString(token, "type", docType))
+                if(getTokenString(tokens.getParam(token), "type", docType))
                 {
                     _isTextDocument = docType.find("text") != std::string::npos;
                 }
 
                 // Store our Kit ViewId
                 int viewId = -1;
-                if(getTokenInteger(token, "viewid", viewId))
+                if(getTokenInteger(tokens.getParam(token), "viewid", viewId))
                     _kitViewId = viewId;
             }
 
@@ -1431,7 +1482,7 @@ bool ClientSession::handleKitToClientMessage(const char* buffer, const int lengt
             const Poco::Dynamic::Var result = parser.parse(stringJSON);
             const auto& object = result.extract<Poco::JSON::Object::Ptr>();
             const std::string rectangle = object->get("rectangle").toString();
-            std::vector<std::string> rectangleTokens(LOOLProtocol::tokenize(rectangle, ','));
+            StringVector rectangleTokens(LOOLProtocol::tokenize(rectangle, ','));
             int x = 0, y = 0, w = 0, h = 0;
             if (rectangleTokens.size() > 2 &&
                 stringToInteger(rectangleTokens[0], x) &&
@@ -1609,6 +1660,9 @@ void ClientSession::onDisconnect()
     docBroker->assertCorrectThread();
     const std::string docKey = docBroker->getDocKey();
 
+    // Keep self alive, so that our own dtor runs only at the end of this function. Without this,
+    // removeSession() may destroy us and then we can't call our own member functions anymore.
+    std::shared_ptr<ClientSession> session = client_from_this();
     try
     {
         // Connection terminated. Destroy session.
@@ -1635,20 +1689,20 @@ void ClientSession::onDisconnect()
             LOG_TRC("Normal close handshake.");
             // Client initiated close handshake
             // respond with close frame
-            shutdown();
+            shutdownNormal();
         }
         else if (!SigUtil::getShutdownRequestFlag())
         {
             // something wrong, with internal exceptions
             LOG_TRC("Abnormal close handshake.");
             closeFrame();
-            shutdown(WebSocketHandler::StatusCodes::ENDPOINT_GOING_AWAY);
+            shutdownGoingAway();
         }
         else
         {
             LOG_TRC("Server recycling.");
             closeFrame();
-            shutdown(WebSocketHandler::StatusCodes::ENDPOINT_GOING_AWAY);
+            shutdownGoingAway();
         }
     }
     catch (const std::exception& exc)
@@ -1670,17 +1724,17 @@ void ClientSession::dumpState(std::ostream& os)
        << "\n\t\ttile size Pixel: " << _tileWidthPixel << "x" << _tileHeightPixel
        << "\n\t\ttile size Twips: " << _tileWidthTwips << "x" << _tileHeightTwips
        << "\n\t\tkit ViewId: " << _kitViewId
-       << "\n\t\thost (un-trusted): " << _hostNoTrust
+       << "\n\t\tour URL (un-trusted): " << _serverURL.getSubURLForEndpoint("")
        << "\n\t\tisTextDocument: " << _isTextDocument
        << "\n\t\tclipboardKeys[0]: " << _clipboardKeys[0]
        << "\n\t\tclipboardKeys[1]: " << _clipboardKeys[1]
-       << "\n\t\tclip sockets: " << _clipSockets.size();
+       << "\n\t\tclip sockets: " << _clipSockets.size()
+       << "\n\t\tproxy access:: " << _proxyAccess;
 
-    std::shared_ptr<StreamSocket> socket = getSocket().lock();
-    if (socket)
+    if (_protocol)
     {
-        uint64_t sent, recv;
-        socket->getIOStats(sent, recv);
+        uint64_t sent = 0, recv = 0;
+        _protocol->getIOStats(sent, recv);
         os << "\n\t\tsent/keystroke: " << (double)sent/_keyEvents << "bytes";
     }
 
@@ -1689,10 +1743,18 @@ void ClientSession::dumpState(std::ostream& os)
 
 }
 
+const std::string &ClientSession::getOrCreateProxyAccess()
+{
+    if (_proxyAccess.size() <= 0)
+        _proxyAccess = Util::rng::getHardRandomHexString(
+            ProxyAccessTokenLengthBytes);
+    return _proxyAccess;
+}
+
 void ClientSession::handleTileInvalidation(const std::string& message,
     const std::shared_ptr<DocumentBroker>& docBroker)
 {
-    docBroker->invalidateTiles(message, getHash());
+    docBroker->invalidateTiles(message, getCanonicalViewId());
 
     // Skip requesting new tiles if we don't have client visible area data yet.
     if(!_clientVisibleArea.hasSurface() ||
@@ -1717,7 +1779,7 @@ void ClientSession::handleTileInvalidation(const std::string& message,
     if( part == -1 ) // If no part is specified we use the part used by the client
         part = _clientSelectedPart;
 
-    int normalizedViewId = getHash();
+    int normalizedViewId = getCanonicalViewId();
 
     std::vector<TileDesc> invalidTiles;
     if(part == _clientSelectedPart || _isTextDocument)
@@ -1751,7 +1813,7 @@ void ClientSession::handleTileInvalidation(const std::string& message,
     {
         TileCombined tileCombined = TileCombined::create(invalidTiles);
         tileCombined.setNormalizedViewId(normalizedViewId);
-        docBroker->handleTileCombinedRequest(tileCombined, shared_from_this());
+        docBroker->handleTileCombinedRequest(tileCombined, client_from_this());
     }
 }
 
