@@ -75,7 +75,7 @@ void sendLastModificationTime(const std::shared_ptr<Session>& session,
 
     std::stringstream stream;
     stream << "lastmodtime: " << documentLastModifiedTime;
-    std::string message = stream.str();
+    const std::string message = stream.str();
     session->sendTextFrame(message);
     if (documentBroker)
         documentBroker->broadcastMessage(message);
@@ -86,9 +86,13 @@ void sendLastModificationTime(const std::shared_ptr<Session>& session,
 Poco::URI DocumentBroker::sanitizeURI(const std::string& uri)
 {
     // The URI of the document should be url-encoded.
+#if !MOBILEAPP
     std::string decodedUri;
     Poco::URI::decode(uri, decodedUri);
     Poco::URI uriPublic(decodedUri);
+#else
+    Poco::URI uriPublic(uri);
+#endif
 
     if (uriPublic.isRelative() || uriPublic.getScheme() == "file")
     {
@@ -168,10 +172,13 @@ public:
 
 std::atomic<unsigned> DocumentBroker::DocBrokerId(1);
 
-DocumentBroker::DocumentBroker(const std::string& uri,
+DocumentBroker::DocumentBroker(ChildType type,
+                               const std::string& uri,
                                const Poco::URI& uriPublic,
                                const std::string& docKey) :
+    _limitLifeSeconds(0),
     _uriOrig(uri),
+    _type(type),
     _uriPublic(uriPublic),
     _docKey(docKey),
     _docId(Util::encodeId(DocBrokerId++, 3)),
@@ -189,8 +196,10 @@ DocumentBroker::DocumentBroker(const std::string& uri,
     _poll(new DocumentBrokerPoll("docbroker_" + _docId, *this)),
     _stop(false),
     _closeReason("stopped"),
+    _lockCtx(new LockContext()),
     _tileVersion(0),
-    _debugRenderedTileCount(0)
+    _debugRenderedTileCount(0),
+    _wopiLoadDuration(0)
 {
     assert(!_docKey.empty());
     assert(!LOOLWSD::ChildRoot.empty());
@@ -199,6 +208,17 @@ DocumentBroker::DocumentBroker(const std::string& uri,
             "] created with docKey [" << _docKey << "]");
 }
 
+void DocumentBroker::setupPriorities()
+{
+#if !MOBILEAPP
+    if (_type == ChildType::Batch)
+    {
+        int prio = LOOLWSD::getConfigValue<int>("per_document.batch_priority", 5);
+        Util::setProcessAndThreadPriorities(_childProcess->getPid(), prio);
+    }
+#endif // !MOBILE
+}
+
 void DocumentBroker::startThread()
 {
     _poll->startThread();
@@ -221,7 +241,7 @@ void DocumentBroker::pollThread()
     do
     {
         static const int timeoutMs = COMMAND_TIMEOUT_MS * 5;
-        _childProcess = getNewChild_Blocks();
+        _childProcess = getNewChild_Blocks(getPublicUri().getPath());
         if (_childProcess ||
             std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::steady_clock::now() -
                                                                   _threadStart).count() > timeoutMs)
@@ -264,6 +284,8 @@ void DocumentBroker::pollThread()
     _childProcess->setDocumentBroker(shared_from_this());
     LOG_INF("Doc [" << _docKey << "] attached to child [" << _childProcess->getPid() << "].");
 
+    setupPriorities();
+
     static const bool AutoSaveEnabled = !std::getenv("LOOL_NO_AUTOSAVE");
 
 #if !MOBILEAPP
@@ -276,7 +298,14 @@ void DocumentBroker::pollThread()
     auto lastBWUpdateTime = std::chrono::steady_clock::now();
     auto lastClipboardHashUpdateTime = std::chrono::steady_clock::now();
 
-    const int limit_load_secs = LOOLWSD::getConfigValue<int>("per_document.limit_load_secs", 100);
+    const int limit_load_secs =
+#ifdef ENABLE_DEBUG
+        // paused waiting for a debugger to attach
+        // ignore load time out
+        std::getenv("PAUSEFORDEBUGGER") ? -1 :
+#endif
+        LOOLWSD::getConfigValue<int>("per_document.limit_load_secs", 100);
+
     const auto loadDeadline = std::chrono::steady_clock::now() + std::chrono::seconds(limit_load_secs);
 #endif
     auto last30SecCheckTime = std::chrono::steady_clock::now();
@@ -284,7 +313,7 @@ void DocumentBroker::pollThread()
     // Main polling loop goodness.
     while (!_stop && _poll->continuePolling() && !SigUtil::getTerminationFlag())
     {
-        _poll->poll(SocketPoll::DefaultPollTimeoutMs);
+        _poll->poll(SocketPoll::DefaultPollTimeoutMicroS);
 
         const auto now = std::chrono::steady_clock::now();
 
@@ -295,9 +324,9 @@ void DocumentBroker::pollThread()
 
         if (!_isLoaded && (limit_load_secs > 0) && (now > loadDeadline))
         {
-            LOG_WRN("Doc [" << _docKey << "] is taking too long to load. Will kill process ["
-                            << _childProcess->getPid() << "]. per_document.limit_load_secs set to "
-                            << limit_load_secs << " secs.");
+            LOG_ERR("Doc [" << _docKey << "] is taking too long to load. Will kill process ["
+                    << _childProcess->getPid() << "]. per_document.limit_load_secs set to "
+                    << limit_load_secs << " secs.");
             broadcastMessage("error: cmd=load kind=docloadtimeout");
 
             // Brutal but effective.
@@ -308,21 +337,50 @@ void DocumentBroker::pollThread()
             continue;
         }
 
+        if (_limitLifeSeconds > 0 &&
+            std::chrono::duration_cast<std::chrono::seconds>(now - _threadStart).count() > _limitLifeSeconds)
+        {
+            LOG_WRN("Doc [" << _docKey << "] is taking too long to convert. Will kill process ["
+                            << _childProcess->getPid() << "]. per_document.limit_convert_secs set to "
+                            << _limitLifeSeconds << " secs.");
+            broadcastMessage("error: cmd=load kind=docexpired");
+
+            // Brutal but effective.
+            if (_childProcess)
+                _childProcess->terminate();
+
+            stop("Load timed out");
+            continue;
+        }
+
         if (std::chrono::duration_cast<std::chrono::milliseconds>
                     (now - lastBWUpdateTime).count() >= COMMAND_TIMEOUT_MS)
         {
             lastBWUpdateTime = now;
-            uint64_t sent, recv;
+            uint64_t sent = 0, recv = 0;
             getIOStats(sent, recv);
+
+            uint64_t deltaSent = 0, deltaRecv = 0;
+
+            // connection drop transiently reduces this.
+            if (sent > adminSent)
+            {
+                deltaSent = sent - adminSent;
+                adminSent = sent;
+            }
+            if (recv > deltaRecv)
+            {
+                deltaRecv = recv - adminRecv;
+                adminRecv = recv;
+            }
+            LOG_TRC("Doc [" << _docKey << "] added stats sent: +" << deltaSent << ", recv: +" << deltaRecv << " bytes to totals.");
+
             // send change since last notification.
-            Admin::instance().addBytes(getDocKey(),
-                                       // connection drop transiently reduces this.
-                                       (sent > adminSent ? (sent - adminSent): uint64_t(0)),
-                                       (recv > adminRecv ? (recv - adminRecv): uint64_t(0)));
-            adminSent = sent;
-            adminRecv = recv;
-            LOG_TRC("Doc [" << _docKey << "] added stats sent: " << sent << ", recv: " << recv << " bytes to totals.");
+            Admin::instance().addBytes(getDocKey(), deltaSent, deltaRecv);
         }
+
+        if (_storage && _lockCtx->needsRefresh(now))
+            refreshLock();
 #endif
 
         if (isSaving() &&
@@ -399,7 +457,7 @@ void DocumentBroker::pollThread()
             _poll->continuePolling() << ", ShutdownRequestFlag: " << SigUtil::getShutdownRequestFlag() <<
             ", TerminationFlag: " << SigUtil::getTerminationFlag() << ", closeReason: " << _closeReason << ". Flushing socket.");
 
-    if (_isModified)
+    if (isModified())
     {
         std::stringstream state;
         dumpState(state);
@@ -407,9 +465,9 @@ void DocumentBroker::pollThread()
     }
 
     // Flush socket data first.
-    constexpr int flushTimeoutMs = POLL_TIMEOUT_MS * 2; // ~1000ms
+    constexpr int64_t flushTimeoutMicroS = POLL_TIMEOUT_MICRO_S * 2; // ~1000ms
     LOG_INF("Flushing socket for doc ["
-            << _docKey << "] for " << flushTimeoutMs << " ms. stop: " << _stop
+            << _docKey << "] for " << flushTimeoutMicroS << " us. stop: " << _stop
             << ", continuePolling: " << _poll->continuePolling()
             << ", ShutdownRequestFlag: " << SigUtil::getShutdownRequestFlag()
             << ", TerminationFlag: " << SigUtil::getTerminationFlag()
@@ -418,11 +476,11 @@ void DocumentBroker::pollThread()
     while (_poll->getSocketCount())
     {
         const auto now = std::chrono::steady_clock::now();
-        const int elapsedMs = std::chrono::duration_cast<std::chrono::milliseconds>(now - flushStartTime).count();
-        if (elapsedMs > flushTimeoutMs)
+        const int64_t elapsedMicroS = std::chrono::duration_cast<std::chrono::microseconds>(now - flushStartTime).count();
+        if (elapsedMicroS > flushTimeoutMicroS)
             break;
 
-        _poll->poll(std::min(flushTimeoutMs - elapsedMs, POLL_TIMEOUT_MS / 5));
+        _poll->poll(std::min(flushTimeoutMicroS - elapsedMicroS, (int64_t)POLL_TIMEOUT_MICRO_S / 5));
     }
 
     LOG_INF("Finished flushing socket for doc [" << _docKey << "]. stop: " << _stop << ", continuePolling: " <<
@@ -563,7 +621,8 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     WopiStorage* wopiStorage = dynamic_cast<WopiStorage*>(_storage.get());
     if (wopiStorage != nullptr)
     {
-        std::unique_ptr<WopiStorage::WOPIFileInfo> wopifileinfo = wopiStorage->getWOPIFileInfo(session->getAuthorization());
+        std::unique_ptr<WopiStorage::WOPIFileInfo> wopifileinfo = wopiStorage->getWOPIFileInfo(
+            session->getAuthorization(), session->getCookies(), *_lockCtx);
         userId = wopifileinfo->getUserId();
         username = wopifileinfo->getUsername();
         userExtraInfo = wopifileinfo->getUserExtraInfo();
@@ -676,10 +735,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     session->setUserName(username);
     session->setUserExtraInfo(userExtraInfo);
     session->setWatermarkText(watermarkText);
-    if(!watermarkText.empty())
-        session->setHash(watermarkText);
-    else
-        session->setHash(0);
+    session->recalcCanonicalViewId(_sessions);
 
     // Basic file information was stored by the above getWOPIFileInfo() or getLocalFileInfo() calls
     const StorageBase::FileInfo fileInfo = _storage->getFileInfo();
@@ -708,9 +764,8 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
                     ", Actual: " << fileInfo.getModifiedTime());
 
             _documentChangedInStorage = true;
-            std::string message = "close: documentconflict";
-            if (_isModified)
-                message = "error: cmd=storage kind=documentconflict";
+            const std::string message = isModified() ? "error: cmd=storage kind=documentconflict"
+                                                     : "close: documentconflict";
 
             session->sendTextFrame(message);
             broadcastMessage(message);
@@ -722,7 +777,11 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
     // Let's load the document now, if not loaded.
     if (!_storage->isLoaded())
     {
-        std::string localPath = _storage->loadStorageFileToLocal(session->getAuthorization(), templateSource);
+        std::string localPath = _storage->loadStorageFileToLocal(
+            session->getAuthorization(), session->getCookies(), *_lockCtx, templateSource);
+
+        if (!_storage->updateLockState(session->getAuthorization(), session->getCookies(), *_lockCtx, true))
+            LOG_ERR("Failed to lock!");
 
 #if !MOBILEAPP
         // Check if we have a prefilter "plugin" for this document format
@@ -732,7 +791,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
             {
                 const std::string extension(plugin->getString("prefilter.extension"));
                 const std::string newExtension(plugin->getString("prefilter.newextension"));
-                const std::string commandLine(plugin->getString("prefilter.commandline"));
+                std::string commandLine(plugin->getString("prefilter.commandline"));
 
                 if (localPath.length() > extension.length()+1 &&
                     strcasecmp(localPath.substr(localPath.length() - extension.length() -1).data(), (std::string(".") + extension).data()) == 0)
@@ -742,27 +801,30 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
 
                     const std::string newRootPath = _storage->getRootFilePath() + "." + newExtension;
 
-                    std::vector<std::string> args(LOOLProtocol::tokenize(commandLine, ' '));
-                    std::string command(args[0]);
-                    args.erase(args.begin()); // strip the commmand
-
                     // The commandline must contain the space-separated substring @INPUT@ that is
                     // replaced with the input file name, and @OUTPUT@ for the output file name.
                     int inputs(0), outputs(0);
-                    for (auto it = args.begin(); it != args.end(); ++it)
+
+                    std::string input("@INPUT");
+                    size_t pos = commandLine.find(input);
+                    if (pos != std::string::npos)
                     {
-                        if (*it == "@INPUT@")
-                        {
-                            *it = _storage->getRootFilePath();
-                            ++inputs;
-                        }
-                        else if (*it == "@OUTPUT@")
-                        {
-                            *it = newRootPath;
-                            ++outputs;
-                        }
+                        commandLine.replace(pos, input.length(), _storage->getRootFilePath());
+                        ++inputs;
                     }
 
+                    std::string output("@OUTPUT@");
+                    pos = commandLine.find(output);
+                    if (pos != std::string::npos)
+                    {
+                        commandLine.replace(pos, output.length(), newRootPath);
+                        ++outputs;
+                    }
+
+                    StringVector args(LOOLProtocol::tokenize(commandLine, ' '));
+                    std::string command(args[0]);
+                    args.erase(args.begin()); // strip the command
+
                     if (inputs != 1 || outputs != 1)
                         throw std::exception();
 
@@ -798,9 +860,8 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         LOG_INF("SHA1 for DocKey [" << _docKey << "] of [" << LOOLWSD::anonymizeUrl(localPath) << "]: " <<
                 Poco::DigestEngine::digestToHex(sha1.digest()));
 
-        // LibreOffice can't open files with '#' in the name
         std::string localPathEncoded;
-        Poco::URI::encode(localPath, "#", localPathEncoded);
+        Poco::URI::encode(localPath, "#?", localPathEncoded);
         _uriJailed = Poco::URI(Poco::URI("file://"), localPathEncoded).toString();
         _uriJailedAnonym = Poco::URI(Poco::URI("file://"), LOOLWSD::anonymizeUrl(localPathEncoded)).toString();
 
@@ -831,6 +892,7 @@ bool DocumentBroker::load(const std::shared_ptr<ClientSession>& session, const s
         auto callDuration = wopiStorage->getWopiLoadDuration();
         // Add the time taken to check file info
         callDuration += getInfoCallDuration;
+        _wopiLoadDuration = std::chrono::duration_cast<std::chrono::milliseconds>(callDuration);
         const std::string msg = "stats: wopiloadduration " + std::to_string(callDuration.count());
         LOG_TRC("Sending to Client [" << msg << "].");
         session->sendTextFrame(msg);
@@ -917,7 +979,7 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId, bool su
     if (!success && !force)
     {
         LOG_ERR("Cannot store docKey [" << _docKey << "] as .uno:Save has failed in LOK.");
-        it->second->sendTextFrame("error: cmd=storage kind=savefailed");
+        it->second->sendTextFrameAndLogError("error: cmd=storage kind=savefailed");
         return false;
     }
 
@@ -944,10 +1006,10 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId, bool su
     if (!isSaveAs && newFileModifiedTime == _lastFileModifiedTime && !isRename)
     {
         // Nothing to do.
-        auto timeInSec = std::chrono::duration_cast<std::chrono::seconds>
+        const auto timeInSec = std::chrono::duration_cast<std::chrono::seconds>
                                             (std::chrono::system_clock::now() - _lastFileModifiedTime);
         LOG_DBG("Skipping unnecessary saving to URI [" << uriAnonym << "] with docKey [" << _docKey <<
-                "]. File last modified " << timeInSec.count() << " seconds ago.");
+                "]. File last modified " << timeInSec.count() << " seconds ago, timestamp unchanged.");
         _poll->wakeup();
         return true;
     }
@@ -955,13 +1017,19 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId, bool su
     LOG_DBG("Persisting [" << _docKey << "] after saving to URI [" << uriAnonym << "].");
 
     assert(_storage && _tileCache);
-    StorageBase::SaveResult storageSaveResult = _storage->saveLocalFileToStorage(auth, saveAsPath, saveAsFilename, isRename);
+    const StorageBase::SaveResult storageSaveResult = _storage->saveLocalFileToStorage(
+        auth, it->second->getCookies(), *_lockCtx, saveAsPath, saveAsFilename, isRename);
     if (storageSaveResult.getResult() == StorageBase::SaveResult::OK)
     {
+#if !MOBILEAPP
+        WopiStorage* wopiStorage = dynamic_cast<WopiStorage*>(_storage.get());
+        if (wopiStorage != nullptr)
+            Admin::instance().setDocWopiUploadDuration(_docKey, std::chrono::duration_cast<std::chrono::milliseconds>(wopiStorage->getWopiSaveDuration()));
+#endif
+
         if (!isSaveAs && !isRename)
         {
             // Saved and stored; update flags.
-            setModified(false);
             _lastFileModifiedTime = newFileModifiedTime;
             _lastSaveTime = std::chrono::steady_clock::now();
 
@@ -1023,14 +1091,14 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId, bool su
         // Make everyone readonly and tell everyone that storage is low on diskspace.
         for (const auto& sessionIt : _sessions)
         {
-            sessionIt.second->sendTextFrame("error: cmd=storage kind=savediskfull");
+            sessionIt.second->sendTextFrameAndLogError("error: cmd=storage kind=savediskfull");
         }
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::UNAUTHORIZED)
     {
         LOG_ERR("Cannot save docKey [" << _docKey << "] to storage URI [" << uriAnonym <<
                 "]. Invalid or expired access token. Notifying client.");
-        it->second->sendTextFrame("error: cmd=storage kind=saveunauthorized");
+        it->second->sendTextFrameAndLogError("error: cmd=storage kind=saveunauthorized");
     }
     else if (storageSaveResult.getResult() == StorageBase::SaveResult::FAILED)
     {
@@ -1045,9 +1113,8 @@ bool DocumentBroker::saveToStorageInternal(const std::string& sessionId, bool su
     {
         LOG_ERR("PutFile says that Document changed in storage");
         _documentChangedInStorage = true;
-        std::string message = "close: documentconflict";
-        if (_isModified)
-            message = "error: cmd=storage kind=documentconflict";
+        const std::string message
+            = isModified() ? "error: cmd=storage kind=documentconflict" : "close: documentconflict";
 
         broadcastMessage(message);
     }
@@ -1066,23 +1133,10 @@ void DocumentBroker::setLoaded()
     }
 }
 
-bool DocumentBroker::autoSave(const bool force, const bool dontSaveIfUnmodified)
+std::string DocumentBroker::getWriteableSessionId() const
 {
     assertCorrectThread();
 
-    LOG_TRC("autoSave(): forceful? " << force);
-    if (_sessions.empty() || _storage == nullptr || !_isLoaded ||
-        !_childProcess->isAlive() || (!_isModified && !force))
-    {
-        // Nothing to do.
-        LOG_TRC("Nothing to autosave [" << _docKey << "].");
-        return false;
-    }
-
-    // Remember the last save time, since this is the predicate.
-    LOG_TRC("Checking to autosave [" << _docKey << "].");
-
-    // Which session to use when auto saving ?
     std::string savingSessionId;
     for (auto& sessionIt : _sessions)
     {
@@ -1101,6 +1155,45 @@ bool DocumentBroker::autoSave(const bool force, const bool dontSaveIfUnmodified)
             break;
         }
     }
+    return savingSessionId;
+}
+
+void DocumentBroker::refreshLock()
+{
+    assertCorrectThread();
+
+    const std::string savingSessionId = getWriteableSessionId();
+    LOG_TRC("Refresh lock " << _lockCtx->_lockToken << " with session " << savingSessionId);
+
+    auto it = _sessions.find(savingSessionId);
+    if (it == _sessions.end())
+        LOG_ERR("No write-able session to refresh lock with");
+    else
+    {
+        std::shared_ptr<ClientSession> session = it->second;
+        if (!session || !_storage->updateLockState(session->getAuthorization(), session->getCookies(), *_lockCtx, true))
+            LOG_ERR("Failed to refresh lock");
+    }
+}
+
+bool DocumentBroker::autoSave(const bool force, const bool dontSaveIfUnmodified)
+{
+    assertCorrectThread();
+
+    LOG_TRC("autoSave(): forceful? " << force);
+    if (_sessions.empty() || _storage == nullptr || !_isLoaded ||
+        !_childProcess->isAlive() || (!isModified() && !force))
+    {
+        // Nothing to do.
+        LOG_TRC("Nothing to autosave [" << _docKey << "].");
+        return false;
+    }
+
+    // Remember the last save time, since this is the predicate.
+    LOG_TRC("Checking to autosave [" << _docKey << "].");
+
+    // Which session to use when auto saving ?
+    const std::string savingSessionId = getWriteableSessionId();
 
     bool sent = false;
     if (force)
@@ -1115,7 +1208,7 @@ bool DocumentBroker::autoSave(const bool force, const bool dontSaveIfUnmodified)
                            dontSaveIfUnmodified, /*isAutosave=*/false,
                            /*isExitSave=*/true);
     }
-    else if (_isModified)
+    else if (isModified())
     {
         const std::chrono::steady_clock::time_point now = std::chrono::steady_clock::now();
         const std::chrono::milliseconds::rep inactivityTimeMs = std::chrono::duration_cast<std::chrono::milliseconds>(now - _lastActivityTime).count();
@@ -1272,7 +1365,9 @@ size_t DocumentBroker::addSessionInternal(const std::shared_ptr<ClientSession>&
 
 #if !MOBILEAPP
     // Tell the admin console about this new doc
-    Admin::instance().addDoc(_docKey, getPid(), getFilename(), id, session->getUserName(), session->getUserId());
+    Admin::instance().addDoc(_docKey, getPid(), getFilename(), id, session->getUserName(),
+                             session->getUserId(), _childProcess->getSMapsFD());
+    Admin::instance().setDocWopiDownloadDuration(_docKey, _wopiLoadDuration);
 #endif
 
     // Add and attach the session.
@@ -1299,20 +1394,21 @@ size_t DocumentBroker::removeSession(const std::string& id)
             LOG_ERR("Invalid or unknown session [" << id << "] to remove.");
             return _sessions.size();
         }
+        std::shared_ptr<ClientSession> session = it->second;
 
         // Last view going away, can destroy.
         _markToDestroy = (_sessions.size() <= 1);
 
-        const bool lastEditableSession = !it->second->isReadOnly() && !haveAnotherEditableSession(id);
+        const bool lastEditableSession = !session->isReadOnly() && !haveAnotherEditableSession(id);
         static const bool dontSaveIfUnmodified = !LOOLWSD::getConfigValue<bool>("per_document.always_save_on_exit", false);
 
         LOG_INF("Removing session ["
                 << id << "] on docKey [" << _docKey << "]. Have " << _sessions.size()
-                << " sessions. IsReadOnly: " << it->second->isReadOnly()
-                << ", IsViewLoaded: " << it->second->isViewLoaded() << ", IsWaitDisconnected: "
-                << it->second->inWaitDisconnected() << ", MarkToDestroy: " << _markToDestroy
-                << ", LastEditableSession: " << lastEditableSession
-                << ", dontSaveIfUnmodified: " << dontSaveIfUnmodified);
+                << " sessions. IsReadOnly: " << session->isReadOnly()
+                << ", IsViewLoaded: " << session->isViewLoaded() << ", IsWaitDisconnected: "
+                << session->inWaitDisconnected() << ", MarkToDestroy: " << _markToDestroy
+                << ", LastEditableSession: " << lastEditableSession << ", DontSaveIfUnmodified: "
+                << dontSaveIfUnmodified << ", IsPossiblyModified: " << isPossiblyModified());
 
         // If last editable, save and don't remove until after uploading to storage.
         if (!lastEditableSession || !autoSave(isPossiblyModified(), dontSaveIfUnmodified))
@@ -1341,7 +1437,16 @@ void DocumentBroker::disconnectSessionInternal(const std::string& id)
             LOOLWSD::dumpEndSessionTrace(getJailId(), id, _uriOrig);
 #endif
 
-            LOG_TRC("Disconnect session internal " << id);
+            LOG_TRC("Disconnect session internal " << id <<
+                    " destroy? " << _markToDestroy <<
+                    " locked? " << _lockCtx->_isLocked);
+
+            if (_markToDestroy && // last session to remove; FIXME: Editable?
+                _lockCtx->_isLocked && _storage)
+            {
+                if (!_storage->updateLockState(it->second->getAuthorization(), it->second->getCookies(), *_lockCtx, false))
+                    LOG_ERR("Failed to unlock!");
+            }
 
             bool hardDisconnect;
             if (it->second->inWaitDisconnected())
@@ -1386,6 +1491,7 @@ void DocumentBroker::finalRemoveSession(const std::string& id)
 
             // Remove. The caller must have a reference to the session
             // in question, lest we destroy from underneath them.
+            it->second->dispose();
             _sessions.erase(it);
             const size_t count = _sessions.size();
 
@@ -1415,6 +1521,39 @@ void DocumentBroker::finalRemoveSession(const std::string& id)
     }
 }
 
+std::shared_ptr<ClientSession> DocumentBroker::createNewClientSession(
+    const std::shared_ptr<ProtocolHandlerInterface> &ws,
+    const std::string& id,
+    const Poco::URI& uriPublic,
+    const bool isReadOnly,
+    const RequestDetails &requestDetails)
+{
+    try
+    {
+        // Now we have a DocumentBroker and we're ready to process client commands.
+        if (ws)
+        {
+            const std::string statusReady = "statusindicator: ready";
+            LOG_TRC("Sending to Client [" << statusReady << "].");
+            ws->sendTextMessage(statusReady.c_str(), statusReady.size());
+        }
+
+        // In case of WOPI, if this session is not set as readonly, it might be set so
+        // later after making a call to WOPI host which tells us the permission on files
+        // (UserCanWrite param).
+        auto session = std::make_shared<ClientSession>(ws, id, shared_from_this(), uriPublic, isReadOnly, requestDetails);
+        session->construct();
+
+        return session;
+    }
+    catch (const std::exception& exc)
+    {
+        LOG_WRN("Exception while preparing session [" << id << "]: " << exc.what());
+    }
+
+    return nullptr;
+}
+
 void DocumentBroker::addCallback(const SocketPoll::CallbackFn& fn)
 {
     _poll->addCallback(fn);
@@ -1478,17 +1617,6 @@ bool DocumentBroker::handleInput(const std::vector<char>& payload)
             LOG_CHECK_RET(kind != "", false);
             Util::alertAllUsers(cmd, kind);
         }
-#if !MOBILEAPP
-        else if (command == "procmemstats:")
-        {
-            int dirty;
-            if (message->getTokenInteger("dirty", dirty))
-            {
-                Admin::instance().updateMemoryDirty(
-                    _docKey, dirty + getMemorySize()/1024);
-            }
-        }
-#endif
         else
         {
             LOG_ERR("Unexpected message: [" << msg << "].");
@@ -1703,7 +1831,6 @@ void DocumentBroker::sendRequestedTiles(const std::shared_ptr<ClientSession>& se
     float tilesOnFlyUpperLimit = 0;
     if (normalizedVisArea.hasSurface() && session->getTileWidthInTwips() != 0 && session->getTileHeightInTwips() != 0)
     {
-
         const int tilesFitOnWidth = std::ceil(normalizedVisArea.getRight() / session->getTileWidthInTwips()) -
                                     std::ceil(normalizedVisArea.getLeft() / session->getTileWidthInTwips()) + 1;
         const int tilesFitOnHeight = std::ceil(normalizedVisArea.getBottom() / session->getTileHeightInTwips()) -
@@ -1899,6 +2026,7 @@ void DocumentBroker::setModified(const bool value)
 #endif
     }
 
+    // Set the X-LOOL-WOPI-IsModifiedByUser header flag sent to storage.
     _storage->setUserModified(value);
 }
 
@@ -1934,14 +2062,14 @@ bool DocumentBroker::forwardToChild(const std::string& viewId, const std::string
     {
         assert(!_uriJailed.empty());
 
-        std::vector<std::string> tokens = LOOLProtocol::tokenize(msg);
-        if (tokens.size() > 1 && tokens[1] == "load")
+        StringVector tokens = LOOLProtocol::tokenize(msg);
+        if (tokens.size() > 1 && tokens.equals(1, "load"))
         {
             // The json options must come last.
             msg = tokens[0] + ' ' + tokens[1] + ' ' + tokens[2];
             msg += " jail=" + _uriJailed;
             msg += " xjail=" + _uriJailedAnonym;
-            msg += ' ' + Poco::cat(std::string(" "), tokens.begin() + 3, tokens.end());
+            msg += ' ' + tokens.cat(std::string(" "), 3);
         }
 
         _childProcess->sendTextFrame(msg);
@@ -2022,7 +2150,7 @@ void DocumentBroker::shutdownClients(const std::string& closeReason)
             else
             {
                 // Notify the client and disconnect.
-                session->shutdown(WebSocketHandler::StatusCodes::ENDPOINT_GOING_AWAY, closeReason);
+                session->shutdownGoingAway(closeReason);
 
                 // Remove session, save, and mark to destroy.
                 removeSession(session->getId());
@@ -2104,13 +2232,14 @@ void DocumentBroker::getIOStats(uint64_t &sent, uint64_t &recv)
     assertCorrectThread();
     for (const auto& sessionIt : _sessions)
     {
-        uint64_t s, r;
+        uint64_t s = 0, r = 0;
         sessionIt.second->getIOStats(s, r);
         sent += s;
         recv += r;
     }
 }
 
+#if !MOBILEAPP
 static std::atomic<size_t> NumConverters;
 
 size_t ConvertToBroker::getInstanceCount()
@@ -2120,10 +2249,65 @@ size_t ConvertToBroker::getInstanceCount()
 
 ConvertToBroker::ConvertToBroker(const std::string& uri,
                                  const Poco::URI& uriPublic,
-                                 const std::string& docKey)
-    : DocumentBroker(uri, uriPublic, docKey)
+                                 const std::string& docKey,
+                                 const std::string& format,
+                                 const std::string& sOptions) :
+    DocumentBroker(ChildType::Batch, uri, uriPublic, docKey),
+    _format(format),
+    _sOptions(sOptions)
 {
+    static const int limit_convert_secs = LOOLWSD::getConfigValue<int>("per_document.limit_convert_secs", 100);
     NumConverters++;
+    _limitLifeSeconds = limit_convert_secs;
+}
+
+bool ConvertToBroker::startConversion(SocketDisposition &disposition, const std::string &id)
+{
+    std::shared_ptr<ConvertToBroker> docBroker = std::static_pointer_cast<ConvertToBroker>(shared_from_this());
+
+    // Create a session to load the document.
+    const bool isReadOnly = true;
+    // FIXME: associate this with moveSocket (?)
+    std::shared_ptr<ProtocolHandlerInterface> nullPtr;
+    RequestDetails requestDetails("convert-to");
+    _clientSession = std::make_shared<ClientSession>(nullPtr, id, docBroker, getPublicUri(), isReadOnly, requestDetails);
+    _clientSession->construct();
+
+    if (!_clientSession)
+        return false;
+
+    disposition.setMove([docBroker] (const std::shared_ptr<Socket> &moveSocket)
+        {
+            // Perform all of this after removing the socket
+
+            // Make sure the thread is running before adding callback.
+            docBroker->startThread();
+
+            // We no longer own this socket.
+            moveSocket->setThreadOwner(std::thread::id(0));
+
+            docBroker->addCallback([docBroker, moveSocket]()
+                 {
+                     auto streamSocket = std::static_pointer_cast<StreamSocket>(moveSocket);
+                     docBroker->_clientSession->setSaveAsSocket(streamSocket);
+
+                     // Move the socket into DocBroker.
+                     docBroker->addSocketToPoll(moveSocket);
+
+                     // First add and load the session.
+                     docBroker->addSession(docBroker->_clientSession);
+
+                     // Load the document manually and request saving in the target format.
+                     std::string encodedFrom;
+                     Poco::URI::encode(docBroker->getPublicUri().getPath(), "", encodedFrom);
+                     const std::string _load = "load url=" + encodedFrom;
+                     std::vector<char> loadRequest(_load.begin(), _load.end());
+                     docBroker->_clientSession->handleMessage(loadRequest);
+
+                     // Save is done in the setLoaded
+                 });
+        });
+    return true;
 }
 
 void ConvertToBroker::dispose()
@@ -2156,6 +2340,32 @@ void ConvertToBroker::removeFile(const std::string &uriOrig)
     }
 }
 
+void ConvertToBroker::setLoaded()
+{
+    DocumentBroker::setLoaded();
+
+    // FIXME: Check for security violations.
+    Poco::Path toPath(getPublicUri().getPath());
+    toPath.setExtension(_format);
+
+    // file:///user/docs/filename.ext normally, file:///<jail-root>/user/docs/filename.ext in the nocaps case
+    const std::string toJailURL = "file://" +
+        (LOOLWSD::NoCapsForKit? getJailRoot(): "") +
+        std::string(JAILED_DOCUMENT_ROOT) + toPath.getFileName();
+
+    std::string encodedTo;
+    Poco::URI::encode(toJailURL, "", encodedTo);
+
+    // Convert it to the requested format.
+    const std::string saveAsCmd = "saveas url=" + encodedTo + " format=" + _format + " options=" + _sOptions;
+
+    // Send the save request ...
+    std::vector<char> saveasRequest(saveAsCmd.begin(), saveAsCmd.end());
+
+    _clientSession->handleMessage(saveasRequest);
+}
+#endif
+
 std::vector<std::shared_ptr<ClientSession>> DocumentBroker::getSessionsTestOnlyUnsafe()
 {
     std::vector<std::shared_ptr<ClientSession>> result;
@@ -2168,7 +2378,7 @@ void DocumentBroker::dumpState(std::ostream& os)
 {
     std::unique_lock<std::mutex> lock(_mutex);
 
-    uint64_t sent, recv;
+    uint64_t sent = 0, recv = 0;
     getIOStats(sent, recv);
 
     auto now = std::chrono::steady_clock::now();
@@ -2186,7 +2396,7 @@ void DocumentBroker::dumpState(std::ostream& os)
                 now - _threadStart).count() << "s";
     os << "\n  sent: " << sent;
     os << "\n  recv: " << recv;
-    os << "\n  modified?: " << _isModified;
+    os << "\n  modified?: " << isModified();
     os << "\n  jail id: " << _jailId;
     os << "\n  filename: " << LOOLWSD::anonymizeUrl(_filename);
     os << "\n  public uri: " << _uriPublic.toString();
@@ -2194,12 +2404,18 @@ void DocumentBroker::dumpState(std::ostream& os)
     os << "\n  doc key: " << _docKey;
     os << "\n  doc id: " << _docId;
     os << "\n  num sessions: " << _sessions.size();
-    const std::time_t t = std::chrono::system_clock::to_time_t(
-        std::chrono::time_point_cast<std::chrono::seconds>(
-            std::chrono::system_clock::now() + (_lastSaveTime - now)));
-    os << "\n  last saved: " << std::ctime(&t);
+    os << "\n  thread start: " << Util::getSteadyClockAsString(_threadStart);
+    os << "\n  last saved: " << Util::getSteadyClockAsString(_lastSaveTime);
+    os << "\n  last save request: " << Util::getSteadyClockAsString(_lastSaveRequestTime);
+    os << "\n  last save response: " << Util::getSteadyClockAsString(_lastSaveResponseTime);
+    os << "\n  last modified: " << Util::getHttpTime(_documentLastModifiedTime);
+    os << "\n  file last modified: " << Util::getHttpTime(_lastFileModifiedTime);
+    if (_limitLifeSeconds)
+        os << "\n  life limit in seconds: " << _limitLifeSeconds;
+    os << "\n  idle time: " << getIdleTimeSecs();
     os << "\n  cursor " << _cursorPosX << ", " << _cursorPosY
       << "( " << _cursorWidth << "," << _cursorHeight << ")\n";
+    _lockCtx->dumpState(os);
     if (_tileCache)
         _tileCache->dumpState(os);
 
