@@ -8,11 +8,12 @@
  */
 
 // Storage abstraction.
-#ifndef INCLUDED_STORAGE_HPP
-#define INCLUDED_STORAGE_HPP
+
+#pragma once
 
 #include <set>
 #include <string>
+#include <chrono>
 
 #include <Poco/URI.h>
 #include <Poco/Util/Application.h>
@@ -24,6 +25,30 @@
 #include "Util.hpp"
 #include <common/Authorization.hpp>
 
+/// Represents whether the underlying file is locked
+/// and with what token.
+struct LockContext
+{
+    /// Do we have support for locking for a storage.
+    bool        _supportsLocks;
+    /// Do we own the (leased) lock currently
+    bool        _isLocked;
+    /// Name if we need it to use consistently for locking
+    std::string _lockToken;
+    /// Time of last successful lock (re-)acquisition
+    std::chrono::steady_clock::time_point _lastLockTime;
+
+    LockContext() : _supportsLocks(false), _isLocked(false) { }
+
+    /// one-time setup for supporting locks & create token
+    void initSupportsLocks();
+
+    /// do we need to refresh our lock ?
+    bool needsRefresh(const std::chrono::steady_clock::time_point &now) const;
+
+    void dumpState(std::ostream& os) const;
+};
+
 /// Base class of all Storage abstractions.
 class StorageBase
 {
@@ -190,13 +215,26 @@ public:
 
     std::string getFileExtension() const { return Poco::Path(_fileInfo.getFilename()).getExtension(); }
 
+    /// Update the locking state (check-in/out) of the associated file
+    virtual bool updateLockState(const Authorization& auth, const std::string& cookies,
+                                 LockContext& lockCtx, bool lock)
+        = 0;
+
     /// Returns a local file path for the given URI.
     /// If necessary copies the file locally first.
-    virtual std::string loadStorageFileToLocal(const Authorization& auth, const std::string& templateUri) = 0;
+    virtual std::string loadStorageFileToLocal(const Authorization& auth,
+                                               const std::string& /*cookies*/, LockContext& lockCtx,
+                                               const std::string& templateUri)
+        = 0;
 
     /// Writes the contents of the file back to the source.
+    /// @param cookies A string representing key=value pairs that are set as cookies.
     /// @param savedFile When the operation was saveAs, this is the path to the file that was saved.
-    virtual SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& saveAsPath, const std::string& saveAsFilename, const bool isRename) = 0;
+    virtual SaveResult
+    saveLocalFileToStorage(const Authorization& auth, const std::string& /*cookies*/,
+                           LockContext& lockCtx, const std::string& saveAsPath,
+                           const std::string& saveAsFilename, const bool isRename)
+        = 0;
 
     static size_t getFileSize(const std::string& filename);
 
@@ -241,6 +279,9 @@ private:
 
     static bool FilesystemEnabled;
     static bool WopiEnabled;
+    /// If true, use only the WOPI URL for whether to use SSL to talk to storage server
+    static bool SSLAsScheme;
+    /// If true, force SSL communication with storage server
     static bool SSLEnabled;
     /// Allowed/denied WOPI hosts, if any and if WOPI is enabled.
     static Util::RegexListMatcher WopiHosts;
@@ -283,9 +324,19 @@ public:
     /// obtained using getFileInfo method
     std::unique_ptr<LocalFileInfo> getLocalFileInfo();
 
-    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& templateUri) override;
+    bool updateLockState(const Authorization&, const std::string&, LockContext&, bool) override
+    {
+        return true;
+    }
 
-    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& saveAsPath, const std::string& saveAsFilename, const bool isRename) override;
+    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& /*cookies*/,
+                                       LockContext& lockCtx,
+                                       const std::string& templateUri) override;
+
+    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& /*cookies*/,
+                                      LockContext& lockCtx, const std::string& saveAsPath,
+                                      const std::string& saveAsFilename,
+                                      const bool isRename) override;
 
 private:
     /// True if the jailed file is not linked but copied.
@@ -302,6 +353,7 @@ public:
                 const std::string& jailPath) :
         StorageBase(uri, localStorePath, jailPath),
         _wopiLoadDuration(0),
+        _wopiSaveDuration(0),
         _reuseCookies(false)
     {
         const auto& app = Poco::Util::Application::instance();
@@ -347,6 +399,7 @@ public:
                      const TriState disableChangeTrackingShow,
                      const TriState disableChangeTrackingRecord,
                      const TriState hideChangeTrackingControls,
+                     const bool supportsLocks,
                      const bool supportsRename,
                      const bool userCanRename,
                      const std::chrono::duration<double> callDuration)
@@ -374,6 +427,7 @@ public:
               _disableChangeTrackingShow(disableChangeTrackingShow),
               _disableChangeTrackingRecord(disableChangeTrackingRecord),
               _hideChangeTrackingControls(hideChangeTrackingControls),
+              _supportsLocks(supportsLocks),
               _supportsRename(supportsRename),
               _userCanRename(userCanRename),
               _callDuration(callDuration)
@@ -427,6 +481,8 @@ public:
 
         bool getSupportsRename() const { return _supportsRename; }
 
+        bool getSupportsLocks() const { return _supportsLocks; }
+
         bool getUserCanRename() const { return _userCanRename; }
 
         std::string& getHideUserList() { return _hideUserList; }
@@ -492,6 +548,8 @@ public:
         TriState _disableChangeTrackingRecord;
         /// If we should hide change-tracking commands for this user.
         TriState _hideChangeTrackingControls;
+        /// If WOPI host supports locking
+        bool _supportsLocks;
         /// If WOPI host supports rename
         bool _supportsRename;
         /// If user is allowed to rename the document
@@ -505,19 +563,32 @@ public:
     /// provided during the initial creation of the WOPI storage.
     /// Also extracts the basic file information from the response
     /// which can then be obtained using getFileInfo()
-    std::unique_ptr<WOPIFileInfo> getWOPIFileInfo(const Authorization& auth);
+    /// Also sets up the locking context for future operations.
+    std::unique_ptr<WOPIFileInfo> getWOPIFileInfo(const Authorization& auth,
+                                                  const std::string& cookies, LockContext& lockCtx);
+
+    /// Update the locking state (check-in/out) of the associated file
+    bool updateLockState(const Authorization& auth, const std::string& cookies,
+                         LockContext& lockCtx, bool lock) override;
 
     /// uri format: http://server/<...>/wopi*/files/<id>/content
-    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& templateUri) override;
+    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& /*cookies*/,
+                                       LockContext& lockCtx,
+                                       const std::string& templateUri) override;
 
-    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& saveAsPath, const std::string& saveAsFilename, const bool isRename) override;
+    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& /*cookies*/,
+                                      LockContext& lockCtx, const std::string& saveAsPath,
+                                      const std::string& saveAsFilename,
+                                      const bool isRename) override;
 
     /// Total time taken for making WOPI calls during load
     std::chrono::duration<double> getWopiLoadDuration() const { return _wopiLoadDuration; }
+    std::chrono::duration<double> getWopiSaveDuration() const { return _wopiSaveDuration; }
 
 private:
     // Time spend in loading the file from storage
     std::chrono::duration<double> _wopiLoadDuration;
+    std::chrono::duration<double> _wopiSaveDuration;
     /// Whether or not to re-use cookies from the browser for the WOPI requests.
     bool _reuseCookies;
 };
@@ -540,14 +611,22 @@ public:
     // Implement me
     // WebDAVFileInfo getWebDAVFileInfo(const Poco::URI& uriPublic);
 
-    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& templateUri) override;
+    bool updateLockState(const Authorization&, const std::string&, LockContext&, bool) override
+    {
+        return true;
+    }
 
-    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& saveAsPath, const std::string& saveAsFilename, const bool isRename) override;
+    std::string loadStorageFileToLocal(const Authorization& auth, const std::string& /*cookies*/,
+                                       LockContext& lockCtx,
+                                       const std::string& templateUri) override;
+
+    SaveResult saveLocalFileToStorage(const Authorization& auth, const std::string& /*cookies*/,
+                                      LockContext& lockCtx, const std::string& saveAsPath,
+                                      const std::string& saveAsFilename,
+                                      const bool isRename) override;
 
 private:
     std::unique_ptr<AuthBase> _authAgent;
 };
 
-#endif
-
 /* vim:set shiftwidth=4 softtabstop=4 expandtab: */
